/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "cvaviio.h"
#include "vfw.h"

// macro
#define SAFE_RELEASE(x) { if (x) x->Release(); x = NULL; }

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    CvAviIo()
//    Purpose:
//        Constructor. Initialize the AVI frame list and variables.
//    Context:
//    Parameters:
//    Returns: 
//	  Authors:
//F*/
CvAviIo::CvAviIo(void) 
{
	m_isAudioLoad = false;
	m_isVideoLoad = false;

	// video
	m_nImageWidth = 0;
	m_nImageHeight = 0;
	m_nImageBand = 0;
	m_fVideoFrameRate = 0.0f;
	m_nVideoFrameNum = 0;				// total frame number of the AVI file
	m_pVideoData = NULL;
	m_nCurrentFrameNumber = 0;			// current frame number in the memory
	m_nStartFrameID = 0;				// the global frame no. of the first frame in the memory

	// audio
	m_nAudioFrameNum = 0;
	m_nAudioSampleRate = 0;
	m_nAudioSampleSize = 0;
	m_pAudioData = NULL;

	return;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    ~CvAviIo()
//    Purpose:
//        Destructor. Release the AVI frames.
//    Context:
//    Parameters:
//    Returns: 
//	  Authors:
//F*/
CvAviIo::~CvAviIo(void)
{
	ClearAvi();
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    LoadVideoFrame(const char szFileName[], int nStart, int nEnd)
//    Purpose:
//        Load AVI frames of the specified file from nStart (include) to nEnd (not include).
//    Context:
//    Parameters:
//			szFileName	--- [I] AVI file name
//			nStart		--- [I] start frame number of the AVI frames loaded (include)
//			nEnd		--- [I] end frame number of the AVI frames loaded (not include)
//    Returns: 
//			the number of frames loaded
//	  Authors:
//			Haibing Ren, Tsinghua Unversity
//			Luhong Liang, Intel China Research Center
//F*/

int CvAviIo::LoadVideoFrame(const char szFileName[], int nStart, int nEnd)
{
	// copy file name & options
	strcpy(m_szCurrentAVIFile, szFileName);

	PAVIFILE pAviFile =NULL;           
	PAVISTREAM pVideoStream =NULL;
	AVISTREAMINFO videoStreamInfo;
	BITMAPINFOHEADER bmpInfo;
	PGETFRAME pGetFrame;

	AVIFileInit();
	switch (AVIFileOpen(&pAviFile, szFileName, OF_READ, 0L)) {
		case AVIERR_BADFORMAT:
			Exception("The file couldn't be read, indicating a corrupt file or an unrecognized format.\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		case AVIERR_MEMORY:
			Exception("The file could not be opened because of insufficient memory.\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		case AVIERR_FILEREAD:
			Exception("A disk error occurred while reading the file.\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		case AVIERR_FILEOPEN:
			Exception("A disk error occurred while opening the file.\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		case REGDB_E_CLASSNOTREG:
			Exception("According to the registry, the type of file specified in AVIFileOpen does not have a handler to process it.\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
	}
	
	//Open video stream
	AVIFileGetStream(pAviFile, &pVideoStream, streamtypeVIDEO, 0);
	AVIFileRelease(pAviFile);

	if(pVideoStream != NULL)
	{
		//get video stream information
		AVIStreamInfo(pVideoStream, &videoStreamInfo, sizeof(AVISTREAMINFO));
		m_nVideoFrameNum = videoStreamInfo.dwLength;
		m_fVideoFrameRate = (float)videoStreamInfo.dwRate / (float)videoStreamInfo.dwScale;

		//read video format (BITMAPINFOHEAD)
		BYTE *lpVideoBuffer;
		long cbVideoBuffer;
		AVIStreamReadFormat(pVideoStream, AVIStreamStart(pVideoStream), NULL, &cbVideoBuffer);
		lpVideoBuffer = new BYTE[cbVideoBuffer];
		AVIStreamReadFormat(pVideoStream, AVIStreamStart(pVideoStream), lpVideoBuffer, &cbVideoBuffer);
		memcpy(&bmpInfo, lpVideoBuffer, sizeof(BITMAPINFOHEADER));
		delete lpVideoBuffer;
		m_nImageWidth = bmpInfo.biWidth;
		m_nImageHeight = bmpInfo.biHeight;
		m_nImageBand = bmpInfo.biBitCount / 8;
		if (m_nImageBand !=1 && m_nImageBand != 3) {
			Exception("Only support 8-bit or 24-bit video!\nIn <CvAviIo::LoadVideoFrame()>.");
			AVIStreamRelease(pVideoStream);
			AVIFileExit();
			return 0;
		}

		// get the video frame pointer
		pGetFrame = AVIStreamGetFrameOpen(pVideoStream, NULL);
		if(pGetFrame == NULL) {
			Exception("Get frame open error!\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		}

		// read the avi file to the memory and change the r g b sequence
		nStart = nStart < 0 ? 0 : nStart;
		m_nStartFrameID = nStart;
		nEnd = nEnd > m_nVideoFrameNum ? m_nVideoFrameNum : nEnd;
		nEnd = (nEnd-nStart > AVSR_MAX_AVI_LENGTH_IN_MEMORY) ? nStart+AVSR_MAX_AVI_LENGTH_IN_MEMORY : nEnd;
		m_nCurrentFrameNumber = 0;
		m_pVideoData = new IplImage*[AVSR_MAX_AVI_LENGTH_IN_MEMORY];
		if(m_pVideoData == NULL){
			Exception("Can't allocate video data buffer!\nNot enough memory!\nIn <CvAviIo::LoadVideoFrame()>.");
			return 0;
		}		
		int i, j;
		for(i=0; i<AVSR_MAX_AVI_LENGTH_IN_MEMORY; i++)
			m_pVideoData[i] = NULL;
		for(i=nStart; i<nEnd; i++) {
			// get image frame
			BYTE *lpstream = (BYTE *)AVIStreamGetFrame(pGetFrame, i);
			if (lpstream == NULL) {
				m_nVideoFrameNum --;
				continue;
			}
			lpstream = lpstream + sizeof(BITMAPINFOHEADER);		// discard the header
			// load image
			IplImage *pBufImage;
			int linewidth;
			if (m_nImageBand == 1) {
				pBufImage = cvCreateImage(cvSize(m_nImageWidth, m_nImageHeight), IPL_DEPTH_8U, 1);
				if (pBufImage == NULL) {
					Exception("Can't create frame image!\nIn <CvAviIo::LoadVideoFrame()>.");
					return 0;
				}
				lpstream += sizeof(RGBQUAD) * 256;
				// horizontally reflect
				linewidth = ((m_nImageWidth + 3) >> 2 ) << 2;
				for (j=0; j<m_nImageHeight; j++) {
					memcpy(pBufImage->imageData+(m_nImageHeight-j-1)*linewidth, lpstream, linewidth);
					lpstream += linewidth;
				}
			}
			else {
				pBufImage = cvCreateImage(cvSize(m_nImageWidth, m_nImageHeight), IPL_DEPTH_8U, 3);
				if (pBufImage == NULL) {
					Exception("Can't create frame image!\nIn <CvAviIo::LoadVideoFrame()>.");
					return 0;
				}
				// horizontally reflect
				linewidth = ((m_nImageWidth * 3 + 3) >> 2 ) << 2;
				for (j=0; j<m_nImageHeight; j++) {
					memcpy(pBufImage->imageData+(m_nImageHeight-j-1)*linewidth, lpstream, linewidth);
					lpstream += linewidth;
				}
			}
			// set video data
			m_pVideoData[m_nCurrentFrameNumber] = pBufImage;
			m_nCurrentFrameNumber ++;
		}
		// close the video stream 
		AVIStreamGetFrameClose(pGetFrame);
		AVIStreamRelease(pVideoStream);
		AVIFileExit();
		
		if (m_nCurrentFrameNumber > 0) {
			m_isVideoLoad = true;
		}

		return m_nCurrentFrameNumber;
		
	}else{
		// Load video stream failed
		AVIFileExit();
		m_isVideoLoad = false;
		return 0;
	}
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: int LoadAvi(const char szFileName[], bool bAudio, bool bVideo)
//    Purpose:
//        Load AVI image frames. If the frame number is larger than the maximial limition, load 
//			one segment.
//    Context:
//    Parameters:
//			szFileName	--- [I] AVI file name
//			bAudio ---		[I] if the audio data will be loaded 
//			bVideo ---		[I] if the video data will be loaded 
//    Returns: 
//	  Authors:
//F*/
bool CvAviIo::LoadAvi(const char szFileName[], bool bAudio, bool bVideo)
{
	if (bAudio == false && bVideo == false) return false;

	ClearAvi();

	// get first segement of video
	if (bVideo) {
		if (LoadVideoFrame(szFileName, 0, AVSR_MAX_AVI_LENGTH_IN_MEMORY) < 1) {
			ClearAvi();
			return false;
		}
	}

	// load audio data
	if (bAudio) {
		if (!ReadWavFromAvi(szFileName)) {
			ClearAvi();
			return false;
		}
	}

	return true;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    GetImage(int nFrameNum)
//    Purpose:
//        Load AVI image frames. If the frame number is larger than the maximial limition, load 
//			one segment.
//    Context:
//    Parameters:
//			szFileName	--- [I] AVI file name
//			nImageWidth --- [O] image width of the AVI
//			nImageHeight -- [O] image height of the AVI
//			nFrameNum	--- [O] frame number of the AVI
//    Returns: 
//			nubmer of frames loaded
//	  Authors:
//F*/
IplImage *CvAviIo::GetImage(int nFrameNum)
{
	if (!m_isVideoLoad) {
		Exception("No AVI video data loaded!\nIn <CvAviIo::GetImage()>.");
		return NULL;
	}

	if (nFrameNum < 0 || nFrameNum >= m_nVideoFrameNum) {
		Exception("Invalid Frame number!\nIn <CvAviIo::GetImage()>.");
		return NULL;
	}

	if (nFrameNum >= m_nStartFrameID && nFrameNum < m_nStartFrameID+m_nCurrentFrameNumber) {
		return m_pVideoData[nFrameNum-m_nStartFrameID];
	}
	else {
		// load image
		if (LoadVideoFrame(m_szCurrentAVIFile, nFrameNum, m_nVideoFrameNum) > 0) {
			return m_pVideoData[0];
		}
		else {
			return NULL;
		}
	}
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    GetVideoInfo(int &nImageWidth, int &nImageHeight, int &nImageBand, int &nFrameNum, float &fFrameRate)
//    Purpose:
//        Get information about the image of the loaded AVI sequence.
//    Context:
//    Parameters:
//			nImageWidth --- [O] image width of the AVI
//			nImageHeight -- [O] image height of the AVI
//			nImageBand	--- [O] color band of the AVI image
//			nFrameNum	--- [O] frame number of the AVI
//			fFrameRate	--- [O] frame rate in Hz
//    Returns: 
//			nubmer of frames loaded
//	  Authors:
//F*/
bool CvAviIo::GetVideoInfo(int &nImageWidth, int &nImageHeight, int &nImageBand, int &nFrameNum, float &fFrameRate)
{
	if (!m_isVideoLoad || m_nCurrentFrameNumber <= 0 || m_pVideoData[0] == NULL) {
		Exception("No AVI data loaded!\nIn <CvAviIo::GetVideoInfo()>.");
		return FALSE;
	}

	nImageWidth = m_pVideoData[0]->width;
	nImageHeight = m_pVideoData[0]->height;
	nImageBand = m_pVideoData[0]->nChannels;
	nFrameNum = m_nVideoFrameNum;
	fFrameRate = m_fVideoFrameRate;

	return TRUE;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    ClearAvi(void)
//    Purpose:
//        Clear the loaded AVI image.
//    Context:
//    Parameters:
//    Returns: 
//			nubmer of frames loaded
//	  Authors:
//F*/
void CvAviIo::ClearAvi(void)
{
	m_isAudioLoad = false;
	m_isVideoLoad = false;
	m_szCurrentAVIFile[0] = 0; 

	// video
	m_nImageWidth = 0;
	m_nImageHeight = 0;
	m_nImageBand = 0;
	m_fVideoFrameRate = 0.0f;
	m_nVideoFrameNum = 0;				// total frame number of the AVI file
	if(m_pVideoData != NULL)
	{
		for (int i=0; i<AVSR_MAX_AVI_LENGTH_IN_MEMORY; i++) {
			if (m_pVideoData[i] != NULL) {
				cvReleaseImage(&(m_pVideoData[i]));
				m_pVideoData[i] = NULL;
			}
		}
		delete []m_pVideoData;
		m_pVideoData = NULL;
	}
	m_nCurrentFrameNumber = 0;			// current frame number in the memory
	m_nStartFrameID = 0;				// the global frame no. of the first frame in the memory

	// audio
	m_nAudioFrameNum = 0;
	m_nAudioSampleRate = 0;
	m_nAudioSampleSize = 0;
	if (m_pAudioData != NULL) {
		delete m_pAudioData;
		m_pAudioData = NULL;
	}
	
	return;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:  ReadWavFromAvi(const char lpszAVIFileName[], int nMode=LChannel)
//    Purpose:
//			Read one channel of wave data from AVI file
//    Context:
//    Parameters:
//			lpszAVIFileName	---- [I] file name to read
//			nMode			---- [I] options to select the channel {Audio_LChannel, Audio_RChannel, Audio_BChannel}
//    Returns: 
//			wave data buffer
//	  Authors:
//F*/
bool CvAviIo::ReadWavFromAvi(const char lpszAVIFileName[], int nMode)
{
	if(!lpszAVIFileName || lpszAVIFileName[0] == 0) { //parameter illegal
		Exception("Invalid input AVI file name!\nIn <CvAviIo::ReadWavFromAvi()>.");
		return FALSE;
	}
	
	// open AVI
	PAVIFILE pAviFile;           
	PAVISTREAM pAudioStream;
	AVISTREAMINFO audioStreamInfo;
	WAVEFORMATEX audioFormatInfo;
	LONG lStreamSize; 
	
	AVIFileInit();
	switch (AVIFileOpen(&pAviFile, lpszAVIFileName, OF_READ, 0L)) {
		case AVIERR_BADFORMAT:
			Exception("The file couldn't be read, indicating a corrupt file or an unrecognized format.\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		case AVIERR_MEMORY:
			Exception("The file could not be opened because of insufficient memory.\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		case AVIERR_FILEREAD:
			Exception("A disk error occurred while reading the file.\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		case AVIERR_FILEOPEN:
			Exception("A disk error occurred while opening the file.\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		case REGDB_E_CLASSNOTREG:
			Exception("According to the registry, the type of file specified in AVIFileOpen does not have a handler to process it.\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
	}
	
	//Open audio stream
	AVIFileGetStream(pAviFile, &pAudioStream, streamtypeAUDIO, 0);
	AVIFileRelease(pAviFile);

	if(pAudioStream != NULL)
	{
		//get audio stream information
		AVIStreamFormatSize(pAudioStream, 0, &lStreamSize);
		AVIStreamInfo(pAudioStream,&audioStreamInfo,sizeof(AVISTREAMINFO));
		AVIStreamReadFormat(pAudioStream, 0, &audioFormatInfo, &lStreamSize);
		m_nAudioFrameNum = audioStreamInfo.dwLength;
		m_nAudioSampleSize = audioStreamInfo.dwSampleSize; // WAVEFORMAT.nBlockAlign
		m_nAudioSampleRate = audioStreamInfo.dwRate / audioStreamInfo.dwScale;
		int nAudioChanles = audioFormatInfo.nChannels;	
		
		//read audio data
		long audioDataLength = m_nAudioFrameNum * m_nAudioSampleSize;
		BYTE *audioTempData = new BYTE[audioDataLength];
		if(audioTempData == NULL){
			Exception("Can't allocate audio source buffer!\nNot enough memory!\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		}
		
		m_pAudioData = new BYTE[audioDataLength/nAudioChanles];
		if(m_pAudioData == NULL){
			Exception("Can't allocate audio data buffer!\nNot enough memory!\nIn <CvAviIo::ReadWavFromAvi()>.");
			return false;
		}
		long tempByte,tempSample;
		AVIStreamRead(pAudioStream, 0, m_nAudioFrameNum, audioTempData, audioDataLength, &tempByte, &tempSample);

		int i,j;
		int nSigleChanleSampleSize;
		switch(nMode) {
			case Audio_LChannel:
				nSigleChanleSampleSize = m_nAudioSampleSize / nAudioChanles;
				for (i=0; i<m_nAudioFrameNum; i++) {
					int index1 = i * nSigleChanleSampleSize;
					int index2 = i * m_nAudioSampleSize;
					for(j=0; j<nSigleChanleSampleSize; j++)
						m_pAudioData[index1+j] = audioTempData[index2+j];
				}
				m_nAudioSampleSize = nSigleChanleSampleSize;
				break;
			case Audio_RChannel:
				nSigleChanleSampleSize = m_nAudioSampleSize / nAudioChanles;
				for (i=0; i<m_nAudioFrameNum; i++) {
					int index1 = i * nSigleChanleSampleSize;
					int index2 = i * m_nAudioSampleSize;
					for(j=0; j<nSigleChanleSampleSize; j++)
					m_pAudioData[index1+j] = audioTempData[index2+nSigleChanleSampleSize+j];	
				}
				m_nAudioSampleSize = nSigleChanleSampleSize;
				break;
		}
		delete audioTempData;		
		
		AVIStreamRelease(pAudioStream);
		AVIFileExit();
		
		m_isAudioLoad = true;
		
		return true;

	}else{
		
		AVIFileExit();

		m_isAudioLoad = false;

		return false;
	}
}

