/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include <stdio.h>
#include <stdlib.h>  
#include <string.h> 
#include "cvAvsrFeatureIO.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: ReadFeatureFileMfcc(char szFilename[], AvsrMfccHeader &header)
//    Purpose:
//			Load visual feature data from mfcc file.
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to read
//			header		---- [O] mfcc header
//    Returns: 
//		  return the data buffer
//F*/
float *CAvsrFeatureDataIO::ReadFeatureFileMfcc(char szFilename[], AvsrMfccHeader &header)
{
	if (szFilename == NULL || szFilename[0] == 0) {
		Exception("Invalid input file name!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		return NULL;
	}

	// open file
	FILE *fin=fopen(szFilename,"rb");
	if (fin == NULL) {
		Exception("Can't open visual feature data file!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		return NULL;
	}

	// read header
	if (fread(&header, sizeof(AvsrMfccHeader), 1, fin) != 1) {
		Exception("Can't read header of the Mfcc file!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		fclose(fin);
		return NULL;
	}
	if (header.sampleSize < 1 || header.nSamples < 1 || header.samplePeriod < 1) {
		Exception("Invalid Mfcc file header or null Mfcc file!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		fclose(fin);
		return NULL;
	}

	// allocate data buffer
	int nVectorLen = header.sampleSize / sizeof(float);
	unsigned int nDataNum = nVectorLen*header.nSamples;
	float *inputData = new float[nDataNum];
	if (inputData == NULL) {
		Exception("Can't allocate data buffer!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		fclose(fin);
		return NULL;
	}

	// read data
	if (fread(inputData, sizeof(float), nDataNum, fin) != nDataNum) {
		Exception("Read data failed!\nIn <CAvsrFeatureDataIO::ReadFeatureFileMfcc()>.");
		fclose(fin);
		return NULL;
	}
 
	// close file
	fclose(fin);

	return inputData;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			bool WriteFeatureFileMfcc(char szFilename[], float pData[], AvsrMfccHeader header)
//			bool WriteFeatureFileMfcc(char szFilename[], double pData[], AvsrMfccHeader header)
//    Purpose:
//			Load visual feature data from mfcc file.
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to read
//			pData		---- [I] feature data buffer
//			header		---- [I] mfcc header
//    Returns: 
//			return the data buffer
//	  Note:
//			No matter what kind of data it is, the input data are written as float type.
//F*/
bool CAvsrFeatureDataIO::WriteFeatureFileMfcc(char szFilename[], float pData[], AvsrMfccHeader header)
{
	if (szFilename == NULL || szFilename[0] == 0) {
		Exception("Invalid input file name!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}
	if (pData == NULL || header.sampleSize < 1 || header.nSamples < 1 || header.samplePeriod < 1) {
		Exception("Invalid input Mfcc header or data!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}

	// open file
	FILE *fout=fopen(szFilename, "wb");
	if (fout == NULL) {
		Exception("Can't write visual feature data file!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}

	// write header
	if (fwrite(&header, sizeof(AvsrMfccHeader), 1, fout) != 1) {
		Exception("Can't write visual feature header!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		fclose(fout);
		return false;
	}
	// write data
	int nVectorLen = header.sampleSize / sizeof(float);
	unsigned int nDataNum = nVectorLen*header.nSamples;
	if (fwrite(pData, sizeof(float), nDataNum, fout) != nDataNum) {
		Exception("Can't write visual feature data file!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		fclose(fout);
		return false;
	}

	// close file
	fclose(fout);

	return true;
}

bool CAvsrFeatureDataIO::WriteFeatureFileMfcc(char szFilename[], double pData[], AvsrMfccHeader header)
{
	if (szFilename == NULL || szFilename[0] == 0) {
		Exception("Invalid input file name!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}
	if (pData == NULL || header.sampleSize < 1 || header.nSamples < 1 || header.samplePeriod < 1) {
		Exception("Invalid input Mfcc header or data!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}

	// open file
	FILE *fout=fopen(szFilename, "wb");
	if (fout == NULL) {
		Exception("Can't write visual feature data file!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		return false;
	}

	// write header
	if (fwrite(&header, sizeof(AvsrMfccHeader), 1, fout) != 1) {
		Exception("Can't write visual feature header!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
		fclose(fout);
		return false;
	}
	// write data
	int nVectorLen = header.sampleSize / sizeof(float);
	int nDataNum = nVectorLen * header.nSamples;
	for (int i=0; i<nDataNum; i++) {
		float fTemp = (float)pData[i];
		if (fwrite(&fTemp, sizeof(float), 1, fout) != 1) {
			Exception("Can't write visual feature data file!\nIn <CAvsrFeatureDataIO::WriteFeatureFileMfcc()>.");
			fclose(fout);
			return false;
		}
	}

	// close file
	fclose(fout);

	return true;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			bool LoadEigenMatrix(const char* szFilename, CvMat** avgMat, CvMat** eValue, CvMat** eVector, bool bBinary)
//    Purpose:
//			Load eigenvalues, eigenvectors for PCA projection
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to read
//			avgMat		---- [I] mean vector
//			eValue		---- [I] eigenvalues
//			eVector		---- [I] eigenvectors
//			bBinary		---- [I] ture if the data file is binary one
//    Returns: 
//			return true if sucess
//	  Note:
//F*/
bool CAvsrFeatureDataIO::LoadEigenMatrix(const char* szFilename, CvMat** avgMat, CvMat** eValue, CvMat** eVector, bool bBinary /*= true*/)
{
	if( szFilename == NULL || avgMat == NULL || eValue == NULL || eVector == NULL ) return false;

	int i = 0, j; 
	int nVecSize, nVecNum, type;
	int nWBytes = 0;
	double* data = NULL;
	FILE* fp = NULL;

	if (bBinary) {
		fp = fopen(szFilename, "rb");
		if(fp == NULL) return false;
		if (fread(&nVecSize, sizeof(int), 1, fp) != 1) {
			fclose(fp);
			return false;
		}
		if (fread(&nVecNum, sizeof(int), 1, fp) != 1) {
			fclose(fp);
			return false;
		}
		if (nVecSize < 1 || nVecNum < 1 || nVecNum > nVecSize) {
			Exception("Error: Invalid matrix size!\nIn <CAvsrFeatureDataIO::LoadEigenMatrix()>.");
			fclose(fp);
			return false;
		}
		if (fread(&type, sizeof(int), 1, fp) != 1) {
			fclose(fp);
			return false;
		}

		// create matrices
		(*avgMat) = cvCreateMat(nVecSize, 1, type);
		(*eValue) = cvCreateMat(nVecNum, 1, type);
		(*eVector) = cvCreateMat(nVecNum, nVecSize, type);

		// load data
		if (type == CV_32F) {
			if (fread((*avgMat)->data.fl, sizeof(float), nVecSize, fp) != (unsigned int)nVecSize) {
				fclose(fp);
				return false;
			}
			if (fread((*eValue)->data.fl, sizeof(float), nVecNum, fp) != (unsigned int)nVecSize) {
				fclose(fp);
				return false;
			}
			if (fread((*eVector)->data.fl, sizeof(float), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) {
				fclose(fp);
				return false;
			}
		}
		else if (type == CV_64F) {
			if (fread((*avgMat)->data.db, sizeof(double), nVecSize, fp) != (unsigned int)nVecSize) {
				fclose(fp);
				return false;
			}
			if (fread((*eValue)->data.db, sizeof(double), nVecNum, fp) != (unsigned int)nVecSize) {
				fclose(fp);
				return false;
			}
			if (fread((*eVector)->data.db, sizeof(double), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) {
				fclose(fp);
				return false;
			}
		}
		else {
			Exception("Error: Invalid data type!\nIn <CAvsrFeatureDataIO::LoadEigenMatrix()>.");
		}

		fclose(fp);

	}else{
		fp = fopen(szFilename, "rt");
		if(fp == NULL) return false;
		fscanf(fp, "Vector Size = %d\n", &nVecSize);
		fscanf(fp, "Vector Num = %d\n", &nVecNum);
		fscanf(fp, "Data Type = %d\n", & type);
		if (nVecSize < 1 || nVecNum < 1 || nVecNum > nVecSize) {
			Exception("Error: Invalid matrix size!\nIn <CAvsrFeatureDataIO::LoadEigenMatrix()>.");
			fclose(fp);
			return false;
		}

		// create matrices
		(*avgMat) = cvCreateMat(nVecSize, 1, type);
		(*eValue) = cvCreateMat(nVecNum, 1, type);
		(*eVector) = cvCreateMat(nVecNum, nVecSize, type);
		char szTemp[256];

		// load data
		if (type == CV_32F) {
			fscanf(fp, "%s\n", szTemp);		// Mean Vector;
			for(i=0; i<nVecSize; i++) {
				fscanf(fp, "%f", &((*avgMat)->data.fl[i]));
			}
			fscanf(fp, "%s\n", szTemp);		// Eigenvalue
			for(i=0; i<nVecNum; i++) {
				fscanf(fp, "%f", &((*eValue)->data.fl[i]));
			}
			fscanf(fp, "%s\n", szTemp);		// Eigenvector
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fscanf(fp, "%f", &((*eVector)->data.fl[i*nVecSize+j]));
				}
			}
		}
		else if (type == CV_64F) {
			fscanf(fp, "%s %s\n", szTemp, szTemp);		// Mean Vector;
			for(i=0; i<nVecSize; i++) {
				fscanf(fp, "%lf", &((*avgMat)->data.db[i]));
			}
			fscanf(fp, "%s\n", szTemp);		// Eigenvalue
			for(i=0; i<nVecNum; i++) {
				fscanf(fp, "%lf", &((*eValue)->data.db[i]));
			}
			fscanf(fp, "%s\n", szTemp);		// Eigenvector
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fscanf(fp, "%lf", &((*eVector)->data.db[i*nVecSize+j]));
				}
			}
		}
		else {
			Exception("Error: Invalid data type!\nIn <CAvsrFeatureDataIO::LoadEigenMatrix()>.");
		}

		fclose(fp);
	}
	return true;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			bool SaveEigenMatrix(const char* szFileName, CvMat* avgMat, CvMat* eValue, CvMat* eVector, bool bBinary)
//    Purpose:
//			write eigenvalues, eigenvectors obtained from PCA
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to write
//			avgMat		---- [I] mean vector
//			eValue		---- [I] eigenvalues
//			eVector		---- [I] eigenvectors
//			bBinary		---- [I] ture if the data file write as a binary one
//    Returns: 
//			return ture if success
//	  Note:
//F*/
bool CAvsrFeatureDataIO::SaveEigenMatrix(const char* szFileName, CvMat* avgMat, CvMat* eValue, CvMat* eVector, bool bBinary /*= true*/ )
{
	if( (szFileName == NULL)) return false;

	int i = 0, j; 
	int nWBytes = 0;
	double* data = NULL;
	FILE* fp = NULL;

	// check data type here
	int type = cvGetElemType(eVector);
	// here
	int nVecSize = eVector->cols;
	int nVecNum = eVector->rows;
	if (eValue->cols * eValue->rows != nVecNum || avgMat->rows*avgMat->cols != nVecSize) {
		printf("Error: Dimensions of eigen vector, eigen value and mean vector do not match!\n");
		return false;
	}

	if (bBinary) {
		fp = fopen(szFileName, "wb");
		if(fp == NULL) return false;
		
		if ( fwrite(&nVecSize, sizeof(int), 1, fp) != 1) return false;

		if ( fwrite(&nVecNum, sizeof(int), 1, fp) != 1) return false;

		if ( fwrite(&type, sizeof(int), 1, fp) != 1) return false;
	
		if (type == CV_32F) {
			if ( fwrite(avgMat->data.fl, sizeof(float), nVecSize, fp) != (unsigned int)nVecSize) return false;
			
			if ( fwrite(eValue->data.fl, sizeof(float), nVecNum, fp) != (unsigned int)nVecSize) return false;

			if ( fwrite(eVector->data.fl, sizeof(float), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) return false;
		}
		else if (type == CV_64F) {
			if ( fwrite(avgMat->data.db, sizeof(double), nVecSize, fp) != (unsigned int)nVecSize) return false;
			
			if ( fwrite(eValue->data.db, sizeof(double), nVecNum, fp) != (unsigned int)nVecSize) return false;

			if ( fwrite(eVector->data.db, sizeof(double), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) return false;
		}
		else {
			Exception("Error: Invalid data type!\nIn <CAvsrFeatureDataIO::SaveEigenMatrix()>.");
		}

		fclose(fp);

	}else{
		fp = fopen(szFileName, "wt");
		if(fp == NULL)
			return false;
		
		fprintf(fp, "Vector Size = %d\nVector Num = %d\nData Type = %d\n", nVecSize, nVecNum, type);
		if (type == CV_32F) {
			fprintf(fp, "Mean Vector\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%f\t", avgMat->data.fl[i]);
			}
			fprintf(fp, "\nEigenvalue\n");
			for(i=0; i<nVecNum; i++) {
				fprintf(fp, "%f\t", eValue->data.fl[i]);
			}
			fprintf(fp, "\nEigenvector\n");
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fprintf(fp, "%f\t", eVector->data.fl[i*nVecSize+j]);
				}
				fprintf(fp, "\n");
			}
		}
		else if (type == CV_64F) {
			fprintf(fp, "Mean Vector\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%1.10lf\t", avgMat->data.db[i]);
			}
			fprintf(fp, "\nEigenvalue\n");
			for(i=0; i<nVecNum; i++) {
				fprintf(fp, "%1.10lf\t", eValue->data.db[i]);
			}
			fprintf(fp, "\nEigenvector\n");
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fprintf(fp, "%1.10lf\t", eVector->data.db[i*nVecSize+j]);
				}
				fprintf(fp, "\n");
			}
		}
		else {
			Exception("Error: Invalid data type!\nIn <CAvsrFeatureDataIO::SaveEigenMatrix()>.");
		}

		fclose(fp);
	}
	return true;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//		bool LoadLdaMatrix(const char* szFileName, CvMat** meanMat, CvMat** devsMat,
//						   CvMat** evMat, CvMat** transMat, bool bBinary)
//    Purpose:
//			Load LDA transform matrix and other data for LDA projection
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to write
//			meanMat		---- [I] mean vector
//			devsMat		---- [I] mean square error vector
//			evMat		---- [I] eigenvalues
//			transMat	---- [I] LDA transformation matrix
//			bBinary		---- [I] ture if the data file is a binary one
//    Returns: 
//			return ture if success
//	  Note:
//F*/
bool CAvsrFeatureDataIO::LoadLdaMatrix(const char* szFileName,
									   CvMat** meanMat,
									   CvMat** devsMat,
									   CvMat** evMat,
									   CvMat** transMat,
									   bool bBinary)
{
	int i,j;
	int type, nVecSize, nVecNum, nRBytes = 0;
	float* fdata = NULL;
	double* ddata = NULL;
	FILE* fp = NULL;
	bool isMatNULL = false;

	if(szFileName == NULL || meanMat==NULL && devsMat==NULL && evMat==NULL && transMat==NULL)
		return false;
	
	if(bBinary){	// read binary lda matrix file
		fp = fopen(szFileName,"rb");
		if(fp==NULL)
			return false;
		
		if (fread(&nVecSize,sizeof(int),1,fp)  != 1) return false;

		if (fread(&nVecNum,sizeof(int),1,fp) != 1) return false;

		if (fread(&type,sizeof(int),1,fp) != 1) return false;
	
		// allocate matrix data
		(*meanMat) = cvCreateMat(nVecSize, 1, type);
		if(*meanMat==NULL) return false;
		(*devsMat) = cvCreateMat(nVecSize, 1, type);
		if(devsMat==NULL) return false;
		(*evMat) = cvCreateMat(nVecNum, 1, type);
		if(*evMat==NULL) return false;
		(*transMat) = cvCreateMat(nVecNum, nVecSize, type);
		if(*transMat==NULL) return false;

		// read data from file into matrix
		if(type==CV_32F){
			// mean value of sample feature
			fdata=(*meanMat)->data.fl;
			if (fread(fdata,sizeof(float),nVecSize,fp) != (unsigned int)nVecSize)	return false;

			// mean squire error of sample feature
			fdata=(*devsMat)->data.fl;
			if (fread(fdata,sizeof(float),nVecSize,fp) != (unsigned int)nVecSize)	return false;
			
			// eigen values corresponding to transform matrix
			fdata=(*evMat)->data.fl;
			if (fread(fdata,sizeof(float),nVecNum,fp) != (unsigned int)nVecNum) return false;

			// transform matrix
			fdata=(*transMat)->data.fl;
			if (fread(fdata,sizeof(float),nVecSize*nVecNum,fp) != (unsigned int)(nVecSize*nVecNum))	return false;
		}
		else if(type==CV_64F){
			// mean value of sample feature
			ddata=(*meanMat)->data.db;
			if (fread(ddata,sizeof(double),nVecSize,fp) != (unsigned int)nVecSize) return false;

			// mean squire error of sample feature
			ddata=(*devsMat)->data.db;
			if (fread(ddata,sizeof(double),nVecSize,fp) != (unsigned int)nVecSize) return false;
			
			// eigen values corresponding to transform matrix
			ddata=(*evMat)->data.db;
			if (fread(ddata,sizeof(double),nVecNum,fp) != (unsigned int)nVecNum) return false;

			// transform matrix
			ddata=(*transMat)->data.db;
			if (fread(ddata,sizeof(double),nVecSize*nVecNum,fp) != (unsigned int)(nVecSize*nVecNum)) return false;
		}else{	// other type not supportted
			return false;
		}
		fclose(fp);
	} else {	// read text lda matrix file
		fp = fopen(szFileName,"rt");
		if(fp==NULL)return false;
		
		fscanf(fp,"Vector Size = %d\n",&nVecSize);
		fscanf(fp,"Vector Num = %d\n",&nVecNum);
		fscanf(fp,"Data Type = %d\n",&type);
		
		// allocate matrix data
		(*meanMat) = cvCreateMat(nVecSize, 1, type);
		if(*meanMat==NULL) return false;
		(*devsMat) = cvCreateMat(nVecSize, 1, type);
		if(devsMat==NULL) return false;
		(*evMat) = cvCreateMat(nVecNum, 1, type);
		if(*evMat==NULL) return false;
		(*transMat) = cvCreateMat(nVecNum, nVecSize, type);
		if(*transMat==NULL) return false;
		
		// read data from file into matrix
		if(type==CV_32F){
			float temp;
			// mean value of sample feature
			fdata = (*meanMat)->data.fl;
			fscanf(fp,"Mean Vector\n");			
			for(i=0;i<nVecSize;i++){
				fscanf(fp,"%f\t",&temp);
				*(fdata+i)=temp;
			}
			// mean squire error of sample feature
			fdata = (*devsMat)->data.fl;
			fscanf(fp,"\nMSE\n");
			for(i=0;i<nVecSize;i++){
				fscanf(fp,"%f\t",&temp);
				*(fdata+i)=temp;
			}
			// eigen values corresponding to transform matrix
			fdata = (*evMat)->data.fl;
			fscanf(fp,"\nEigenvalue\n");
			for(i=0;i<nVecNum;i++){
				fscanf(fp,"%f\t",&temp);
				*(fdata+i)=temp;
			}
			// transform matrix
			fdata = (*transMat)->data.fl;
			fscanf(fp,"\nTransform Matrix\n");
			for(i=0;i<nVecNum;i++){
				for(j=0;j<nVecSize;j++){
					fscanf(fp,"%f\t",&temp);
					*(fdata+i*nVecSize+j)=temp;
				}
				fscanf(fp,"\n");
			}
		}
		else if(type==CV_64F){	// type==CV_64F
			double temp;
			// mean value of sample feature
			ddata = (*meanMat)->data.db;
			fscanf(fp,"Mean Vector\n");
			for(i=0;i<nVecSize;i++){
				fscanf(fp,"%lf\t",&temp);
				*(ddata+i)=temp;
			}
			// mean squire error of sample feature
			ddata = (*devsMat)->data.db;
			fscanf(fp,"\nMSE\n");
			for(i=0;i<nVecSize;i++){
				fscanf(fp,"%lf\t",&temp);
				*(ddata+i)=temp;
			}
			// eigen values corresponding to transform matrix
			ddata = (*evMat)->data.db;
			fscanf(fp,"\nEigenvalue\n");
			for(i=0;i<nVecNum;i++){
				fscanf(fp,"%lf\t",&temp);
				*(ddata+i)=temp;
			}
			// transform matrix
			ddata = (*transMat)->data.db;
			fscanf(fp,"\nTransform Matrix\n");
			for(i=0;i<nVecNum;i++){
				for(j=0;j<nVecSize;j++){
					fscanf(fp,"%lf\t",&temp);
					*(ddata+i*nVecSize+j)=temp;
				}
				fscanf(fp,"\n");
			}
		}else{	// other type not supportted
			return false;
		}
		fclose(fp);
	}

	return true;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//		bool SaveLdaMatrix(const char* szFileName, CvMat* meanMat, CvMat* devsMat,
//						   CvMat* evMat, CvMat* transMat, bool bBinary)
//    Purpose:
//			Load LDA transform matrix and other data for LDA projection
//    Context:
//    Parameters:
//			szFilename	---- [I] name of the file to write
//			meanMat		---- [I] mean vector
//			devsMat		---- [I] mean square error vector
//			evMat		---- [I] eigenvalues
//			transMat	---- [I] LDA transformation matrix
//			bBinary		---- [I] ture if the data file is written as a binary one
//    Returns: 
//			return ture if success
//	  Note:
//F*/
bool CAvsrFeatureDataIO::SaveLdaMatrix( const char* szFileName, CvMat* meanMat, CvMat* devsMat, 
										CvMat* evMat, CvMat* transMat, bool bBinary)
{
	if( (szFileName == NULL)) return false;

	int i = 0, j; 
	int nWBytes = 0;
	double* data = NULL;
	FILE* fp = NULL;

	// check data type here
	int type = cvGetElemType(transMat);
	// here
	int nVecSize = transMat->cols;
	int nVecNum = transMat->rows;
	if (meanMat->rows != nVecSize || devsMat->rows != nVecSize || evMat->rows != nVecNum) {
		printf("Error: Dimensions of mean, devation, eigen value and transform matrices do not match!\n");
		return false;
	}

	if (bBinary) {
		fp = fopen(szFileName, "wb");
		if(fp == NULL) return false;
		
		if (fwrite(&nVecSize, sizeof(int), 1, fp) != 1) return false;

		if (fwrite(&nVecNum, sizeof(int), 1, fp) != 1) return false;

		if (fwrite(&type, sizeof(int), 1, fp) != 1) return false;
	
		if (type == CV_32F) {
			// mean value of sample feature
			if (fwrite(meanMat->data.fl, sizeof(float), nVecSize, fp) != (unsigned int)nVecSize) return false;

			// mean squire  error of sample feature
			if ( fwrite(devsMat->data.fl, sizeof(float), nVecSize, fp) != (unsigned int)nVecSize) return false;
			
			// eigen values corresponding to Transform matrix
			if ( fwrite(evMat->data.fl, sizeof(float), nVecNum, fp) != (unsigned int)nVecNum) return false;

			// transform matrix
			if ( fwrite(transMat->data.fl, sizeof(float), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) return false;
		}
		else {
			// mean value of sample feature
			if ( fwrite(meanMat->data.db, sizeof(double), nVecSize, fp) != (unsigned int)nVecSize) return false;

			// mean squire  error of sample feature
			if ( fwrite(devsMat->data.db, sizeof(double), nVecSize, fp) != (unsigned int)nVecSize) return false;
			
			// eigen values corresponding to Transform matrix
			if ( fwrite(evMat->data.db, sizeof(double), nVecNum, fp) != (unsigned int)nVecNum) return false;

			// transform matrix
			if ( fwrite(transMat->data.db, sizeof(double), nVecSize*nVecNum, fp) != (unsigned int)(nVecSize*nVecNum)) return false;
		}

		fclose(fp);

	}else{
		fp = fopen(szFileName, "wt");
		if(fp == NULL)
			return false;
		
		fprintf(fp, "Vector Size = %d\nVector Num = %d\nData Type = %d\n", nVecSize, nVecNum, type);
		if (type == CV_32F) {
			// mean value of sample feature
			fprintf(fp, "Mean Vector\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%e\t", meanMat->data.fl[i]);
			}
			// mean squire  error of sample feature
			fprintf(fp, "\nMSE\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%e\t", devsMat->data.fl[i]);
			}
			// eigen values corresponding to Transform matrix
			fprintf(fp, "\nEigenvalue\n");
			for(i=0; i<nVecNum; i++) {
				fprintf(fp, "%e\t", evMat->data.fl[i]);
			}
			// transform matrix
			fprintf(fp, "\nTransform Matrix\n");
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fprintf(fp, "%e\t", transMat->data.fl[i*nVecSize+j]);
				}
				fprintf(fp, "\n");
			}
		}
		else {		// type == CV64F
			// mean value of sample feature
			fprintf(fp, "Mean Vector\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%e\t", meanMat->data.db[i]);
			}
			// mean squire  error of sample feature
			fprintf(fp, "\nMSE\n");
			for(i=0; i<nVecSize; i++) {
				fprintf(fp, "%e\t", devsMat->data.db[i]);
			}
			// eigen values corresponding to Transform matrix
			fprintf(fp, "\nEigenvalue\n");
			for(i=0; i<nVecNum; i++) {
				fprintf(fp, "%e\t", evMat->data.db[i]);
			}
			// transform matrix
			fprintf(fp, "\nTransform Matrix\n");
			for(i=0; i<nVecNum; i++) {
				for(j=0; j<nVecSize; j++) {
					fprintf(fp, "%e\t", transMat->data.db[i*nVecSize+j]);
				}
				fprintf(fp, "\n");
			}
		}

		fclose(fp);
	}
	return true;
}

// Replace a file's extension
bool CAvsrFeatureDataIO::ReplaceFileExt(char szFileOld[], char szNewExt[], char szFileNew[])
{
	if (szFileOld == NULL || szNewExt == NULL || szFileNew == NULL) return false;

	char *pDst;
	pDst = strrchr(szFileOld, '.');
	if (pDst == NULL) return false;
	int nLenTitle = strlen(szFileOld) - strlen(pDst);

	int nLen = strlen(szNewExt);
	strncpy(szFileNew, szFileOld, nLenTitle);			// copy title with dot
	for (int i=0; i<nLen; i++) {
		szFileNew[nLenTitle+i] = szNewExt[i];
	}
	szFileNew[nLenTitle+nLen] = 0;

	return true;
}

void CAvsrFeatureDataIO::Exception(char szMsg[])
{
	//AfxMessageBox(szMsg);
}

