/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include <stdio.h>
#include <io.h>
#include <direct.h>

#include "cv.h"
#include "CvAvsrFeatureIO.h"
#include "CvAviIO.h"
#include "cvvisualspeech.h"
#include "cvmouthtracking.h"

void PringUsage() 
{
	printf("PCA training:\tpca.exe -train <AVI file folder> <result file>  <subspace dim>\n");
	printf("PCA projection:\tpca.exe -proj <PCA data file> <AVI file folder> <result file folder>\n");
	return;
}

bool PcaTrain(char *szAviFolder, char *szResultFile, int nEigenNum); 

bool PcaProjection(char *szPcaFile, char *szAviFolder, char *szFeatFolder);

IplImage ** CreateImageList(int nWidth, int nHeight, int nBand, int nFrameNum);

void ClearImageList(IplImage **pImageList, int nFrameNum);

const int g_nStdWidth = 32;
const int g_nStdHeight = 32;

// for debug
IplImage *AverageDown(IplImage *iplSrcImage);

int main(int argc, char* argv[])
{
	if (argc != 5) {
		printf("Error: Invalid input argument!\n");
		PringUsage(); 
		return 0;
	}

	int nEigenNum;
	if (strcmp(strlwr(argv[1]), "-train") == 0) {
		sscanf(argv[4], "%d", &nEigenNum);
		if (nEigenNum < 1 || nEigenNum > g_nStdWidth * g_nStdHeight) {
			printf("Error: Invalid number of eigen values!\n");
			return 0;
		}
		// PCA training
		PcaTrain(argv[2], argv[3], nEigenNum);
	}
	else if (strcmp(strlwr(argv[1]), "-proj") == 0) {
		// PCA projection
		PcaProjection(argv[2], argv[3], argv[4]);
	}
	else {
		printf("Error: Invalid argument <%s>!\n", argv[1]);
		PringUsage();
		return 0;
	}

	return 0;
}

bool PcaTrain(char *szAviFolder, char *szResultFile, int nEigenNum)
{
	if (szAviFolder == NULL || szResultFile == NULL) {
		printf("Error: Invalid AVI folder or result file name!\n");
		return false;
	}

	char szAviPath[512];
	long hFile = -1L;
	struct _finddata_t c_file;
	CvAviIo aviIo;
	CAvsrFeatureDataIO featIo;
	int processed = 0;	
	int nImageWidth, nImageHeight, nImageBand, nFrameNum;
	float fFrameRate;
	IplImage **pNormImageList;
	IplImage *iplGrayImage;
	int i;

	// create matrices
	int nTotalImageNum = 0, nVectorSize = g_nStdWidth * g_nStdHeight;
	const int type = CV_64F;//CV_32F;
	CvMat *sumMat, *cosqrMat, *meanMat, *covMat, *eigenValue, *eigenVector;
	sumMat = cvCreateMat(nVectorSize, 1, type);
	cosqrMat = cvCreateMat(nVectorSize, nVectorSize, type);
	meanMat = cvCreateMat(nVectorSize, 1, type);
	covMat = cvCreateMat(nVectorSize, nVectorSize, type);
	eigenValue = cvCreateMat(nEigenNum, 1, type);
	eigenVector = cvCreateMat(nEigenNum, nVectorSize, type);

	sprintf(szAviPath, "%s\\*.avi", szAviFolder);
	bool bFirst = TRUE;
	if ((hFile = _findfirst(szAviPath, &c_file )) == -1L ) 
	{
		printf("Error: Can't find AVI file in folder <%s>.", szAviFolder);
		return false;
	}
	do {
		// load AVI
		printf("Process file <%s>...\t", c_file.name);
		sprintf(szAviPath, "%s\\%s", szAviFolder, c_file.name);
		if (!aviIo.LoadAvi(szAviPath, false, true)) {
			printf("Error: Fail to load data in file <%s>!\n", c_file.name);
			continue;
		}
		// check image size
		if (!aviIo.GetVideoInfo(nImageWidth, nImageHeight, nImageBand, nFrameNum, fFrameRate)) {
			printf("Error: Invalid AVI video data!\n");
			aviIo.ClearAvi();
			continue;
		}
		// normalize images
		pNormImageList = CreateImageList(g_nStdWidth, g_nStdHeight, 1, nFrameNum);
		if (pNormImageList == NULL) {
			printf("Error: Fail to create image list!\n");
			continue;
		}
		if (nImageBand != 1) {
			iplGrayImage = cvCreateImage( cvSize(nImageWidth, nImageHeight), IPL_DEPTH_8U, 1 );
			if (iplGrayImage == NULL) {
				printf("Error: Fail to create temporary gray image!\n");
				ClearImageList(pNormImageList, nFrameNum);
				continue;
			}
			if (nImageWidth != g_nStdWidth || nImageHeight != g_nStdHeight) {
				for (i=0; i<nFrameNum; i++) {
					// get gray image
					cvCvtColor( aviIo.GetImage(i), iplGrayImage, CV_BGR2GRAY );
					// resample
					cvResize(iplGrayImage, pNormImageList[i]);
				}
			} else {
				for (i=0; i<nFrameNum; i++) {
					// get gray image
					cvCvtColor( aviIo.GetImage(i), pNormImageList[i], CV_BGR2GRAY );
				}
			}
			cvReleaseImage(&iplGrayImage);
		}
		else {
			if (nImageWidth != g_nStdWidth || nImageHeight != g_nStdHeight) {
				for (i=0; i<nFrameNum; i++) {
					// resample
					cvResize(aviIo.GetImage(i), pNormImageList[i]);
				}
			} else {
				for (i=0; i<nFrameNum; i++) {
					// copy images
					cvCopyImage(aviIo.GetImage(i), pNormImageList[i]);
				}
			}
		}
		aviIo.ClearAvi();
		for (i=0; i<nFrameNum; i++) {
			// IGC
			//cvIGC(pNormImageList[i]);
		}
		//  PCA training
		if (bFirst) {
			cvInitPCA((const IplImage **)pNormImageList, nFrameNum, sumMat, cosqrMat);
			nTotalImageNum = nFrameNum;
			bFirst = FALSE;
		} else {
			cvUpdatePCA((const IplImage **)pNormImageList, nFrameNum, nTotalImageNum, sumMat, cosqrMat);
		}
		ClearImageList(pNormImageList, nFrameNum);
		processed ++;
		printf("Ok!\n");
	} while (_findnext( hFile, &c_file ) == 0);
	printf("%d AVI file processed.\n", processed);

	if (nTotalImageNum > 1) {
		// calculate eigen vectors and values
		printf("Compute eigen vectors and eigen values...\n");
		cvCalcPCAMatrix(nTotalImageNum, sumMat, cosqrMat, meanMat, covMat, eigenValue, eigenVector);
		// save result
		printf("Save result...\n");
		featIo.SaveEigenMatrix(szResultFile, meanMat, eigenValue, eigenVector, false);
	}
	else {
		printf("Error: No image in loaded!\n");
	}

	// release matrices
	cvReleaseMat(&sumMat);
	cvReleaseMat(&cosqrMat);
	cvReleaseMat(&meanMat);
	cvReleaseMat(&covMat);
	cvReleaseMat(&eigenValue);
	cvReleaseMat(&eigenVector);

	return true;
}

bool PcaProjection(char *szPcaFile, char *szAviFolder, char *szFeatFolder)
{
	if (szPcaFile == NULL || szAviFolder == NULL || szFeatFolder == NULL) {
		printf("Error: Invalid PCA file, AVI folder or result file name!\n");
		return false;
	}

	char szAviPath[512];
	char szFeatPath[512];
	char szFeatFile[256];
	long hFile = -1L;
	struct _finddata_t c_file;
	CvAviIo aviIo;
	CAvsrFeatureDataIO featIo;
	int processed = 0;	
	int nImageWidth, nImageHeight, nImageBand, nFrameNum;
	float fFrameRate;
	IplImage **pNormImageList;
	IplImage *iplGrayImage;
	int i;
	AvsrMfccHeader mfccHeader;

	// load PCA file
	CvMat *meanMat, *eigenValue, *eigenVector, *featMat;
	if (!featIo.LoadEigenMatrix(szPcaFile, &meanMat, &eigenValue, &eigenVector, false)) {
		printf("Error: Fail to load PCA file <%s>!\n", szPcaFile);
		return false;
	}
	int nVectorSize = eigenVector->cols;
	int nOutVecSize = eigenVector->rows;
	if (nVectorSize != g_nStdWidth * g_nStdHeight) {
		printf("Error: Eigen vectors do not match the image size!\n");
		return false;
	}

	// create destination folder
	_mkdir(szFeatFolder);

	// projection
	sprintf(szAviPath, "%s\\*.avi", szAviFolder);
	bool bFirst = TRUE;
	if ((hFile = _findfirst(szAviPath, &c_file )) == -1L ) 
	{
		printf("Error: Can't find AVI file in folder <%s>.", szAviFolder);
		return false;
	}
	do {
		// load AVI
		printf("Process file <%s>...\t", c_file.name);
		sprintf(szAviPath, "%s\\%s", szAviFolder, c_file.name);
		if (!aviIo.LoadAvi(szAviPath, false, true)) {
			printf("Error: Fail to load data in file <%s>!\n", c_file.name);
			continue;
		}
		// check image size
		if (!aviIo.GetVideoInfo(nImageWidth, nImageHeight, nImageBand, nFrameNum, fFrameRate)) {
			printf("Error: Invalid AVI video data!\n");
			aviIo.ClearAvi();
			continue;
		}
		if (fFrameRate <= 0.0f) {
			printf("Error: Invalid frame rate!\n");
			continue;
		}
		// normalize images
		pNormImageList = CreateImageList(g_nStdWidth, g_nStdHeight, 1, nFrameNum);
		if (pNormImageList == NULL) {
			printf("Error: Fail to create image list!\n");
			continue;
		}
		if (nImageBand != 1) {
			iplGrayImage = cvCreateImage( cvSize(nImageWidth, nImageHeight), IPL_DEPTH_8U, 1 );
			if (iplGrayImage == NULL) {
				printf("Error: Fail to create temporary gray image!\n");
				ClearImageList(pNormImageList, nFrameNum);
				continue;
			}
			if (nImageWidth != g_nStdWidth || nImageHeight != g_nStdHeight) {
				for (i=0; i<nFrameNum; i++) {
					// get gray image
					cvCvtColor( aviIo.GetImage(i), iplGrayImage, CV_BGR2GRAY );
					// resample
					cvResize(iplGrayImage, pNormImageList[i]);
				}
			} else {
				for (i=0; i<nFrameNum; i++) {
					// get gray image
					cvCvtColor( aviIo.GetImage(i), pNormImageList[i], CV_BGR2GRAY );
				}
			}
			cvReleaseImage(&iplGrayImage);
		}
		else {
			if (nImageWidth != g_nStdWidth || nImageHeight != g_nStdHeight) {
				for (i=0; i<nFrameNum; i++) {
					// resample
					cvResize(aviIo.GetImage(i), pNormImageList[i]);
				}
			} else {
				for (i=0; i<nFrameNum; i++) {
					// copy images
					cvCopyImage(aviIo.GetImage(i), pNormImageList[i]);
				}
			}
		}
		aviIo.ClearAvi();
		for (i=0; i<nFrameNum; i++) {
			// IGC
			//cvIGC(pNormImageList[i]);
		}
		//  PCA projection
		featMat = cvCreateMat(nFrameNum, nOutVecSize, CV_32F);
		cvPCAProjection((const IplImage **)pNormImageList, nFrameNum, eigenValue, eigenVector, 
						meanMat, featMat, false);
		ClearImageList(pNormImageList, nFrameNum);
		// save result
		if (!featIo.ReplaceFileExt(c_file.name, ".pmfc", szFeatFile)) {
			printf("Error: Fail to create feature file name!\n");
			cvReleaseMat(&featMat);
			continue;
		}
		sprintf(szFeatPath, "%s\\%s", szFeatFolder, szFeatFile);
		mfccHeader.nSamples = nFrameNum;
		mfccHeader.paramKind = AVSR_MFCC_VISUAL;
		mfccHeader.samplePeriod = (int)(10000000 / fFrameRate);
		mfccHeader.sampleSize = nOutVecSize * sizeof(float);
		featIo.WriteFeatureFileMfcc(szFeatPath, featMat->data.fl, mfccHeader); 
		//featIo.WriteFeatureFileText(szFeatPath, featMat->data.fl, nOutVecSize, nFrameNum); 
		cvReleaseMat(&featMat);
		processed ++;
		printf("Ok!\n");
	} while (_findnext( hFile, &c_file ) == 0);
	printf("%d AVI file processed.\n", processed);

	// release matrices
	cvReleaseMat(&meanMat);
	cvReleaseMat(&eigenValue);
	cvReleaseMat(&eigenVector);

	return true;
}

IplImage **CreateImageList(int nWidth, int nHeight, int nBand, int nFrameNum)
{
	int i, j;

	IplImage **pImageList = (IplImage **)malloc(nFrameNum * sizeof(IplImage*));
	if (pImageList == NULL) {
		printf("Error: Fail to allocate image list!\n");
		return NULL;
	}

	for (i=0; i<nFrameNum; i++) {
		pImageList[i] = cvCreateImage( cvSize(nWidth, nHeight), IPL_DEPTH_8U, nBand );
		if (pImageList[i] == NULL) {
			printf("Error: Fail to create image in list!\n");
			for (j=0; j<i; j++) {
				cvReleaseImage(&(pImageList[i]));
			}
			delete pImageList;
			return NULL;
		}
	}

	return pImageList;
}

void ClearImageList(IplImage **pImageList, int nFrameNum)
{
	if (pImageList == NULL || nFrameNum < 1) return;

	int i;
	for (i=0; i<nFrameNum; i++) {
		if (pImageList[i] != NULL) {
			cvReleaseImage(&(pImageList[i]));
		}
	}
	delete pImageList;

	return;
}

IplImage *AverageDown(IplImage *iplSrcImage)
{
	if (iplSrcImage == NULL) {
		printf("Invalid input images!\nIn <CAvsrImageProcess::AverageDown()>.");
		return NULL;
	}

	int nSrcW = iplSrcImage->width;
	int nSrcH = iplSrcImage->height;
	int nDstW = nSrcW / 2;
	int nDstH = nSrcH / 2;
	IplImage *iplImage = cvCreateImage(cvSize(nDstW, nDstH), IPL_DEPTH_8U, 1);
	if (iplImage == NULL) {
		printf("Can't allocate result image!\nIn <CAvsrImageProcess::AverageDown()>.");
		return NULL;
	}
	int nSrcLineW = iplSrcImage->widthStep;
	int nDstLineW = iplImage->widthStep;

	unsigned char *pArray = (unsigned char *)iplSrcImage->imageData;
	unsigned char *pDst = (unsigned char *)iplImage->imageData;
	for (int y=0; y<nDstH; y++) {
		for (int x=0; x<nDstW; x++) {
			int index = nSrcLineW * y * 2 + x * 2;
			int nTotal = pArray[index] + pArray[index+1] + pArray[index+nSrcLineW] + pArray[index+nSrcLineW+1];
			pDst[nDstLineW*y+x] = nTotal / 4;
		}
	}

	return iplImage;
}
