/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#ifndef _CV_MOUTH_TRACK_H_
#define _CV_MOUTH_TRACK_H_

#define CV_MOUTH_TRAJECTORY_LIST_SIZE	64

#include "cvtypes.h"
#include "_cvaux.h"

#ifdef __cplusplus
extern "C" {
#endif
	
enum CvTrackRltSts 
{ 
	TRACKSUCCESSFUL, 
	FACEDETECTERROR, 
	MOUTHDETECTERROR,
	MOUTHTRACKERROR, 
};

enum CvTrackState
{
	FACEMOUTHDETECTION,
	MOUTHTRACKING,
};

typedef struct _CvMouthLocateResult 
{
	float	CenterX;		// location of the mouth
	float	CenterY;
	float	Width;
	float	Height;
	float	RotAngle;		// rotation angle of the mouth, clockwise, in degree
	float	Likeness;		// final likeness of the region, from 0 to 1
	int		face_left;
	int		face_right;
	int		face_top;
	int		face_bottom;
} 
CvMouthLocateResult;


typedef struct _CvTrackResult
{
	CvTrackState TrackState;
	CvMouthLocateResult MouthResult;
	CvTrackRltSts  TrackRltSts;
}
CvTrackResult;

typedef struct _CvMouthTracker
{
	
	// Down sample image and its parameter
	float DownSmpRate;
	IplImage *iplDownSmpImage;
	
	// Face detector
	CvHidHaarClassifierCascade *hidCascadeFace;
	
	// Mouth detector
	CvHidHaarClassifierCascade	*hidCascadeMouth0;
	CvHidHaarClassifierCascade	*hidCascadeMouth1;
	CvMemStorage *StorageDual0;
	CvMemStorage *StorageDual1;
	int			MouthWndWidth;

	// for tracking state
	CvTrackState TrackState;	// tracking state
	int InitialCounter;			// counter for initialize stage
	int ErrorCounter;

	// for AvsrKalmanFilter
	CvKalman *KalmanXY;
	int   KinematicOrder;		// 2 -- position and velocity, 3 -- position, velocity and acceleration
	CvMat *MeasureVectXY;
	float PrevX, PrevY;
	float PrevVx, PrevVy;
	int iskalmanCreated;
	
	// TrackPostprocess for tracking
	int	RltPoolLen;
	int	RltPoolHeader;
	int	isRltPoolFull;
	CvMouthLocateResult	ResultPool[CV_MOUTH_TRAJECTORY_LIST_SIZE];
}
CvMouthTracker;

/* Image Processing Function */
OPENCVAUXAPI CVStatus
cvEqualizeHist(IplImage *img);

// illumination gradient correct
OPENCVAUXAPI CVStatus 
cvNormalizeIllum(IplImage *img, const float destMean CV_DEFAULT(128), const float destMse CV_DEFAULT(64));

/* object detection function */
OPENCVAUXAPI CvSeq*
cvHaarDetectObjectsR( const IplImage* img,
                      CvHidHaarClassifierCascade* cascade,
                      CvMemStorage* storage, double scale_factor CV_DEFAULT(1.1),
                      int min_neighbors CV_DEFAULT(3), int flags CV_DEFAULT(0),
					  int min_width CV_DEFAULT(24), int max_width CV_DEFAULT(256));

/* Mouth Tracking Function */
OPENCVAUXAPI CVStatus
cvInitTrackMouth( CvHidHaarClassifierCascade *hidCascadeFace, 
				  CvHidHaarClassifierCascade *hidCascadeMouth0, 
				  CvHidHaarClassifierCascade *hidCascadeMouth1,
				  CvMouthTracker *MouthTracker);

OPENCVAUXAPI CVStatus 
cvResetTrackMouth(const int imagewidth, const int imageheight, CvMouthTracker* MouthTracker);

OPENCVAUXAPI CVStatus 
cvUpdateTrackMouth(IplImage * img, CvMouthTracker* MouthTracker, CvTrackResult & result); 

OPENCVAUXAPI CVStatus 
cvPostProcessTrackMouth( CvMouthLocateResult * TrackList, const int framenum ); 
	
OPENCVAUXAPI CVStatus 
cvReleaseTrackMouth(CvMouthTracker* MouthTracker);

/* icv Function*/
/* for face detector */
OPENCVAUXAPI int
icvHaarDetectMaxFace(const IplImage *iplImage, CvHidHaarClassifierCascade* hidCascade, CvRect &face);

OPENCVAUXAPI CVStatus
icvHaarEstimateMouthRegion(const CvRect face, CvRect &rectSearch, int &MouthWidth);

/* for mouth detector Haar*/
OPENCVAUXAPI CVStatus 
icvHaarDetectMouthDualCascades(IplImage *iplImage, CvMouthTracker* MouthTracker, double scale_factor, int min_width, int max_width, CvRect &mouth, bool &isMouthDetected);

/* for mouth tracking */
OPENCVAUXAPI CVStatus
icvResultPoolInit(const int PoolSize, CvMouthTracker* MouthTracker);

OPENCVAUXAPI void
icvResultPoolPush(const CvMouthLocateResult &result, CvMouthTracker* MouthTracker);

OPENCVAUXAPI CVStatus
icvResultPoolAverage(const CvMouthTracker* MouthTracker, CvMouthLocateResult &result);

OPENCVAUXAPI CVStatus
icvResultPoolClear(CvMouthTracker* MouthTracker);

/* for Kalman filter*/
OPENCVAUXAPI CVStatus
icvCreateKalmanXY(const int KinOrder, const float Tx, const float Ty, const float Q, const float R, CvMouthTracker* MouthTracker);

OPENCVAUXAPI CVStatus
icvInitKalmanXY( const float X0, const float Y0, const float Vx0, const float Vy0, CvMouthTracker* MouthTracker);

OPENCVAUXAPI CVStatus
icvUpdateKalmanXY(const float X, const float Y, CvMouthTracker* MouthTracker, float &SmX, float &SmY);

OPENCVAUXAPI void 
icvReleaseKalmanXY(CvMouthTracker* MouthTracker);

/* for mouth tracking postprocess */
OPENCVAUXAPI bool
icvCheckMouthTrackResult(CvMouthLocateResult &result);

// regularize tracking location
OPENCVAUXAPI CVStatus
icvRegularTrack(CvMouthLocateResult * Trajectory, const int FrameNum);

// gaussian filtering
OPENCVAUXAPI void 
icvGaussianSmooth(double* Array, const int ArrayLength, double *Filter, const int FilterLength, const int Variance);

OPENCVAUXAPI CVStatus
icvGaussianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance);

OPENCVAUXAPI CVStatus
icvGaussianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance);

// mid-value filtering
OPENCVAUXAPI void 
icvMedianValueSmooth(double *Array, const int ArrayLength, const int FilterRadius);

OPENCVAUXAPI CVStatus
icvMedianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius);

OPENCVAUXAPI CVStatus
icvMedianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius);

#ifdef __cplusplus
}
#endif

#endif		// #ifndef _CV_MOUTH_TRACK_H_