/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "cvVisualSpeech.h"

#include "cvavsrfeatureIO.h"


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			cvInitLDA( const CvMat* labels, const CvMat* inVectors, const int numClasses, 
//					   CvMat *numSamplesInClass, CvMat *sumVectorInClass, 
//			           CvMat **pCovMatrixInClass, CvMat *normFactor);
//    Purpose:
//			The function cvInitLDA calculates the number of samples, sum vectors and partial 
//			covariance matrices for each class, and the normalization factor of all input vectors. 
//			The output matrices must be allocated before call this function. Given the c-class 
//			n-dimensional input vectors, the matrices numSamplesInClass, sumVectorInClass, 
//			pCovMatrixInClass and normFactor are of size c1, cn, c list of nn and 1n 
//			matrices respectively.
//    Context:
//    Parameters:
//			labels		-- [I] class labels, each element is corresponding to one input vectors in inVectors.
//							   For C classes the value of the labels are in the range [0,C-1].
//			inVectors	-- [I] input data, each row represents one sample vector.
//			numClasses	-- [I] number of classes
//			numSamplesInClass -- [O] the number of samples in each class, the number of samples in the ith class are stored in the ith row
//          sumVectorInClass  -- [O] the sum vector for each class, the ith contains the sum vector of all vectors in the  ith class
//			pCovMatrixInClass -- [O] the sequence of partial covariance matrices for each class
//			normFactor -- [O] normalization factor for each dimension of the input vectors
//    Returns: 
//			return Cv_StsOK if success, else report error;
//    Notes:
//F*/
CV_IMPL  CVStatus 
cvInitLDA(const CvMat* labels, const CvMat* inVectors, const int numClasses, 
		  CvMat* numSamplesInClass, CvMat* sumVectorInClass, 
		  CvMat** pCovMatrixInClass, CvMat* normFactor)
{
	int i, j, n;
	int labelnum;	 
	int vecsize;	// vector length
	int fmnum;		// # of frames in each feature matrix
	int featuretype, labeltype, outtype;	// input/output element data type	
	double temp1, temp2;
	CvMat* temp_input_total_per_class = NULL;
	CvMat* temp_input_sq_total = NULL;
	CvMat* temp_class_cov[CV_LDA_MAX_CLASSES];

	CV_FUNCNAME( "cvInitLDA" );

   	__BEGIN__;

	if( !CV_IS_MAT(labels) || !CV_IS_MAT(inVectors) )
		CV_ERROR(CV_StsBadArg, "Invalid input matrix!");
	
	if( !CV_IS_MAT(numSamplesInClass) || !CV_IS_MAT(sumVectorInClass) 
		|| !CV_IS_MAT(pCovMatrixInClass[0]) || !CV_IS_MAT(normFactor) ) 
		CV_ERROR(CV_StsBadArg, "Invalid output matrix!");
		
	if( !CV_ARE_TYPES_EQ(normFactor, sumVectorInClass) 
		|| !CV_ARE_TYPES_EQ(normFactor, pCovMatrixInClass[0]) 
		|| !CV_ARE_TYPES_EQ(normFactor, numSamplesInClass) )
		CV_ERROR(CV_StsBadArg, "All output matrices should have same type.");
	
	CV_CALL( outtype = cvGetElemType(normFactor) );
	if (outtype != CV_32F && outtype != CV_64F)
		CV_ERROR(CV_StsBadArg, "Invalid data type if output data!");
	
	CV_CALL( featuretype = cvGetElemType(inVectors) );
	if (featuretype != CV_32F && featuretype != CV_64F)
		CV_ERROR(CV_StsBadArg, "Invalid data type in input feature data.");
	
	CV_CALL( labeltype = cvGetElemType(labels) );
	if (labeltype != CV_8U)
		CV_ERROR(CV_StsBadArg, "Invalid data type in input labels data.");
	
	vecsize = inVectors->cols;
	fmnum = inVectors->rows;
	if(fmnum < 1) 
		CV_ERROR(CV_StsBadSize, "Feature vector should be loaded before.");
	
	// initialize matrices to 0
	for( i = 0; i < numClasses; i++ )
		cvSetZero( pCovMatrixInClass[i]);
	cvSetZero(normFactor);
	cvSetZero(sumVectorInClass);
	cvSetZero(numSamplesInClass);
	
	if( outtype == CV_32F )
	{
		// allocate temp buffer
		CV_CALL( temp_input_total_per_class = cvCreateMat(numClasses, vecsize, CV_64F) );
		if(temp_input_total_per_class == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		
		CV_CALL( temp_input_sq_total = cvCreateMat(1, vecsize, CV_64F) );
		if(temp_input_sq_total == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );

		for(i=0; i < numClasses; i++)
		{
			CV_CALL( temp_class_cov[i] = cvCreateMat(vecsize, vecsize, CV_64F) );
			if(temp_class_cov[i] == NULL)
				CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		}
		// init temp buffer to 0
		cvSetZero(temp_input_total_per_class);
		cvSetZero(temp_input_sq_total);		
		for(i=0; i < numClasses; i++ )
			cvSetZero( temp_class_cov[i]);
		
		// begin to calculate feature
		for (n=0; n<fmnum; n++) 
		{
			labelnum = labels->data.i[n];
			if (labelnum < -1 || labelnum >= numClasses)  
				CV_ERROR(CV_StsBadSize, "Invalid class labels.");
		
			if (labelnum == -1) 
				continue;		// empty feature
			// calc normFactor, sumVectorInClass, pCovMatrixInClass.
			if (featuretype == CV_32F) 
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = (double)(inVectors->data.fl[n*vecsize+i]);
					temp_input_total_per_class->data.db[labelnum*vecsize+i] += temp1;
					temp_input_sq_total->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = (double)(inVectors->data.fl[n*vecsize+j]);
						temp_class_cov[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					} 
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.fl[labelnum] += 1.0;
			}
			else  // featuretype == CV_64F
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = inVectors->data.db[n*vecsize+i];
					temp_input_total_per_class->data.db[labelnum*vecsize+i] += temp1;
					temp_input_sq_total->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = inVectors->data.db[n*vecsize+j];
						temp_class_cov[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					}
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.fl[labelnum] += 1.0;
			}
		} 
		// end for calculating
		
		// copy data from temp buffer to output matrix
		for(n=0; n < numClasses; n++)
			for(i=0; i < vecsize; i++)
				sumVectorInClass->data.fl[n*vecsize+i] = 
				(float)temp_input_total_per_class->data.db[n*vecsize+i];
		
		for(i=0; i < vecsize; i++)
			normFactor->data.fl[i] = 
			(float)temp_input_sq_total->data.db[i];

		for(n=0; n < numClasses; n++)
			for(i=0; i < vecsize; i++)
				for(j=0; j < vecsize; j++)
					pCovMatrixInClass[n]->data.fl[i*vecsize+j] = 
					(float)temp_class_cov[n]->data.db[i*vecsize+j];
		// free temp buffer
		cvReleaseMat(&temp_input_total_per_class);
		cvReleaseMat(&temp_input_sq_total);
		for(i=0; i < numClasses; i++)
			cvReleaseMat(&temp_class_cov[i]);
	}
	else	// outtype == CV_64F
	{
		// begin to calculate feature
		for (n=0; n<fmnum; n++) 
		{
			labelnum = labels->data.i[n];
			if (labelnum < -1 || labelnum >= numClasses)  
				CV_ERROR(CV_StsBadSize, "Invalid class labels.");
		
			if (labelnum == -1) 
				continue;		// empty feature
			// calc input_total, normFactor, sumVectorInClass, pCovMatrixInClass.
			if (featuretype == CV_32F) 
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = (double)(inVectors->data.fl[n*vecsize+i]);
					sumVectorInClass->data.db[labelnum*vecsize+i] += temp1;
					normFactor->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = (double)(inVectors->data.fl[n*vecsize+j]);
						pCovMatrixInClass[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					} 
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.db[labelnum] += 1.0;
			}
			else  // featuretype == CV_64F
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = inVectors->data.db[n*vecsize+i];
					sumVectorInClass->data.db[labelnum*vecsize+i] += temp1;
					normFactor->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = inVectors->data.db[n*vecsize+j];
						pCovMatrixInClass[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					}
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.db[labelnum] += 1.0;
			}
		} 
		// end for calculating
	} // end if
		
	return CV_StsOk;

	__END__;
	
	cvFree((void**)&temp_input_total_per_class);
	cvFree((void**)&temp_input_sq_total);
	for(i=0; i < numClasses; i++)
		cvFree((void**)&temp_class_cov[i]);

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvUpdateLDA(const CvMat* labels, const CvMat* inVectors, const int numClasses, 
//					    CvMat *numSamplesInClass, CvMat *sumVectorInClass, 
//						CvMat **pCovMatrixInClass, CvMat *normFactor);
//    Purpose:
//			The function cvUpdateLDA updates the number of samples, sum vectors and partial covariance 
//			matrices without for each class, and the normalization factor, from the values obtained 
//			with cvInitLDA. The sizes of matrices numSamplesInClass, sumVectorInClass, pCovMatrixInClass 
//			and normFactor must match that of the input feature data and class number. 
//			See the description of the size of each matrix in function cvInitLDA.
//    Context:
//    Parameters:
//			labels		-- [I] class labels, each element is corresponding to one input vectors in inVectors. 
//								For C classes the value of the labels are in the range [0,C-1].
//			inVectors	-- [I] input data, each row represents one sample vector.
//			numClasses	-- [I] number of classes 
//			numSamplesInClass	-- [I/O] the number of samples in each class updated "in place". 
//          sumVectorInClass	-- [I/O] the values of the sum vectors updated "in place".
//			pCovMatrixInClass	-- [I/O] the partial covariance matrices for each class updated 'in place". 
//			normFactor	-- [I/O] the normalization factor updated "in place".
//    Returns: 
//			return Cv_StsOK if success, else report error;
//    Notes:
//F*/

CV_IMPL  CVStatus 
cvUpdateLDA( const CvMat* labels, const CvMat* inVectors, const int numClasses, 
			 CvMat *numSamplesInClass, CvMat *sumVectorInClass, 
			 CvMat **pCovMatrixInClass, CvMat *normFactor)
{
	int i, j, n;
	int labelnum;	 
	int vecsize;	// vector length
	int fmnum;		// # of frames in each feature matrix
	int featuretype, outtype;	// input element data type	
	double temp1, temp2;
	CvMat* temp_input_total_per_class = NULL;
	CvMat* temp_input_sq_total = NULL;
	CvMat* temp_class_cov[CV_LDA_MAX_CLASSES];

	CV_FUNCNAME( "cvUpdateLDA" );

   	__BEGIN__;

	if( !CV_IS_MAT(labels) || !CV_IS_MAT(inVectors) )
		CV_ERROR(CV_StsBadArg, "Invalid input matrix!");
	
	if( !CV_IS_MAT(numSamplesInClass) || !CV_IS_MAT(sumVectorInClass) 
		|| !CV_IS_MAT(pCovMatrixInClass[0]) || !CV_IS_MAT(normFactor) ) 
		CV_ERROR(CV_StsBadArg, "Invalid output matrix!");
		
	if( !CV_ARE_TYPES_EQ(normFactor, sumVectorInClass) 
		|| !CV_ARE_TYPES_EQ(normFactor, pCovMatrixInClass[0]) 
		|| !CV_ARE_TYPES_EQ(normFactor, numSamplesInClass) )
		CV_ERROR(CV_StsBadArg, "All output matrices should have same type.");

	CV_CALL( outtype = cvGetElemType(normFactor) );
	if (outtype != CV_32F && outtype != CV_64F)
		CV_ERROR(CV_StsBadArg, "Invalid data type in output data!");

	CV_CALL( featuretype = cvGetElemType(inVectors) );
	if (featuretype != CV_32F && featuretype != CV_64F)
		CV_ERROR(CV_StsBadArg, "Invalid data type in input feature data.");

	vecsize = inVectors->cols;
	fmnum = inVectors->rows;
	if(fmnum < 1) 
		CV_ERROR(CV_StsBadSize, "Feature vector should be loaded before.");

	if( outtype == CV_32F )
	{
		// allocate temp buffer
		CV_CALL( temp_input_total_per_class = cvCreateMat(numClasses, vecsize, CV_64F) );
		if(temp_input_total_per_class == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		
		CV_CALL( temp_input_sq_total = cvCreateMat(1, vecsize, CV_64F) );
		if(temp_input_sq_total == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );

		for(i=0; i < numClasses; i++)
		{
			CV_CALL( temp_class_cov[i] = cvCreateMat(vecsize, vecsize, CV_64F) );
			if(temp_class_cov[i] == NULL)
				CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		}
		// init temp buffer to 0
		cvSetZero(temp_input_total_per_class);
		cvSetZero(temp_input_sq_total);		
		for(i=0; i < numClasses; i++ )
			cvSetZero( temp_class_cov[i]);
		
		// begin to calculate feature
		for (n=0; n<fmnum; n++) 
		{
			labelnum = labels->data.i[n];
			if (labelnum < -1 || labelnum >= numClasses)  
				CV_ERROR(CV_StsBadSize, "Invalid class labels.");
		
			if (labelnum == -1) 
				continue;		// empty feature
			// calc normFactor, sumVectorInClass, pCovMatrixInClass.
			if (featuretype == CV_32F) 
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = (double)(inVectors->data.fl[n*vecsize+i]);
					temp_input_total_per_class->data.db[labelnum*vecsize+i] += temp1;
					temp_input_sq_total->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = (double)(inVectors->data.fl[n*vecsize+j]);
						temp_class_cov[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					} 
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.fl[labelnum] += 1.0;
			}
			else  // featuretype == CV_64F
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = inVectors->data.db[n*vecsize+i];
					temp_input_total_per_class->data.db[labelnum*vecsize+i] += temp1;
					temp_input_sq_total->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = inVectors->data.db[n*vecsize+j];
						temp_class_cov[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					}
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.fl[labelnum] += 1.0;
			}
		} 
		// end for calculating
		
		// inc data from temp buffer to output matrix
		for(n=0; n < numClasses; n++)
			for(i=0; i < vecsize; i++)
				sumVectorInClass->data.fl[n*vecsize+i] += 
				(float)temp_input_total_per_class->data.db[n*vecsize+i];
		
		for(i=0; i < vecsize; i++)
			normFactor->data.fl[i] += 
			(float)temp_input_sq_total->data.db[i];

		for(n=0; n < numClasses; n++)
			for(i=0; i < vecsize; i++)
				for(j=0; j < vecsize; j++)
					pCovMatrixInClass[n]->data.fl[i*vecsize+j] += 
					(float)temp_class_cov[n]->data.db[i*vecsize+j];
		// free temp buffer
		cvReleaseMat(&temp_input_total_per_class);
		cvReleaseMat(&temp_input_sq_total);
		for(i=0; i < numClasses; i++)
			cvReleaseMat(&temp_class_cov[i]);
	}
	else	// outtype == CV_64F
	{
		// begin to calculate feature
		for (n=0; n<fmnum; n++) 
		{
			labelnum = labels->data.i[n];
			if (labelnum < -1 || labelnum >= numClasses)  
				CV_ERROR(CV_StsBadSize, "Invalid class labels.");
		
			if (labelnum == -1) 
				continue;		// empty feature
			// calc input_total, normFactor, sumVectorInClass, pCovMatrixInClass.
			if (featuretype == CV_32F) 
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = (double)(inVectors->data.fl[n*vecsize+i]);
					sumVectorInClass->data.db[labelnum*vecsize+i] += temp1;
					normFactor->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = (double)(inVectors->data.fl[n*vecsize+j]);
						pCovMatrixInClass[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					} 
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.db[labelnum] += 1.0;
			}
			else  // featuretype == CV_64F
			{
				for ( i = 0; i < vecsize; i++ ) 
				{
					temp1 = inVectors->data.db[n*vecsize+i];
					sumVectorInClass->data.db[labelnum*vecsize+i] += temp1;
					normFactor->data.db[i] += temp1 * temp1;
					for ( j = 0; j < vecsize; j++ )
					{
						temp2 = inVectors->data.db[n*vecsize+j];
						pCovMatrixInClass[labelnum]->data.db[i*vecsize+j] += temp1 * temp2;
					}
				}
				//accumulate feature frames # of per class
				numSamplesInClass->data.db[labelnum] += 1.0;
			}
		} 
		// end for calculating
	} // end if
		
	return CV_StsOk;

	__END__;
	
	cvFree((void**)&temp_input_total_per_class);
	cvFree((void**)&temp_input_sq_total);
	for(i=0; i < numClasses; i++)
		cvFree((void**)&temp_class_cov[i]);

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvCalcLDASwSb( const CvMat* numSamplesInClass, const CvMat* sumVectorInClass, 
//						   CvMat** pCovMatrixInClass, const CvMat* normFactor,
//			               CvMat* withinClassScatter, CvMat* betweenClassScatter,
//			               CvMat* meanVectors, CvMat* mse);
//    Purpose:
//			The function cvCalcLDASwSb calculates within-class scatter and between-scatter matrices 
//			as well as the means and mean square error of all the sample data used. The output matrices 
//			must be allocated before calling this function. Suppose the input features are n-dimensional, 
//			the matrices withinClassScatter, betweenClassScatter, meanVectors and mse are of size nn, 
//			nn, n1 and n1 respectively.
//    Context:
//    Parameters:
//			numSamplesInClass -- [I] the number of samples in each class calculated by 
//									 function cvInitLDA or cvUpdateLDA.
//			sumVectorInClass  -- [I] the sum vectors of the samples in each class calculated by function cvInitLDA or cvLDATotal
//			pCovMatrixInClass -- [I] the partial covariance matrices for each class calculated by function cvInitLDA or cvUpdateLDA
//			normFactor	-- [I] the normalization factor calculated by function cvInitLDA or cvUpdateLDA.
//			withinClassScatter	-- [O] the within-class scatter matrix.
//			betweenClassScatter -- [O] the between-class scatter matrix.
//			meanVectors	-- [O] the mean vector of all the sample data
//			mse			-- [O] the mean square error of all the sample data used.
//    Returns: 
//			return Cv_StsOK if successfully detect a face, else report error;
//    Notes:
//F*/

CV_IMPL  CVStatus 
cvCalcLDASwSb( const CvMat* numSamplesInClass, const CvMat* sumVectorInClass, 
			   CvMat** pCovMatrixInClass, const CvMat* normFactor,
               CvMat* withinClassScatter, CvMat* betweenClassScatter,
               CvMat* meanVectors, CvMat* mse)
{
	int i, j, k;
	int numClasses;
	int vecsize;	// vector length
	int type;		// input element data type	
	double total_trials = 0;	// total # of samples	
	double temp1, var_tmp;	
	CvMat *input_total = NULL;	// accumulate total input featrues.
	CvMat *class_means = NULL;	// class sample means	
	CvMat *tempbuf1 = NULL, *tempbuf2 = NULL;

	CV_FUNCNAME( "cvUpdateLDA" );

   	__BEGIN__;

	if( !CV_IS_MAT(numSamplesInClass) || !CV_IS_MAT(sumVectorInClass) 
		|| !CV_IS_MAT(pCovMatrixInClass[0]) || !CV_IS_MAT(normFactor) )
		CV_ERROR(CV_StsBadArg, "Invalid input matrix!");
	
	if( !CV_IS_MAT(withinClassScatter) || !CV_IS_MAT(betweenClassScatter) 
		|| !CV_IS_MAT(meanVectors) || !CV_IS_MAT(mse) ) 
		CV_ERROR(CV_StsBadArg, "Invalid output matrix!");
		
	if( !CV_ARE_TYPES_EQ(sumVectorInClass, numSamplesInClass) 		
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, pCovMatrixInClass[0]) 
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, normFactor) 
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, withinClassScatter) 
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, betweenClassScatter) 
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, meanVectors) 
		|| !CV_ARE_TYPES_EQ(sumVectorInClass, mse) )
		CV_ERROR(CV_StsBadArg, "All input output matrices should have same type.");

	CV_CALL( type = cvGetElemType(meanVectors) );
	if (type != CV_32F && type != CV_64F)
		CV_ERROR(CV_StsBadArg, "Invalid data type in output data.");

	numClasses = sumVectorInClass->rows;
	vecsize = sumVectorInClass->cols;

	if (type == CV_32F) 
	{
		// if any class samples missed in the feature data
		for (k = 0; k < numClasses; k++) 
		{
			if (numSamplesInClass->data.fl[k] == 0) 
			{
				CV_ERROR( CV_DIV_BY_ZERO_ERR, "A class of samples don't exist.");
			}
		} // end for k

		// calculate total value of all the feature data, total_trails
		CV_CALL( input_total = cvCreateMat(1, vecsize, type) );
		if( input_total == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		cvSetZero(input_total);
		total_trials = 0;
		for ( i = 0; i < numClasses; i++ ) 
		{
			total_trials += numSamplesInClass->data.fl[i];
			for (j=0; j<vecsize; j++) 
			{
				input_total->data.fl[j] += sumVectorInClass->data.fl[i*vecsize+j];
			}
		}
		
		// assign matrix and initialize to 0
		CV_CALL( class_means = cvCreateMat(numClasses, vecsize, type) );
		if( class_means == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		cvSetZero(class_means);
		// calc meanVectors, mse and class_means
   		for (i = 0; i < vecsize; i++) 
		{
			temp1 = input_total->data.fl[i] / total_trials;
			meanVectors->data.fl[i] = (float)temp1;
			var_tmp = normFactor->data.fl[i] / total_trials - (temp1 * temp1);
			if(var_tmp < CV_LDA_SMALLEST) 
			{				
				var_tmp = CV_LDA_SMALLEST;
			}
			mse->data.fl[i] = (float)sqrt(var_tmp);
			for (k = 0; k < numClasses; k++)
			{
				class_means->data.fl[k * vecsize + i] = 
					sumVectorInClass->data.fl[k * vecsize + i] / numSamplesInClass->data.fl[k];
			} 
		} //end for i
		
		// compute within_scatter_matrix and between_scatter_matrix
		CV_CALL( tempbuf1  =  cvCreateMat(numClasses, vecsize, type) );	// kxn
		if( tempbuf1 == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		CV_CALL( tempbuf2  =  cvCreateMat(vecsize, numClasses, type) );	// nxk
		if( tempbuf2 == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );

		// begin to compute within-class scatter matrix
		for ( j = 0; j < vecsize; j++ ) 
		{
			for ( i = 0; i < vecsize; i++ ) 
			{
				temp1 = 0;
				for (k = 0; k < numClasses; k++ ) 
				{
					temp1 += pCovMatrixInClass[k]->data.fl[i*vecsize+j] /numSamplesInClass->data.fl[k] -
						(class_means->data.fl[k*vecsize+i] * class_means->data.fl[k*vecsize+j]);
				}
				withinClassScatter->data.fl[i*vecsize+j] = (float)temp1;
			}
		}

		// begin to compute between-class scatter matrix
		// tempbuf1 = class_mean(c)-feature_mean
		for( k = 0; k < numClasses; k++ ) 
		{
			for( i = 0; i < vecsize; i++)
			{
				tempbuf1->data.fl[k*vecsize+i] = class_means->data.fl[k*vecsize+i] - meanVectors->data.fl[i];
			}
		}

		// tempbuf2 = (class_mean(c)-feature_mean)(t)
		CV_CALL( cvTranspose(tempbuf1, tempbuf2) );

		// tempbuf2[i][k] = L(c) * (class_mean(c)-feature_mean)(t)
		for(i = 0; i < vecsize; i++ )
		{
			for ( k = 0; k < numClasses; k++ )
			{
				tempbuf2->data.fl[i*numClasses+k] *= numSamplesInClass->data.fl[k];
			}
		}
				
		// Sb= sigma(c)L(c)(class_mean(c)-feature_mean)(class_mean(c)-feature_mean)(t)
		CV_CALL( cvMatMulAdd(tempbuf2, tempbuf1, 0, betweenClassScatter) ); 
		// Sb= sigma(c)Pr(c)(class_mean(c)-feature_mean)(class_mean(c)-feature_mean)(t)
		for(i = 0; i < vecsize; i++ )
		{
			for ( j = 0; j < vecsize; j++ )
			{
				betweenClassScatter->data.fl[i*vecsize+j] /= (float)total_trials;		
			}
		}

		// release memory
		cvReleaseMat(&input_total);	
		cvReleaseMat(&class_means);
		cvReleaseMat(&tempbuf1);
		cvReleaseMat(&tempbuf2);
	}
	else // type == CV_64F
	{
		// if any class samples missed in the feature data
		for (k = 0; k < numClasses; k++) 
		{
			if (numSamplesInClass->data.db[k] == 0) 
			{
				CV_ERROR( CV_DIV_BY_ZERO_ERR, "A class of samples don't exist.");
			}
		} // end for k

		// calculate total value of all the feature data, total_trails
		CV_CALL( input_total = cvCreateMat(1, vecsize, type) );
		if( input_total == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		cvSetZero(input_total);
		total_trials = 0;
		for ( i = 0; i < numClasses; i++ ) 
		{
			total_trials += numSamplesInClass->data.db[i];
			for (j=0; j<vecsize; j++) 
			{
				input_total->data.db[j] += sumVectorInClass->data.db[i*vecsize+j];
			}
		}
		
		// assign matrix and initialize to 0
		CV_CALL( class_means = cvCreateMat(numClasses, vecsize, type) );
		if( class_means == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		cvSetZero(class_means);
		// calc meanVectors, mse and class_means
   		for (i = 0; i < vecsize; i++) 
		{
			temp1 = input_total->data.db[i] / total_trials;
			meanVectors->data.db[i] = temp1;
			var_tmp = normFactor->data.db[i] / total_trials - (temp1 * temp1);
			if(var_tmp < CV_LDA_SMALLEST) 
			{	
				var_tmp = CV_LDA_SMALLEST;
			}
			mse->data.db[i] = (float)sqrt(var_tmp);
			for (k = 0; k < numClasses; k++)
			{
				class_means->data.db[k * vecsize + i] = 
					sumVectorInClass->data.db[k * vecsize + i] / numSamplesInClass->data.db[k];
			} 
		} //end for i
		
		// compute within_scatter_matrix and between_scatter_matrix
		CV_CALL( tempbuf1  =  cvCreateMat(numClasses, vecsize, type) );	// kxn
		if( tempbuf1 == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );
		CV_CALL( tempbuf2  =  cvCreateMat(vecsize, numClasses, type) );	// nxk
		if( tempbuf2 == NULL)
			CV_ERROR( CV_StsNoMem, "Can't allocate temp matrix! Not enough memory!" );

		// begin to compute within-class scatter matrix
		for ( j = 0; j < vecsize; j++ ) 
		{
			for ( i = 0; i < vecsize; i++  ) 
			{
				temp1 = 0;
				for (k = 0; k < numClasses; k++ ) 
				{
					temp1 += pCovMatrixInClass[k]->data.db[i*vecsize+j] /numSamplesInClass->data.db[k] -
						(class_means->data.db[k*vecsize+i] * class_means->data.db[k*vecsize+j]);
				}
				withinClassScatter->data.db[i*vecsize+j] = temp1;
			}
		}

		// begin to compute between-class scatter matrix
		// tempbuf1 = class_mean(c)-feature_mean
		for( k = 0; k < numClasses; k++ ) 
		{
			for( i = 0; i < vecsize; i++)
			{
				tempbuf1->data.db[k*vecsize+i] = class_means->data.db[k*vecsize+i] - meanVectors->data.db[i];
			}
		}

		// tempbuf2 = (class_mean(c)-feature_mean)(t)
		CV_CALL( cvTranspose(tempbuf1, tempbuf2) );

		// tempbuf2[i][k] = L(c) * (class_mean(c)-feature_mean)(t)
		for(i = 0; i < vecsize; i++ )
		{
			for ( k = 0; k < numClasses; k++ )
			{
				tempbuf2->data.db[i*numClasses+k] *= numSamplesInClass->data.db[k];
			}
		}
				
		// Sb= sigma(c)L(c)(class_mean(c)-feature_mean)(class_mean(c)-feature_mean)(t)
		CV_CALL( cvMatMulAdd(tempbuf2, tempbuf1, 0, betweenClassScatter) ); 
		//CV_CALL( icvStdMatMul(tempbuf2, tempbuf1, betweenClassScatter) ); 
		// Sb= sigma(c)Pr(c)(class_mean(c)-feature_mean)(class_mean(c)-feature_mean)(t)
		for(i = 0; i < vecsize; i++ )
		{
			for ( j = 0; j < vecsize; j++ )
			{
				betweenClassScatter->data.db[i*vecsize+j] /= total_trials;		
			}
		}

		// release memory
		cvReleaseMat(&input_total);	
		cvReleaseMat(&class_means);
		cvReleaseMat(&tempbuf1);		
		cvReleaseMat(&tempbuf2);		
	}

	return CV_StsOk;

	__END__;
	
	cvFree((void**)&input_total);
	cvFree((void**)&class_means);
	cvFree((void**)&tempbuf1);
	cvFree((void**)&tempbuf2);

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			cvCalcLDAMatrix( CvMat* withinClassScatter, CvMat* betweenClassScatter, 
//							 CvMat* projMatrix, CvMat* eigenvalues);
//    Purpose:
//			The function cvCalcLDAMatrix calculates the LDA projection matrix A and 
//			the corresponding eigenvalues using the algorithm described above. 
//			The output matrices must be allocated before calling this function. Suppose the input features 
//			are n-dimensional and the destination feature space is m-dimensional, the size of projMatrix 
//			and eigenvalue should be mn and m1 respectively.
//    Context:
//    Parameters:
//			withinClassScatter	--- [I] the within-class scatter matrix.
//			betweenClassScatter	--- [I] the between-class scatter matrix.
//			projMatrix		--- [O] the LDA projection matrix 
//			eigenvalues	--- [O] eigenvalues obtained by the 2nd SVD in the algorithm
//    Returns: 
//			return Cv_StsOK if successfully detect a face, else report error;
//F*/
CV_IMPL  CVStatus 
cvCalcLDAMatrix( CvMat* withinClassScatter, CvMat* betweenClassScatter, 
				 CvMat* projMatrix, CvMat* eigenvalues)
{
	int i, j, k, index;
	int vecsize;			// square matrix size n
	int maxfeatures1, maxfeatures2, rowcopy; // coordinate of the max # of features due to eigenvalue threshold
	int type;	
	double sum = 0.0;		// trace of eigen vectors
    double mineigenvalue;	// relative min matDb, Dw threshold
	double tmp;				// temp variables 
	
	CvMat *ytranspose = NULL; //y(t)*Sb*y = Db
	CvMat *utranspose = NULL; //z(t)*Sw*z = Dw
	CvMat *matDb = NULL;	
	CvMat *matDw = NULL;	
	CvMat *ztsbz = NULL;
	CvMat *tempmat = NULL;
	CvMat *zbuf = NULL;		// z= y*d^(-1/2)
	CvMat *ztranspose = NULL;	 
	CvMat *mata = NULL;

	CV_FUNCNAME( "cvCalcLDAMatrix" );

    __BEGIN__;

	if( !CV_IS_MAT(withinClassScatter) || !CV_IS_MAT(betweenClassScatter) || 
		!CV_IS_MAT(projMatrix) || !CV_IS_MAT(eigenvalues) )
		CV_ERROR(CV_StsBadArg, "Invalid input argument!");
		
	if( !CV_ARE_TYPES_EQ(withinClassScatter, betweenClassScatter) 
		|| !CV_ARE_TYPES_EQ(withinClassScatter, projMatrix) 
		|| !CV_ARE_TYPES_EQ(withinClassScatter, eigenvalues) )
		CV_ERROR(CV_StsBadArg, "All input/ouput matrices should have the same type!");
		
	vecsize = withinClassScatter->cols;
	CV_CALL( type = cvGetElemType(withinClassScatter) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR(CV_StsUnmatchedFormats, "Input data type must be float or double!");	
	if( type == CV_32F )
	{
		// assign memory
		CV_CALL( ytranspose = cvCreateMat(vecsize, vecsize, type) );	// n*n
		CV_CALL( matDb = cvCreateMat(vecsize, vecsize, type) );	// n*n
		
		//singular value decomposition Y(t)*Sw*Y = Dw
		CV_CALL( cvSVD( betweenClassScatter, matDb, ytranspose, 0, CV_SVD_U_T) ); // ytranspose = Y(t)
		// set eigenvectors matrix to positive
		CV_CALL( sum = cvmTrace(ytranspose) );
		if (sum < 0.0) 
		{
			index = 0;
			for (i=0; i<vecsize; i++) 
			{
				for (j=0; j<vecsize; j++) 
				{
					ytranspose->data.fl[index] = - ytranspose->data.fl[index];
					index++;
				}
			}
		}

		//set the Db threshold
		mineigenvalue = matDb->data.fl[0] / CV_LDA_MAX_RATIO;
		maxfeatures1 = vecsize;
		for (i=0; i<matDb->rows; i++)
		{
			if ( matDb->data.fl[i*matDb->rows+i] < mineigenvalue ) 
			{
				maxfeatures1 = i; 
				break;
			}
		}
		// allocate buffer
		CV_CALL( ztranspose = cvCreateMat(maxfeatures1, vecsize, type) );
		CV_CALL( zbuf = cvCreateMat(vecsize, maxfeatures1, type) );
		CV_CALL( tempmat = cvCreateMat(maxfeatures1, vecsize, type) );
		CV_CALL( ztsbz = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( matDw = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( utranspose = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( mata = cvCreateMat(maxfeatures1, vecsize, type) );	// n*n

		// Z(t)=(D^(-1/2)) * Y(t)
		index = 0;
		for(i=0; i<ztranspose->rows; i++) 
		{
			tmp = sqrt(matDb->data.fl[i*matDb->cols+i]);		
			for (j=0; j<ztranspose->cols; j++) 
			{
				ztranspose->data.fl[index] = (float)(ytranspose->data.fl[index] / tmp);
				index ++;
			}
		}
		// Y(t) * Sw
        CV_CALL( cvMatMulAdd(ztranspose, withinClassScatter, 0, tempmat) ); 
		// compute Z
		CV_CALL( cvTranspose(ztranspose, zbuf) ); // zbuf = Z
		// Z(t) * Sw * Z
		CV_CALL( cvMatMulAdd(tempmat, zbuf, 0, ztsbz) ); 
		//singular value decomposition Z(t)*Sb*Z
		CV_CALL( cvSVD( ztsbz, matDw, utranspose, 0, CV_SVD_U_T) ); //yt=U(t)
		// set eigenvectors matrix to positive
		CV_CALL( sum = cvmTrace(utranspose) );
		if (sum < 0.0) 
		{
			index = 0;
			for (i=0; i<utranspose->rows; i++) 
			{
				for (j=0; j<utranspose->cols; j++) 
				{
					utranspose->data.fl[index] = - utranspose->data.fl[index];
					index ++;
				}
			}
		}		
		
		//set the Dw threshold
		mineigenvalue = matDw->data.fl[0] / CV_LDA_MAX_RATIO;
		maxfeatures2 = vecsize;
		for (i=0; i<matDw->rows; i++)
		{
			if ( matDw->data.fl[i*matDw->cols+i] < mineigenvalue ) 
			{
				maxfeatures2 = i; 
				break;
			}
		}
		
		// make transform matrix A=U(t)*Z(t)
		CV_CALL( cvMatMulAdd(utranspose, ztranspose, 0, mata) );	// TranMat = U(t)*Z(t)
		
		// sorted projMatrix according to Dw of Z(t)*Sb*Z 
		for (i=0; i<matDw->rows - 1; i++)
		{
			for (j=0; j<matDw->cols - 1; j++)
			{
				if (matDw->data.fl[(j+1)*matDw->cols+j+1] > matDw->data.fl[j*matDw->cols+j]) 
				{
					tmp = (double)matDw->data.fl[j*matDw->cols+j];
					matDw->data.fl[j*matDw->cols+j] = matDw->data.fl[(j+1)*matDw->cols+j+1];
					matDw->data.fl[(j+1)*matDw->cols+j+1] = (float)tmp;
					for (k=0; k<mata->cols; k++) 
					{
						tmp = (double)mata->data.fl[j*mata->cols+k];
						mata->data.fl[j*mata->cols+k] = mata->data.fl[(j+1)*mata->cols+k];
						mata->data.fl[(j+1)*mata->cols+k] = (float)tmp;
					}
				} 
			} 
		}// end for sorting
		
		rowcopy = mata->rows < projMatrix->rows ? mata->rows : projMatrix->rows;
		index = 0;
		for(i=0; i<rowcopy; i++) 
		{
			eigenvalues->data.fl[i] = matDw->data.fl[i*matDw->cols+i];
			for(j=0;j<vecsize;j++) 
			{
				projMatrix->data.fl[index] = mata->data.fl[index];// / (float)tmp;
				index ++;
			}
		}		
		// set values and eigenvetors beyond rowcopy to zero
		for(i=rowcopy; i<projMatrix->rows; i++) 
		{
			eigenvalues->data.fl[i] = 0.0f;
			for(j=0;j<vecsize;j++) 
			{
				projMatrix->data.fl[index] = 0.0f;
				index ++;
			}
		}		
		// free temp buffer
		CV_CALL( cvReleaseMat(&ytranspose) );
		CV_CALL( cvReleaseMat(&matDb) );
		CV_CALL( cvReleaseMat(&ztranspose) );
		CV_CALL( cvReleaseMat(&zbuf) );
		CV_CALL( cvReleaseMat(&tempmat) );
		CV_CALL( cvReleaseMat(&ztsbz) );
		CV_CALL( cvReleaseMat(&matDw) );
		CV_CALL( cvReleaseMat(&utranspose) );
		CV_CALL( cvReleaseMat(&mata) );
	}
	else	// type == CV_64F
	{
		// assign memory
		CV_CALL( ytranspose = cvCreateMat(vecsize, vecsize, type) );	// n*n
		CV_CALL( matDb = cvCreateMat(vecsize, vecsize, type) );	// n*n
		
		//singular value decomposition Y(t)*Sw*Y = Dw
		CV_CALL( cvSVD( betweenClassScatter, matDb, ytranspose, 0, CV_SVD_U_T) ); // ytranspose = Y(t)
		
		// set eigenvectors matrix to positive
		CV_CALL( sum = cvmTrace(ytranspose) );
		if (sum < 0.0) 
		{
			index = 0;
			for (i=0; i<vecsize; i++) 
			{
				for (j=0; j<vecsize; j++) 
				{
					ytranspose->data.db[index] = - ytranspose->data.db[index];
					index++;
				}
			}
		}
		//set the Db threshold
		mineigenvalue = matDb->data.db[0] / CV_LDA_MAX_RATIO;
		maxfeatures1 = vecsize;
		for (i=0; i<matDb->rows; i++)
		{
			if ( matDb->data.db[i*matDb->rows+i] < mineigenvalue ) 
			{
				maxfeatures1 = i; 
				break;
			}
		}
		// allocate buffer
		CV_CALL( ztranspose = cvCreateMat(maxfeatures1, vecsize, type) );
		CV_CALL( zbuf = cvCreateMat(vecsize, maxfeatures1, type) );
		CV_CALL( tempmat = cvCreateMat(maxfeatures1, vecsize, type) );
		CV_CALL( ztsbz = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( matDw = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( utranspose = cvCreateMat(maxfeatures1, maxfeatures1, type) );
		CV_CALL( mata = cvCreateMat(maxfeatures1, vecsize, type) );	// n*n

		// Z(t)=(D^(-1/2)) * Y(t)
		index = 0;
		for(i=0; i<ztranspose->rows; i++) 
		{
			tmp = sqrt(matDb->data.db[i*matDb->cols+i]);		
			for (j=0; j<ztranspose->cols; j++) 
			{
				ztranspose->data.db[index] = ytranspose->data.db[index] / tmp;
				index ++;
			}
		}
		// Y(t) * Sw
		CV_CALL( cvMatMulAdd(ztranspose, withinClassScatter, 0, tempmat) ); 
		// compute Z
		CV_CALL( cvTranspose(ztranspose, zbuf) ); // zbuf = Z
		// Z(t) * Sw * Z
		CV_CALL( cvMatMulAdd(tempmat, zbuf, 0, ztsbz) ); 
		//singular value decomposition Z(t)*Sb*Z
		CV_CALL( cvSVD( ztsbz, matDw, utranspose, 0, CV_SVD_U_T) ); //yt=U(t)
		// set eigenvectors matrix to positive
		CV_CALL( sum = cvmTrace(utranspose) );
		if (sum < 0.0) 
		{
			index = 0;
			for (i=0; i<utranspose->rows; i++) 
			{
				for (j=0; j<utranspose->cols; j++) 
				{
					utranspose->data.db[index] = - utranspose->data.db[index];
					index ++;
				}
			}
		}		
		
		//set the Dw threshold
		mineigenvalue = matDw->data.db[0] / CV_LDA_MAX_RATIO;
		maxfeatures2 = vecsize;
		for (i=0; i<matDw->rows; i++)
		{
			if ( matDw->data.db[i*matDw->cols+i] < mineigenvalue ) 
			{
				maxfeatures2 = i; 
				break;
			}
		}
		
		// make transform matrix A=U(t)*Z(t)
		CV_CALL( cvMatMulAdd(utranspose, ztranspose, 0, mata) );	// TranMat = U(t)*Z(t)
		
		// sorted projMatrix according to Dw of Z(t)*Sw*Z 
		for (i=0; i<matDw->rows - 1; i++)
		{
			for (j=0; j<matDw->cols - 1; j++)
			{
				if (matDw->data.db[(j+1)*matDw->cols+j+1] > matDw->data.db[j*matDw->cols+j]) 
				{
					tmp = matDw->data.db[j*matDw->cols+j];
					matDw->data.db[j*matDw->cols+j] = matDw->data.db[(j+1)*matDw->cols+j+1];
					matDw->data.db[(j+1)*matDw->cols+j+1] = tmp;
					for (k=0; k<mata->cols; k++) 
					{
						tmp = mata->data.db[j*mata->cols+k];
						mata->data.db[j*mata->cols+k] = mata->data.db[(j+1)*mata->cols+k];
						mata->data.db[(j+1)*mata->cols+k] = tmp;
					}
				} 
			} 
		}// end for sorting
		
		rowcopy = mata->rows < projMatrix->rows ? mata->rows : projMatrix->rows;
		index = 0;
		for(i=0; i<rowcopy; i++) 
		{
			eigenvalues->data.db[i] = matDw->data.db[i*matDw->cols+i];
			for(j=0;j<vecsize;j++) 
			{
				projMatrix->data.db[index] = mata->data.db[index]; 
				index ++;
			}
		}		
		// set values and eigenvetors beyond maxfeatures2 to zero
		for(i=rowcopy; i<projMatrix->rows; i++) 
		{
			eigenvalues->data.db[i] = 0.0;
			for(j=0;j<vecsize;j++) 
			{
				projMatrix->data.db[index] = 0.0;
				index ++;
			}
		}
		// free temp buffer
		CV_CALL( cvReleaseMat(&ytranspose) );
		CV_CALL( cvReleaseMat(&matDb) );
		CV_CALL( cvReleaseMat(&ztranspose) );
		CV_CALL( cvReleaseMat(&zbuf) );
		CV_CALL( cvReleaseMat(&tempmat) );
		CV_CALL( cvReleaseMat(&ztsbz) );
		CV_CALL( cvReleaseMat(&matDw) );
		CV_CALL( cvReleaseMat(&utranspose) );
		CV_CALL( cvReleaseMat(&mata) );
	} // end if type
	
	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			cvNormLDAMatrix( const CvMat* eigenvalues, CvMat* projMatrix ); 
//    Purpose:
//			Divides the vectors in the rows of LDA projection matrix by the square roots of the corresponding eigenvalues
//    Context:
//    Parameters:
//			eigenvalues	--- [I] the eigenvalues obtained using function cvCalcLDAMatrix. 
//			projMatrix		--- [I/O] the LDA projection matrix obtained by function cvCalcLDAMatrix. 
//								  After the function call projMatrix is normalized.
//    Returns: 
//			return Cv_StsOK if successfully detect a face, else report error;
//F*/
CV_IMPL  CVStatus 
cvNormLDAMatrix( const CvMat* eigenvalues, CvMat* projMatrix )
{
	int i, j, index;
	int vecsize, vecnum;	// square matrix size n
	int type;	
	double tmp;				// temp variables 
	
	CV_FUNCNAME( "cvCalcLdaUnitTransMat" );

    __BEGIN__;

	if( !CV_IS_MAT(eigenvalues) || !CV_IS_MAT(projMatrix) )
		CV_ERROR(CV_StsBadArg, "Invalid input argument!");
		
	if( !CV_ARE_TYPES_EQ(eigenvalues, projMatrix) )
		CV_ERROR(CV_StsBadArg, "All input/ouput matrices should have the same type!");
		
	vecsize = projMatrix->cols;
	vecnum = projMatrix->rows;
	if (eigenvalues->cols * eigenvalues->rows != vecnum )
		CV_ERROR(CV_StsBadArg, "The size of input/ouput matrices do not match!");

	CV_CALL( type = cvGetElemType(projMatrix) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR(CV_StsUnmatchedFormats, "Input data type must be float or double!");	

	if( type == CV_32F )
	{
		// normalize the projMatrix		
		for (i=0; i<vecnum; i++) 
		{
			if ( eigenvalues->data.fl[i] <= 0.0f) continue;
			tmp = 1.0f / (float)sqrt(eigenvalues->data.fl[i]);
			index = i * vecsize;
			for(j=0; j<vecsize; j++)
			{
				projMatrix->data.fl[index] *= (float)tmp;
				index ++;
			}
		} 
	}
	else	// type == CV_64F
	{
		// normalize the projMatrix		
		for (i=0; i<vecnum; i++) 
		{
			if ( eigenvalues->data.db[i] <= 0.0 ) continue;
			tmp = 1.0 / sqrt(eigenvalues->data.db[i]);
			index = i * vecsize;
			for(j=0; j<vecsize; j++)
			{
				projMatrix->data.db[index] *= tmp;
				index ++;
			}
		} 
	} // end if type
	
	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvLDAProjection( const CvMat* projMatrix, const CvMat* inVectors, CvMat* outVectors );
//    Purpose:
//			The function cvLDAProjection transforms the input data into the transformed space 
//			obtained by LDA. Suppose that the input vectors space is n-dimensional, the LDA space is
//			m-dimension and the number of input vectors is p, then the matrices projMatrix, inVectors 
//			and outVectors are of size mn, pn and pm respectively.
//    Context:
//    Parameters:
//		projMatrix	-- [I] the LDA projection matrix obtained by function CalcEigenvectorsLDA or cvNormLDAMatrix. 
//		inVectors	-- [I] the input vectors stored as the rows of inVectors.
//		outVectors	-- [O] the transformed vectors stored as the rows of outVectors. 
//    Returns: 
//		return Cv_StsOK if successfully detect a face, else report error;
//	  Notes:
//F*/
CV_IMPL CVStatus
cvLDAProjection( const CvMat* projMatrix, const CvMat* inVectors, CvMat* outVectors )
{
	int i, j, k;
	int srcvecsize;			// inVectors vector length
	int dstvecsize;			// outVectors vector length
	int fmnum;				// # of frame
	int transtype, feattype;// element type
	float tmpfl, *pflsrc, *pfldst, *pfl;
	double tmpdb, *pdbsrc, *pdbdst, *pdb;

	CV_FUNCNAME( "cvLdaTransform1" );

    __BEGIN__;
	
	// check arguments
	if( !CV_IS_MAT(projMatrix) || !CV_IS_MAT(inVectors) || !CV_IS_MAT(outVectors) )
	{
		CV_ERROR( CV_StsBadArg, "Invalid input matrices!" );
	}

	if ( !CV_ARE_TYPES_EQ( inVectors, outVectors ) )
	{
		CV_ERROR( CV_StsUnmatchedFormats, "Input and output feature matrices must have the same type." );
	}

	CV_CALL( transtype = cvGetElemType(projMatrix) );
	CV_CALL( feattype = cvGetElemType(inVectors) );

	srcvecsize = projMatrix->cols;
	dstvecsize = projMatrix->rows;
	fmnum = inVectors->rows;

	if ( srcvecsize < 1 || fmnum < 1 || dstvecsize < 1)
	{
		CV_ERROR( CV_StsBadSize, "Invalid input matrics!" );
	}
	
	if( inVectors->cols != srcvecsize || outVectors->cols != dstvecsize || inVectors->rows != outVectors->rows)
	{
		CV_ERROR( CV_StsBadArg, "Input and output feature matrices do not match!" );
	}
	
	if (transtype == CV_32F)
	{
		if (feattype == CV_32F)
		{
			for (i=0; i<fmnum; i++)
			{
				pflsrc = &(inVectors->data.fl[srcvecsize*i]);
				pfldst = &(outVectors->data.fl[dstvecsize*i]);
				// multiply matrix
				for (j=0; j<dstvecsize; j++)
				{
					tmpfl = 0;
					pfl = &(projMatrix->data.fl[j*srcvecsize]);
					for (k=0; k<srcvecsize; k++) 
					{
						tmpfl += pfl[k] * pflsrc[k];
					}
					pfldst[j] = tmpfl;
				}
			}
		}
		else if (feattype == CV_64F)
		{
			for (i=0; i<fmnum; i++)
			{
				pdbsrc = &(inVectors->data.db[srcvecsize*i]);
				pdbdst = &(outVectors->data.db[dstvecsize*i]);
				// multiply matrix
				for (j=0; j<dstvecsize; j++)
				{
					tmpdb = 0;
					pfl = &(projMatrix->data.fl[j*srcvecsize]);
					for (k=0; k<srcvecsize; k++) 
					{
						tmpdb += pfl[k] * pdbsrc[k];
					}
					pdbdst[j] = tmpdb;
				}
			}
		}
		else
		{
			CV_ERROR( CV_StsBadArg, "Invalid data type of input and output feature matrices!" );
		}
	}
	else if (transtype == CV_64F)
	{
		if (feattype == CV_32F)
		{
			for (i=0; i<fmnum; i++)
			{
				pflsrc = &(inVectors->data.fl[srcvecsize*i]);
				pfldst = &(outVectors->data.fl[dstvecsize*i]);
				// multiply matrix
				for (j=0; j<dstvecsize; j++)
				{
					tmpdb = 0;
					pdb = &(projMatrix->data.db[j*srcvecsize]);
					for (k=0; k<srcvecsize; k++) 
					{
						tmpdb += pdb[k] * pflsrc[k];
					}
					pfldst[j] = (float)tmpdb;
				}
			}
		}
		else if (feattype == CV_64F)
		{
			for (i=0; i<fmnum; i++)
			{
				pdbsrc = &(inVectors->data.db[srcvecsize*i]);
				pdbdst = &(outVectors->data.db[dstvecsize*i]);
				// multiply matrix
				for (j=0; j<dstvecsize; j++)
				{
					tmpdb = 0;
					pdb = &(projMatrix->data.db[j*srcvecsize]);
					for (k=0; k<srcvecsize; k++) 
					{
						tmpdb += pdb[k] * pdbsrc[k];
					}
					pdbdst[j] = tmpdb;
				}
			}
		}
		else
		{
			CV_ERROR( CV_StsBadArg, "Invalid data type of input and output feature matrices!" );
		}
	}
	else 
	{
		CV_ERROR( CV_StsBadArg, "Invalid data type of transform matrix!" );
	}
	
	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}
