/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "cvmouthtracking.h"

typedef int sumtype;
typedef double sqsumtype;


typedef struct CvHidHaarFeature
{
    struct
    {
        sumtype *p0, *p1, *p2, *p3;
        float weight;
    }
    rect[CV_HAAR_FEATURE_MAX];
}
CvHidHaarFeature;


typedef struct CvHidHaarClassifier
{
    int count;
    CvHaarFeature* origFeature;
    CvHidHaarFeature* feature;
    float* threshold;
    int* left;
    int* right;
    float* alpha;
}
CvHidHaarClassifier;


typedef struct CvHidHaarStageClassifier
{
    int  count;
    float threshold;
    CvHidHaarClassifier* classifier;
}
CvHidHaarStageClassifier;


struct CvHidHaarClassifierCascade
{
    int  headerSize;
    int  count;
    int  is_stump_based;
    int  has_tilted_features;
    CvSize origWindowSize;
    CvSize realWindowSize;
    double scale, invWindowArea;
    CvHidHaarStageClassifier* stageClassifier;
    CvMat sum;
    CvMat tiltedSum;
    CvMat sqsum;
    sqsumtype *pq0, *pq1, *pq2, *pq3;
    sumtype *p0, *p1, *p2, *p3;
};

static int is_equal( const void* _r1, const void* _r2, void* )
{
    const CvRect* r1 = (const CvRect*)_r1;
    const CvRect* r2 = (const CvRect*)_r2;
    int distance = cvRound(r1->width*0.2);

    return r2->x <= r1->x + distance &&
           r2->x >= r1->x - distance &&
           r2->y <= r1->y + distance &&
           r2->y >= r1->y - distance &&
           r2->width <= cvRound( r1->width * 1.2 ) &&
           cvRound( r2->width * 1.2 ) >= r1->width;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvHaarDetectObjectsR( const IplImage* img,
//			                      CvHidHaarClassifierCascade* cascade,
//								  CvMemStorage* storage, double scale_factor,
//								  int min_neighbors, int flags,
//								  int min_width, int max_width)
//    Purpose:
//          The function cvHaarDetectObjectsR finds rectangular regions in the given image 
//			that are likely to contain objects the cascade has been trained for and returns 
//			those regions as a sequence of rectangles. CvHaarDetectObjectsR has the same 
//			functionality as cvHaarDetectObjects from OpenCV except that it restricts the width 
//			of the detected object to the range [min_width, max_width].
//    Context:
//    Parameters:
//			img		--- [I] Input Ipl image.
//			cascade --- [I] Haar classifier cascade in internal representation. 
//			storage --- [I] Memory storage to store the resultant sequence of the object 
//							candidate rectangles. 
//			scale_factor -- [I] The factor by which the search window is scaled between 
//								the subsequent scans, for example, 1.1 means increasing window 
//								by 10%. 
//			min_neighbors - [I] Minimum number (minus 1) of neighbor rectangles that makes up 
//								an object. All the groups of a smaller number of rectangles than 
//								min_neighbors-1 are rejected. If min_neighbors is 0, the function 
//								does not any grouping at all and returns all the detected candidate 
//								rectangles, which may be useful if the user wants to apply a 
//								customized grouping procedure
//			flags	--- [I] Mode of operation. Currently the only flag that may be specified is 
//							CV_HAAR_DO_CANNY_PRUNING. If it is set, the function uses Canny edge 
//							detector to reject some image regions that contain too few or too much 
//							edges and thus can not contain the searched object. The particular 
//							threshold values are tuned for face detection and in this case the pruning 
//							speeds up the processing.  
//			min_width --- [I] Minimum width of the object to detect. 
//			max_width --- [I] Maximum width of the object to detect. 
//    Returns:
//			Return the list of detected objects
//    Notes:
//F*/
CV_IMPL CvSeq*
cvHaarDetectObjectsR( const IplImage* img,
                      CvHidHaarClassifierCascade* cascade,
					  CvMemStorage* storage, double scale_factor,
					  int min_neighbors, int flags,
					  int min_width, int max_width)
{
    int split_stage = 2;
    CvSize size = cvGetSize( img ), sumsize = { size.width + 1, size.height + 1 };
    IplImage *temp = 0, *sum = 0, *tilted = 0, *sqsum = 0;
    IplImage *sumcanny = 0;
    CvSeq* seq = 0;
    CvSeq* seq2 = 0;
    CvSeq* idx_seq = 0;
    CvSeq* result_seq = 0;
    CvMemStorage* temp_storage = 0;
    CvAvgComp* comps = 0;
    int ncomp;
    int do_canny_pruning = flags & CV_HAAR_DO_CANNY_PRUNING;
	double maxFactor, minFactor;
    
    CV_FUNCNAME( "cvHaarDetectObjectsR" );

    __BEGIN__;

    double factor;
    int i, npass = 2;
    int hid_cascade_count0;

    if( !CV_IS_IMAGE(img) )
        CV_ERROR( CV_StsBadArg, "Invalid source image!" );

    if( !cascade || cascade->headerSize != sizeof(*cascade))
        CV_ERROR( CV_StsBadArg, "Invalid classifier cascade!" );

    if( !storage )
        CV_ERROR( CV_StsNullPtr, "Invalid storage pointer!" );

    temp = cvCreateImage( size, 8, 1 );
    sum = cvCreateImage( sumsize, IPL_DEPTH_32S, 1 );
    sqsum = cvCreateImage( sumsize, IPL_DEPTH_64F, 1 );
    temp_storage = cvCreateChildMemStorage( storage );

    if( cascade->has_tilted_features )
        tilted = cvCreateImage( sumsize, IPL_DEPTH_32S, 1 );

    seq = cvCreateSeq( 0, sizeof(CvSeq), sizeof(CvRect), temp_storage );
    seq2 = cvCreateSeq( 0, sizeof(CvSeq), sizeof(CvAvgComp), temp_storage );
    result_seq = cvCreateSeq( 0, sizeof(CvSeq), sizeof(CvAvgComp), storage );

    if( min_neighbors == 0 )
        seq = result_seq;

    if( img->nChannels == 3 )
    {
        cvCvtColor( img, temp, CV_BGR2GRAY );
        img = temp;
    }
    
    cvIntegral( img, sum, sqsum, tilted );
    
    if( do_canny_pruning )
    {
        sumcanny = cvCreateImage( sumsize, IPL_DEPTH_32S, 1 );
        cvCanny( img, temp, 0, 50, 3 );
        cvIntegral( temp, sumcanny );
    }
    
    hid_cascade_count0 = cascade->count;

    if( split_stage >= hid_cascade_count0 || split_stage == 0 )
    {
        split_stage = hid_cascade_count0;
        npass = 1;
    }

	// constraint in scale here
	minFactor = (double)min_width / (double)(cascade->origWindowSize.width);
	minFactor = minFactor < 1.0 ? 1.0 : minFactor;
	maxFactor = (double)max_width / (double)(cascade->origWindowSize.width);
    for( factor = minFactor; (factor*cascade->origWindowSize.width < size.width - 4) && 
		 (factor*cascade->origWindowSize.height < size.height - 4) && (factor <= maxFactor); factor *= scale_factor )
	{
        double const ystep = MAX( 2, factor );
        
        CvSize winSize = cvSize( cvRound( cascade->origWindowSize.width * factor ),
                                 cvRound( cascade->origWindowSize.height * factor ));
        CvRect equ_rect = { 0, 0, 0, 0 };
        int *p0 = 0, *p1 = 0, *p2 = 0, *p3 = 0;
        int *pq0 = 0, *pq1 = 0, *pq2 = 0, *pq3 = 0;
        int pass, stage_offset = 0;
        int stopHeight = cvRound((size.height - winSize.height - ystep) / ystep);

        cvSetImagesForHaarClassifierCascade( cascade, sum, sqsum, tilted, factor );
        cvZero( temp );

        if( do_canny_pruning )
        {
            equ_rect.x = cvRound(winSize.width*0.3);
            equ_rect.y = cvRound(winSize.height*0.3);
            equ_rect.width = cvRound(winSize.width*0.7);
            equ_rect.height = cvRound(winSize.height*0.7);

            p0 = (int*)(sumcanny->imageData + equ_rect.y*sumcanny->widthStep) + equ_rect.x;
            p1 = (int*)(sumcanny->imageData + equ_rect.y*sumcanny->widthStep)
                        + equ_rect.x + equ_rect.width;
            p2 = (int*)(sumcanny->imageData + (equ_rect.y + equ_rect.height)*sumcanny->widthStep) + equ_rect.x;
            p3 = (int*)(sumcanny->imageData + (equ_rect.y + equ_rect.height)*sumcanny->widthStep)
                        + equ_rect.x + equ_rect.width;

            pq0 = (int*)(sum->imageData + equ_rect.y*sum->widthStep) + equ_rect.x;
            pq1 = (int*)(sum->imageData + equ_rect.y*sum->widthStep)
                        + equ_rect.x + equ_rect.width;
            pq2 = (int*)(sum->imageData + (equ_rect.y + equ_rect.height)*sum->widthStep) + equ_rect.x;
            pq3 = (int*)(sum->imageData + (equ_rect.y + equ_rect.height)*sum->widthStep)
                        + equ_rect.x + equ_rect.width;
        }

        cascade->count = split_stage;

        for( pass = 0; pass < npass; pass++ )
        {
#ifdef _OPENMP
    #pragma omp parallel for shared(cascade, stopHeight, seq, ystep, temp, size, winSize, pass, npass, sum, p0, p1, p2 ,p3, pq0, pq1, pq2, pq3, stage_offset) 
#endif // _OPENMP

            for( int yInt = 0; yInt < stopHeight; yInt++ )
            {
                int iy = cvRound(yInt*ystep);
                int xIntStep = 1;
                int xInt, stopWidth = cvRound((size.width - winSize.width - 2*ystep) / ystep);

                char* mask_row = temp->imageData + temp->widthStep * iy;

                for( xInt = 0; xInt < stopWidth; xInt += xIntStep )
                {
                    int ix = cvRound(xInt*ystep);
                    
                    if( pass == 0 )
                    {
                        int result;

                        xIntStep = 2;

                        if( do_canny_pruning )
                        {
                            int offset;
                            int s, sq;
                        
                            offset = iy*(sum->widthStep/sizeof(p0[0])) + ix;
                            s = p0[offset] - p1[offset] - p2[offset] + p3[offset];
                            sq = pq0[offset] - pq1[offset] - pq2[offset] + pq3[offset];
                            if( s < 100 || sq < 20 )
                                continue;
                        }

                        result = cvRunHaarClassifierCascade( cascade, cvPoint(ix,iy),
                                                             0/*stage_offset*/);
#ifdef _OPENMP
#pragma omp critical
#endif
                        if( result > 0 )
                        {
                            if( pass < npass - 1 )
                            {
                                mask_row[ix] = 1;
                            }
                            else
                            {
                                CvRect rect = cvRect(ix,iy,winSize.width,winSize.height);
                                cvSeqPush( seq, &rect );
                            }
                        }
                        if( result < 0 )
                            xIntStep = 1;
                    }
                    else if( mask_row[ix] )
                    {
                        int result = cvRunHaarClassifierCascade( cascade, cvPoint(ix,iy),
                                                                 stage_offset );
#ifdef _OPENMP
#pragma omp critical
#endif
                        if( result > 0 )
                        {
                            if( pass == npass - 1 )
                            {
                                CvRect rect = cvRect(ix,iy,winSize.width,winSize.height);
                                cvSeqPush( seq, &rect );
                            }
                        }
                        else
                            mask_row[ix] = 0;
                    }
                }
            }
            stage_offset = cascade->count;
            cascade->count = hid_cascade_count0;
        }
    }

    if( min_neighbors != 0 )
    {
        // group retrieved rectangles in order to filter out noise 
        ncomp = cvPartitionSeq( seq, 0, &idx_seq, is_equal, 0, 0 );
        CV_CALL( comps = (CvAvgComp*)cvAlloc( (ncomp+1)*sizeof(comps[0])));
        memset( comps, 0, (ncomp+1)*sizeof(comps[0]));

        // count number of neighbors
        for( i = 0; i < seq->total; i++ )
        {
            CvRect r1 = *(CvRect*)cvGetSeqElem( seq, i );
            int idx = *(int*)cvGetSeqElem( idx_seq, i );
            assert( (unsigned)idx < (unsigned)ncomp );

            comps[idx].neighbors++;
             
            comps[idx].rect.x += r1.x;
            comps[idx].rect.y += r1.y;
            comps[idx].rect.width += r1.width;
            comps[idx].rect.height += r1.height;
        }

        // calculate average bounding box
        for( i = 0; i < ncomp; i++ )
        {
            int n = comps[i].neighbors;
            if( n >= min_neighbors )
            {
                CvAvgComp comp;
                comp.rect.x = (comps[i].rect.x*2 + n)/(2*n);
                comp.rect.y = (comps[i].rect.y*2 + n)/(2*n);
                comp.rect.width = (comps[i].rect.width*2 + n)/(2*n);
                comp.rect.height = (comps[i].rect.height*2 + n)/(2*n);
                comp.neighbors = comps[i].neighbors;

                cvSeqPush( seq2, &comp );
            }
        }

        // filter out small face rectangles inside large face rectangles
        for( i = 0; i < seq2->total; i++ )
        {
            CvAvgComp r1 = *(CvAvgComp*)cvGetSeqElem( seq2, i );
            int j, flag = 1;

            for( j = 0; j < seq2->total; j++ )
            {
                CvAvgComp r2 = *(CvAvgComp*)cvGetSeqElem( seq2, j );
                int distance = cvRound( r2.rect.width * 0.2 );
            
                if( i != j &&
                    r1.rect.x >= r2.rect.x - distance &&
                    r1.rect.y >= r2.rect.y - distance &&
                    r1.rect.x + r1.rect.width <= r2.rect.x + r2.rect.width + distance &&
                    r1.rect.y + r1.rect.height <= r2.rect.y + r2.rect.height + distance &&
                    (r2.neighbors > MAX( 3, r1.neighbors ) || r1.neighbors < 3) )
                {
                    flag = 0;
                    break;
                }
            }

            if( flag )
            {
                cvSeqPush( result_seq, &r1 );
                /* cvSeqPush( result_seq, &r1.rect ); */
            }
        }
    }

    __END__;

    cvReleaseMemStorage( &temp_storage );
    cvReleaseImage( &sum );
    cvReleaseImage( &sqsum );
    cvReleaseImage( &tilted );
    cvReleaseImage( &temp );
    cvReleaseImage( &sumcanny );

    cvFree( (void**)&comps );

    return result_seq;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvInitTrackMouth(CvHidHaarClassifierCascade *hidCascadeFace, 
//							 CvHidHaarClassifierCascade	*hidCascadeMouth0, 
//							 CvHidHaarClassifierCascade	*hidCascadeMouth1,
//							 CvMouthTracker* MouthTracker)
//    Purpose:
//          The function cvInitTrackMouth initialize the data structure MouthTracker 
//			that contains the pointers of classifiers, tracking status and records used 
//			in mouth detection and tracking. The input arguments hidCascadeFace, 
//			hidCascadeMouth0 and hidCascadeMouth1 must be initialized by function 
//			cvLoadHaarClassifierCascade and function cvCreateHidHaarClassifierCascade in OpenCV 
//			before calling this function. The structure MouthTracker must allocated 
//			before using this function.
//    Context:
//    Parameters:
//			hidCascadeFace	 --- [I] Input the pointer of boosted cascaded classifier for face detection
//			hidCascadeMouth0 --- [I] Input the pointer of boosted cascaded classifier for mouth without beard.
//			hidCascadeMouth1 --- [I] Input the pointer of boosted cascaded classifier for mouth with beard.
//			MouthTracker	 --- [O] Output the data structure that contains the pointers of 
//									 mouth tracking classifiers, status and records. 
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus
cvInitTrackMouth(CvHidHaarClassifierCascade *hidCascadeFace, 
				 CvHidHaarClassifierCascade	*hidCascadeMouth0, 
				 CvHidHaarClassifierCascade	*hidCascadeMouth1,
				 CvMouthTracker* MouthTracker)
{

	CV_FUNCNAME( "cvInitTrackMouth" );

	__BEGIN__;
	
	if (hidCascadeFace == NULL) 
		CV_ERROR(CV_StsNullPtr, "Invalid pointer of face detector!");
	
	if(hidCascadeMouth0 == NULL || hidCascadeMouth1 == NULL)
		CV_ERROR(CV_StsNullPtr, "Invalid pointer of mouth detector!");

	// initial variables
	MouthTracker->DownSmpRate = 0.25f;
	MouthTracker->iplDownSmpImage = NULL;
	MouthTracker->iskalmanCreated = FALSE;	
	MouthTracker->RltPoolLen = 0;
	MouthTracker->RltPoolHeader = 0;
	MouthTracker->isRltPoolFull = FALSE;	
	MouthTracker->KalmanXY = NULL;
	MouthTracker->MeasureVectXY = NULL;

	// initialize the face detector
	MouthTracker->hidCascadeFace = hidCascadeFace;
	
	// initialize the mouth detector
	MouthTracker->hidCascadeMouth0 = hidCascadeMouth0;
	MouthTracker->hidCascadeMouth1 = hidCascadeMouth1;
	
	// allocate buffer 0
	CV_CALL( MouthTracker->StorageDual0 = cvCreateMemStorage() );
	if (MouthTracker->StorageDual0 == NULL)
		CV_ERROR(CV_StsNullPtr, "Fail to allocate buffer 0 for storage!");

	// allocate buffer 1
	CV_CALL( MouthTracker->StorageDual1 = cvCreateMemStorage() );
	if (MouthTracker->StorageDual1 == NULL)
		CV_ERROR(CV_StsNullPtr, "Fail to allocate buffer 1 for storage!");

	MouthTracker->MouthWndWidth = 32;

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvResetTrackMouth(const int imagewidth, const int imageheight, CvMouthTracker* MouthTracker)
//    Purpose:
//			The function cvResetTrackMouth updates the data structure MouthTracker according to 
//			the image size of the video sequence, and resets the Kalman filter and the finite state machine 
//			for mouth tracking. The data structure MouthTracker must be initialized by function 
//			cvInitTrackMouth before calling this function.
//    Context:
//    Parameters:
//			imagewidth	---- [I] Input image width of the video sequence to process.
//			imageheight ---- [I] Input image height of the video sequence to process.
//			MouthTracker --- [I/O] Input the data structure initialized by function cvInitMouthTracker. 
//								   Output the updated data structure
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvResetTrackMouth(const int imagewidth, const int imageheight, CvMouthTracker* MouthTracker)
{
	const int DampFrameNum = 8;

	CV_FUNCNAME( "cvResetTrackMouth" );

	__BEGIN__;
	
	// Create down sampled image	
	if (MouthTracker->iplDownSmpImage != NULL) 
	{
		cvReleaseImage(&MouthTracker->iplDownSmpImage);
		MouthTracker->iplDownSmpImage = NULL;
	}
	CV_CALL( MouthTracker->iplDownSmpImage = cvCreateImage(cvSize(cvRound(imagewidth * MouthTracker->DownSmpRate), cvRound(imageheight * MouthTracker->DownSmpRate)), IPL_DEPTH_8U, 1) );
	if (MouthTracker->iplDownSmpImage == NULL) 
		CV_ERROR(CV_StsNullPtr, "Fail to create downsampled image!");

	// for tracking state
	MouthTracker->TrackState = FACEMOUTHDETECTION;		// tracking state
	MouthTracker->InitialCounter = 0;	// counter for initialize stage
	MouthTracker->ErrorCounter = 0;
	
	CV_CALL( icvResultPoolInit(DampFrameNum, MouthTracker) );

	// for tracking options and result
	if (MouthTracker->iskalmanCreated) 
		CV_CALL( icvReleaseKalmanXY(MouthTracker) );
	
	CV_CALL( icvCreateKalmanXY(3, 0.04f, 0.04f, 1.0f, 1.0f, MouthTracker) );
	MouthTracker->iskalmanCreated = TRUE;

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvUpdateTrackMouth(IplImage * img, CvMouthTracker* MouthTracker, CvTrackResult & result); 
//    Purpose:
//          The function cvUpdateTrackMouth detects or tracks the mouth region in one give image, 
//			and updates the data structure CvTrackResult according to the detection/tracking result 
//			and the previous status of the finite state machine.
//    Context:
//    Parameters:
//			img			---- [I] Input 8-bit grayscale or 24-bit true color image.
//			MouthTracker --- [I/O] Input the data structure CvMouthTracker that contains the pointers 
//									of classifiers, tracking status and records. Output the updated one.
//			result		---- [O] Output the data structure CvTrackResult that contains the mouth tracking result. 
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvUpdateTrackMouth (IplImage * img, CvMouthTracker* MouthTracker, CvTrackResult & result)
{
	CV_FUNCNAME( "cvUpdateTrackMouth" );

	__BEGIN__;

	const int PrErrMin = 16;
	const float PrErrRate = 0.15f;
	const int MaxMouthErrorNum = 3;
	const int InitialFrameNum = 5;
	const int MaxContErrInTrack = 1;
	const float ResampleInterval = 1.1f;
	const float MouthSizeDev = 1.31f;

	static CvMouthLocateResult mouth;
	static CvMouthLocateResult oldMouth;
	static CvMouthLocateResult aveMouth;		
	static int	ErrFrameInTrack;	
	
	IplImage *iplRoiGrayImage = NULL;
	CvRect	searchRoi;
	CvRect	maxface;		// detected maxsize face rectangle
	CvRect	mouth_rect;
	bool isMouthDetected;
	int	MouthMin, MouthMax;	
	int	facenum;
	int  MouthWidth;
	float EstX, EstY;
	int PrErrX, PrErrY;
	
	// ------------------------------------------------------------------------
	// Haar cascade classifier based mouth tracking
	// ------------------------------------------------------------------------
	// get image
	if (img == NULL) 
		CV_ERROR(CV_StsBadArg, "Invalid input image!");
	
	// detection	
	switch (MouthTracker->TrackState) 
	{
		case FACEMOUTHDETECTION:
			if(img->nChannels == 3)
			{
				// convert color image to gray scale image
				CV_CALL( iplRoiGrayImage = cvCreateImage(cvSize(img->width, img->height), IPL_DEPTH_8U, 1) );
				if (iplRoiGrayImage == NULL) 
					CV_ERROR(CV_StsBadFunc, "Fail to allocate region image!");			
				CV_CALL( cvCvtColor(img, iplRoiGrayImage, CV_BGR2GRAY ) );

				// down sample
				CV_CALL( cvResize(iplRoiGrayImage, MouthTracker->iplDownSmpImage) );
			}else
			{
				// down sample for gray scale image
				CV_CALL( cvResize(img, MouthTracker->iplDownSmpImage) );
			}
		
			// face detection
			CV_CALL( facenum = icvHaarDetectMaxFace(MouthTracker->iplDownSmpImage, MouthTracker->hidCascadeFace, maxface) ); 
			if (facenum <= 0) 
			{	// face detection error				
				result.TrackRltSts = FACEDETECTERROR;
				cvReleaseImage(&iplRoiGrayImage);
				return CV_StsOk;
			}
			
			// modify face size
			maxface.x = cvRound(maxface.x / MouthTracker->DownSmpRate);
			maxface.width = cvRound(maxface.width / MouthTracker->DownSmpRate);
			maxface.y = cvRound(maxface.y / MouthTracker->DownSmpRate);
			maxface.height = cvRound(maxface.height / MouthTracker->DownSmpRate);

			// set mouth region
			mouth.face_left = maxface.x;
			mouth.face_right = maxface.x + maxface.width;
			mouth.face_top = maxface.y;
			mouth.face_bottom = maxface.y + maxface.height;
							
			// predict search region ROI
			CV_CALL( icvHaarEstimateMouthRegion(maxface, searchRoi, MouthWidth) );	

			MouthMin = (int)(MouthWidth / MouthSizeDev);
			MouthMax = (int)(MouthWidth * MouthSizeDev);
			
			// mouth detect	
			CV_CALL( cvSetImageROI( iplRoiGrayImage, searchRoi) );
			CV_CALL( icvHaarDetectMouthDualCascades( iplRoiGrayImage, MouthTracker, ResampleInterval, MouthMin, MouthMax, mouth_rect, isMouthDetected) );
			cvReleaseImage(&iplRoiGrayImage);
					
			// state transfer
			if (isMouthDetected) 
			{
				mouth.Likeness = 1.0f;
				mouth.RotAngle = 0.0f;
				mouth.CenterX = (float)(mouth_rect.x + mouth_rect.width / 2);
				mouth.CenterY = (float)(mouth_rect.y + mouth_rect.height / 2);
				mouth.Width = (float)mouth_rect.width;
				mouth.Height = (float)mouth_rect.height;
				
				// record track in initial stack
				CV_CALL( icvResultPoolPush(mouth, MouthTracker) );
				MouthTracker->InitialCounter ++;
				if (MouthTracker->InitialCounter >= InitialFrameNum) 
				{
					// success in continuous 5 frames, transfer state to tracking
					MouthTracker->TrackState = MOUTHTRACKING;
					// intitial kalman filter
					CV_CALL( icvResultPoolAverage(MouthTracker,aveMouth) );					
					CV_CALL( icvInitKalmanXY(aveMouth.CenterX, aveMouth.CenterY, 0, 0, MouthTracker) );	
				}
				// tracking successful, save result
				result.MouthResult = mouth;
				result.TrackRltSts = TRACKSUCCESSFUL;
				oldMouth = mouth;
			} 
			else 
			{
				// reset counter
				CV_CALL( icvResultPoolClear(MouthTracker) );
				// set track result status				
				result.TrackRltSts = MOUTHDETECTERROR;
			} // end if (isMouthDetected)
			break;

		case MOUTHTRACKING:
			// use search region calculated according to the tracking result			
			// time update Kalman filter
			CV_CALL( cvKalmanUpdateByTime(MouthTracker->KalmanXY) );
			// get estimated obs (omit multiple by H)
			// EstX, EstY ---- estimated new state (converted to measurement)
			EstX = MouthTracker->KalmanXY->PriorState[0];
			EstY = MouthTracker->KalmanXY->PriorState[MouthTracker->KinematicOrder];
			PrErrX = cvRound(aveMouth.Width * PrErrRate);
			PrErrX = PrErrX < PrErrMin ? PrErrMin : PrErrX;
			PrErrY = PrErrX;
			// set face region NULL
			mouth.face_left = 0;
			mouth.face_right = 0;
			mouth.face_top = 0;
			mouth.face_bottom = 0;			
			// set predicted search roi
			searchRoi.x = cvRound(EstX - aveMouth.Width / 2) - PrErrX;
			searchRoi.width = (int)(aveMouth.Width + PrErrX * 2);
			searchRoi.y = cvRound(EstY - aveMouth.Height / 2) - PrErrY;
			searchRoi.height = (int)(aveMouth.Height + PrErrY * 2);
			// mouth locating	
			MouthMin = cvRound(aveMouth.Width / ResampleInterval);
			MouthMax = cvRound(aveMouth.Width * ResampleInterval);
			// set input gray scale image
			if(img->nChannels == 3)
			{
				// convert color image to gray scale image
				CV_CALL( iplRoiGrayImage = cvCreateImage(cvSize(img->width, img->height), IPL_DEPTH_8U, 1) );
				if (iplRoiGrayImage == NULL) 
					CV_ERROR(CV_StsBadFunc, "Fail to allocate region image!");			
				CV_CALL( cvCvtColor(img, iplRoiGrayImage, CV_BGR2GRAY ) );
			}
			else
			{
				iplRoiGrayImage = img;
			}
			// mouth detect
			CV_CALL( cvSetImageROI( iplRoiGrayImage, searchRoi) );
			CV_CALL( icvHaarDetectMouthDualCascades(iplRoiGrayImage, MouthTracker, ResampleInterval, MouthMin, MouthMax,  mouth_rect, isMouthDetected) );
			cvReleaseImage(&iplRoiGrayImage);		
			if (isMouthDetected) 
			{
				ErrFrameInTrack = 0;
				result.TrackRltSts = TRACKSUCCESSFUL;
				// get mouths
				mouth.CenterX = (float)(mouth_rect.x + mouth_rect.width / 2);
				mouth.CenterY = (float)(mouth_rect.y + mouth_rect.height / 2);
				mouth.Width = (float)mouth_rect.width;
				mouth.Height = (float)mouth_rect.height;				
				mouth.Likeness = 1.0f;
				mouth.RotAngle = 0.0f;
				// kalman measurement update				
				CV_CALL( icvUpdateKalmanXY(mouth.CenterX, mouth.CenterY, MouthTracker, mouth.CenterX, mouth.CenterY) );
				// Save result
				result.MouthResult = mouth;
				oldMouth = mouth;
				// record track in initial stack
				CV_CALL( icvResultPoolPush(mouth, MouthTracker) );			
			} 
			else 
			{
				// mouth tracking error
				ErrFrameInTrack ++;
				result.TrackRltSts = MOUTHTRACKERROR;
				if (ErrFrameInTrack <= MaxContErrInTrack) 
				{
					// error number is less than Max, use old mouth position to update Kalman filter
					// kalman measurement update
					CV_CALL( icvUpdateKalmanXY(oldMouth.CenterX, oldMouth.CenterY, MouthTracker, oldMouth.CenterX, oldMouth.CenterY) );					
				}
				else 
				{
					// too much error, detect face again
					// reset counter
					CV_CALL( icvResultPoolClear(MouthTracker) );
					//state transfer
					MouthTracker->TrackState = FACEMOUTHDETECTION;
				}
			}
			break;
	}	
	
	return CV_StsOk;
	
	__END__;	

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			cvPostProcessTrackMouth( CvMouthLocateResult * TrackList, const int framenum ); 
//    Purpose:
//          The function cvPostProcessTrackMouth refine the trajectory of the mouth region 
//			using the linear interpolation, median filter and Gaussian filter. 
//    Context:
//    Parameters:
//			TrackList	---- [I/O] Input the list of mouth detection and tracking results. 
//									Output the updated list.
//			framenum	---- [I] Input the number of results in TrackList.
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvPostProcessTrackMouth( CvMouthLocateResult * TrackList, const int framenum )
{

	int WindowRadius;

	CV_FUNCNAME( "cvPostProcessTrackMouth" );

	__BEGIN__;

	if (TrackList == NULL || framenum < 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input argument!");

	if (framenum < 5) 
		return CV_StsOk;

	CV_CALL( icvRegularTrack(TrackList, framenum) );

	if (framenum < 5)
		return CV_StsOk;

	// smoothing location
	CV_CALL( icvMedianSmoothTrackXY(TrackList, framenum, 5) );	
	CV_CALL( icvGaussianSmoothTrackXY(TrackList, framenum, 5, 4) );

	// smoothing size
	if (framenum >= 101) 
	{
		CV_CALL( icvMedianSmoothTrackSize(TrackList, framenum, 50) );
		CV_CALL( icvGaussianSmoothTrackSize(TrackList, framenum, 50, 50) );
	}
	else 
	{
		WindowRadius = (framenum - 1) / 2;
		CV_CALL( icvMedianSmoothTrackSize(TrackList, framenum, WindowRadius) );
		CV_CALL( icvGaussianSmoothTrackSize(TrackList, framenum, WindowRadius, WindowRadius) );
	}

	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvReleaseTrackMouth(CvMouthTracker* MouthTracker)
//    Purpose:
//          The function cvReleaseTrackMouth releases three boosted cascade classifiers, 
//			the buffers and the Kalman filter used in mouth detection and tracking. 
//			The data structure CvMouthTracker itself will not be released.
//    Context:
//    Parameters:
//			MouthTracker ---- [I] MouthTracker to be released
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvReleaseTrackMouth(CvMouthTracker* MouthTracker)
{
	CV_FUNCNAME( "cvReleaseTrackMouth" );

	__BEGIN__;
	
	// check if mouth tracker is null
	if( MouthTracker == NULL)
		CV_ERROR( CV_StsNullPtr, "Invalid pointer of mouth tracker!");

	// Clear classifier objects
	if (MouthTracker->hidCascadeMouth0 != NULL) 
		CV_CALL( cvReleaseHidHaarClassifierCascade(&MouthTracker->hidCascadeMouth0) );
	
	if (MouthTracker->hidCascadeMouth1!= NULL)
		CV_CALL( cvReleaseHidHaarClassifierCascade(&MouthTracker->hidCascadeMouth1) );
	
	// release kalman filter
	if (MouthTracker->iskalmanCreated) 
		CV_CALL( icvReleaseKalmanXY(MouthTracker) );
	
	MouthTracker->iskalmanCreated = FALSE;
	
	// release face detector
	if (MouthTracker->hidCascadeFace != NULL)
	{
		cvReleaseHidHaarClassifierCascade(&MouthTracker->hidCascadeFace);
		MouthTracker->hidCascadeFace = NULL;
	}

	// release down sample image
	if (MouthTracker->iplDownSmpImage) 
	{
		cvReleaseImage(&MouthTracker->iplDownSmpImage);
		MouthTracker->iplDownSmpImage = NULL;
	}
	
	return CV_StsOk;
	
	__END__;	

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvHaarDetectMaxFace(const IplImage *iplImage, CvHidHaarClassifierCascade* hidCascade, CvRect &face)
//    Purpose:
//          Detect face using Haar based face detector, get the maxium face in size.
//    Context:
//    Parameters:
//			iplImage ---- [I] input image, must be a grayscale one.
//			hidCascade -- [I] input face detector model.
//			face	 ---- [O] the maxium face detected.
//    Returns:
//			return the number of face detected in the image, -1 for error case
//    Notes:
//F*/
CV_IMPL int
icvHaarDetectMaxFace(const IplImage *iplImage, CvHidHaarClassifierCascade* hidCascade, CvRect &face)
{
	const float ResRate = 1.1f;
	const int	Neighbour = 1;
	CvMemStorage* storage;
	CvSeq* faces;
	int FaceNum;
	int MaxSize;
	int i;
	CvRect r;

	CV_FUNCNAME( "icvHaarDetectMaxFace" );

	__BEGIN__;

	// check input parameters
	if (iplImage == NULL || iplImage->nChannels != 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input image!");

	if (hidCascade == NULL) 
		CV_ERROR(CV_StsBadArg, "Invalid input model for face detection!");

	// face detection
	CV_CALL( storage = cvCreateMemStorage() );
	cvClearMemStorage( storage );	
	CV_CALL( faces = cvHaarDetectObjects(iplImage, hidCascade, storage, ResRate, Neighbour) );
	if (faces == NULL) 
	{
		cvReleaseMemStorage(&storage);
		return 0;	// not found face
	}

	// get maxium face in size
	FaceNum = faces->total;
	MaxSize = 0;
	for (i=0; i<FaceNum; i++) 
	{
		r = *(CvRect*)cvGetSeqElem(faces, i);
		if (r.width > MaxSize) 
		{
			face.x = r.x;
			face.width = r.width;
			face.y = r.y;
			face.height = r.height;
			MaxSize = r.width;
		}
	}

	cvReleaseMemStorage(&storage);

	return FaceNum;

	__END__;

	return -1;	// any detect error
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvHaarEstimateMouthRegion(const CvRect face, CvRect &rectSearch, int &MouthWidth)
//    Purpose:
//			Estimated the search region and the mouth width from the face detection result.
//    Context:
//    Parameters:
//			face	   ---- [I] input face rectangle
//			rectSearch ---- [O] estimated search region for mouth detector
//			MouthWidth ---- [O] estimated width of the mouth region
//    Returns:
//			return CV_StsOk if successes, else for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus
icvHaarEstimateMouthRegion(const CvRect face, CvRect &rectSearch, int &MouthWidth)
{
	const float MouthVsFace = 0.3f;
	const float MouthFromFaceTop = 0.806f;
	const float VarX = 0.06f;
	const float VarY = 0.10f;
	const float MouthRegionShape = 0.875;

	int FaceWidth;
	int FaceHeight;
	int HalfMouthWidth;
	int HalfMouthHeight;
	int CenterX;
	int CenterY;
	int AddSpaceX;
	int AddSpaceY;
	
	CV_FUNCNAME( "icvHaarEstimateMouthRegion" );

	__BEGIN__;

	FaceWidth = face.width;
	FaceHeight = face.height;
	if (FaceWidth < 1 || FaceHeight < 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input face rectangle!");

	// compute mouth size
	MouthWidth = cvRound(FaceWidth * MouthVsFace);
	HalfMouthWidth = MouthWidth/ 2;
	HalfMouthHeight = cvRound(HalfMouthWidth * MouthRegionShape);

	// compute center
	CenterX = face.x + face.width / 2;
	CenterY = face.y + cvRound(FaceHeight * MouthFromFaceTop);

	// compute additional space
	AddSpaceX = cvRound(FaceWidth * VarX);
	AddSpaceY = cvRound(FaceHeight * VarY);

	// compute search region
	rectSearch.x = CenterX - HalfMouthWidth - AddSpaceX;
	rectSearch.width = MouthWidth + AddSpaceX * 2;
	rectSearch.y = CenterY - HalfMouthHeight - AddSpaceY;
	rectSearch.height = ( HalfMouthHeight + AddSpaceY ) * 2;

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: icvHaarDetectMouthDualCascades( const IplImage *iplImage, 
//											CvMouthTracker* MouthTracker, 
//											double scale_factor, 
//											int min_width, int max_width,
//											CvRect &mouth,
//											BOOL &isMouthDetected)
//
//    Purpose:
//          detect mouth using two boosted cascade classifiers in the ROI of the input image.
//    Context:
//    Parameters:
//			iplImage	---- [I] input gray scale image.
//			scale_factor --- [I] rate for downsample the image
//			min_width, max_width - [I] the minimal and maximal with of the mouth to detect
//			mouth		---- [O] detected mouth region
//			isMouthDetected ---- [O] if detect mouth success TURE, otherwise FALSE
//    Returns:
//			return CV_StsOk if complete detecting, else for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
icvHaarDetectMouthDualCascades(IplImage *iplImage, CvMouthTracker* MouthTracker, double scale_factor, int min_width, int max_width, CvRect &mouth, bool &isMouthDetected)
{

	const int Neighbour = 2;
	
	CvSeq *mouths0, *mouths1;
	IplImage *iplRoi;
	CvRect *r, roi;
	int i;	
	int maxWidth;

	CV_FUNCNAME( "icvHaarDetectMouthDualCascades" );

	__BEGIN__;

	if (iplImage == NULL || min_width < MouthTracker->MouthWndWidth || max_width < min_width) 
		CV_ERROR(CV_StsBadArg, "Invalid input argument!");

	if (MouthTracker->StorageDual0 == NULL || MouthTracker->hidCascadeMouth0 == NULL 
		|| MouthTracker->StorageDual1 == NULL || MouthTracker->hidCascadeMouth1 == NULL) 
		CV_ERROR(CV_StsBadArg, "Invalid cascade classifier for mouth detection!");

	if (iplImage->nChannels != 1)
		CV_ERROR(CV_StsBadArg, "Invalid input argument!");

	isMouthDetected = FALSE;

	CV_CALL( cvClearMemStorage(MouthTracker->StorageDual0) );
	CV_CALL( cvClearMemStorage(MouthTracker->StorageDual1) );

	// copy ROI region
	CV_CALL( roi = cvGetImageROI(iplImage) );
	CV_CALL( iplRoi = cvCreateImage(cvSize(roi.width, roi.height), IPL_DEPTH_8U, 1) );
	CV_CALL( cvCopyImage(iplImage, iplRoi) );

	// detect mouth				
	CV_CALL( mouths0 = cvHaarDetectObjectsR(iplRoi, MouthTracker->hidCascadeMouth0, MouthTracker->StorageDual0, scale_factor, Neighbour, 0, min_width, max_width) );
	CV_CALL( mouths1 = cvHaarDetectObjectsR(iplRoi, MouthTracker->hidCascadeMouth1, MouthTracker->StorageDual1, scale_factor, Neighbour, 0, min_width, max_width) );
	cvReleaseImage(&iplRoi);
		
	// shift mouth result
	if (mouths0 != NULL && mouths0->total > 0) 
	{
		for (i=0; i<mouths0->total; i++) 
		{
			r = (CvRect*)cvGetSeqElem( mouths0, i );
			r->x += roi.x;
			r->y += roi.y;
		}
	}
	if (mouths1 != NULL && mouths1->total > 0) 
	{
		for (i=0; i<mouths1->total; i++) 
		{
			r = (CvRect*)cvGetSeqElem( mouths1, i );
			r->x += roi.x;
			r->y += roi.y;
		}
	}

	// post process to find maxsize mouth region
	maxWidth = 0;
	if (mouths0 != NULL && mouths0->total > 0)
	{
		isMouthDetected = TRUE;
		for ( i=0; i<mouths0->total; i++) 
		{
			r = (CvRect*)cvGetSeqElem( mouths0, i );
			if (r->width > maxWidth) 
			{
				mouth.x = r->x;
				mouth.y = r->y;
				mouth.width = r->width;
				mouth.height = r->height;
				maxWidth = r->width;
			}
		}
	}
	if (mouths1 != NULL && mouths1->total > 0) {
		isMouthDetected = TRUE;		
		for (int i=0; i<mouths1->total; i++) 
		{
			r = (CvRect*)cvGetSeqElem( mouths1, i );
			if (r->width > maxWidth) {
				mouth.x = r->x;
				mouth.y = r->y;
				mouth.width = r->width;
				mouth.height = r->height;
				maxWidth = r->width;
			}
		}
	} 

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			icvResultPoolInit(const int PoolSize, CvMouthTracker* MouthTracker)
//    Purpose:
//          Initialize the pool for tracking results
//    Context:
//    Parameters:
//			PoolSize		---- [I] length of the pool
//			MouthTracker --- [O] MouthTracker containing tracking result pool to be initialized
//    Returns:
//			return CV_StsOk if successful, CV_StsError for failure
//    Notes:
//F*/
CV_IMPL CVStatus
icvResultPoolInit(const int PoolSize, CvMouthTracker* MouthTracker)
{
	CV_FUNCNAME( "icvResultPoolInit" );

	__BEGIN__;

	if (PoolSize > 64 || PoolSize < 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input pool size!");

	MouthTracker->RltPoolLen = PoolSize;
	MouthTracker->RltPoolHeader = 0;
	MouthTracker->isRltPoolFull = FALSE;

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			icvResultPoolPush(const CvMouthLocateResult &result, CvMouthTracker* MouthTracker)
//    Purpose:
//          push a result into the pool
//    Context:
//    Parameters:
//			result		---- [I] result to push
//			MouthTracker --- [I/O] MouthTracker containing the updated tracking result pool
//    Returns:
//    Notes:
//F*/
CV_IMPL void
icvResultPoolPush(const CvMouthLocateResult &result, CvMouthTracker* MouthTracker)
{
	CV_FUNCNAME( "icvResultPoolPush" );

	__BEGIN__;

	MouthTracker->ResultPool[MouthTracker->RltPoolHeader] = result;
	MouthTracker->RltPoolHeader ++;
	if (MouthTracker->RltPoolHeader >= MouthTracker->RltPoolLen) 
	{
		MouthTracker->RltPoolHeader = 0;
		MouthTracker->isRltPoolFull = TRUE;
	}

	__END__;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			icvResultPoolAverage(const CvMouthTracker* MouthTracker, 
//								 CvMouthLocateResult &result)
//    Purpose:
//          Count the average result of the results in the pool
//    Context:
//    Parameters:
//			MouthTracker --- [I] MouthTracker containing the tracking result pool
//			result		---- [O] average of the results in the pool
//    Returns:
//			return CV_StsOk if success, else CV_StsError for error
//    Notes:
//F*/
CV_IMPL CVStatus
icvResultPoolAverage(const CvMouthTracker* MouthTracker, CvMouthLocateResult &result)
{

	int EndIndex;
	int i;
	int isItemValid[CV_MOUTH_TRAJECTORY_LIST_SIZE];
	float LargestSize = 0.0f;
	int LargestIndex = -1;
	float Size;
	float SmallestSize;
	int SmallestIndex = -1;
	int Count = 0;

	CV_FUNCNAME( "icvResultPoolAverage" );

	__BEGIN__;
	
	result.CenterX = 0;
	result.CenterY = 0;
	result.Height = 0;
	result.Width = 0;
	result.Likeness = 0;
	result.RotAngle = 0;

	if (MouthTracker->RltPoolLen < 1) 
		CV_ERROR(CV_StsBadArg, "Invalid result pool!");

	EndIndex = MouthTracker->isRltPoolFull ? MouthTracker->RltPoolLen : MouthTracker->RltPoolHeader;
	if (EndIndex <= 0)
		CV_ERROR(CV_StsBadArg, "Empty result pool!");

	// validate damp list
	for (i=0; i<EndIndex; i++) 
		isItemValid[i] = TRUE;
	
	if (EndIndex > 3) 
	{
		// ignore the one with the largest size
		for (i=0; i<EndIndex; i++) 
		{
			Size = MouthTracker->ResultPool[i].Width;
			if (Size > LargestSize) 
			{
				LargestSize = Size;
				LargestIndex = i;
			}
		}
		isItemValid[LargestIndex] = FALSE;
		
		// ignore the one with the mallest size
		SmallestSize = LargestSize;
		for (i=0; i<EndIndex; i++) 
		{
			Size = MouthTracker->ResultPool[i].Width;
			if (Size < SmallestSize) 
			{
				SmallestSize = Size;
				SmallestIndex = i;
			}
		}
		isItemValid[SmallestIndex] = FALSE;
	}
	
	for (i=0; i<EndIndex; i++) 
	{
		if (!isItemValid[i]) 
			continue;
		
		result.CenterX += MouthTracker->ResultPool[i].CenterX;
		result.CenterY  += MouthTracker->ResultPool[i].CenterY;
		result.Width += MouthTracker->ResultPool[i].Width;
		result.Height += MouthTracker->ResultPool[i].Height;
		result.RotAngle += MouthTracker->ResultPool[i].RotAngle;
		result.Likeness += MouthTracker->ResultPool[i].Likeness;
		Count ++;
	}
	result.CenterX = result.CenterX / (float)Count;
	result.CenterY = result.CenterY / (float)Count;
	result.Width = result.Width / (float)Count;
	result.Height = result.Height / (float)Count;
	result.RotAngle = result.RotAngle / (float)Count;
	result.Likeness = result.Likeness / (float)Count;

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			icvResultPoolClear(const CvMouthTracker* MouthTracker)
//    Purpose:
//          Clear the tracking result pool
//    Context:
//    Parameters:
//			MouthTracker -- [I/O] MouthTracker containing the tracking result pool
//    Returns:
//			return CV_StsOk if success, else CV_StsError for error
//    Notes:
//F*/
CV_IMPL CVStatus
icvResultPoolClear(CvMouthTracker* MouthTracker)
{
	CV_FUNCNAME( "icvResultPoolClear" );

	__BEGIN__;
	
	if( MouthTracker == NULL)
		CV_ERROR( CV_StsNullPtr, "Invalid pointer of mouth tracker!");
	
	MouthTracker->RltPoolHeader = 0;
	MouthTracker->isRltPoolFull = FALSE;
	MouthTracker->InitialCounter = 0;	
	
	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvCreateKalmanXY(const int KinOrder, const float Tx, const float Ty, const float Q, const float R, CvMouthTracker* MouthTracker)
//    Purpose:
//			Create Kalman filter for X Y trajectory prediction
//    Context:
//    Parameters:
//			KinOrder ---- [I] For kinematic model. 2, use position and velocity; 3, use acceleration in addition
//			Tx, Ty  ---- [I] factor in kalman matrix A, recomendated to use frame interval in second
//			Q  ---- [I] multiple factor in process noice matrix which is a identity one.
//			R  ---- [I] multiple factor in measurement noice matrix which is a identity one.
//			MouthTracker -- [O] containing created kalman filter
//    Returns: 
//			return CV_StsOk if successfully create a kalman filter, else return CV_StsError;
//    Notes:
//F*/
CV_IMPL CVStatus
icvCreateKalmanXY(const int KinOrder, const float Tx, const float Ty, const float Q, const float R, CvMouthTracker* MouthTracker)
{
	CvMat Dyn;
	CvMat Mes;
	CvMat PNC;
	CvMat MNC;
	CvMat PriErr;
	CvMat PostErr;
	CvMat PriState;

	CV_FUNCNAME( "icvCreateKalmanXY" );

	__BEGIN__;

	// create Kalman filter
	if (KinOrder == 2) 
	{
		CV_CALL( icvReleaseKalmanXY(MouthTracker) );
		MouthTracker->KinematicOrder = 2;
		MouthTracker->KalmanXY = cvCreateKalman(4,4);
		CV_CALL( Dyn = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->DynamMatr) );
		CV_CALL( Mes = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->MeasurementMatr) );
		CV_CALL( PNC = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->PNCovariance) );
		CV_CALL( MNC = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->MNCovariance) );
		CV_CALL( PriErr = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->PriorErrorCovariance) );
		CV_CALL( PostErr = cvMat(4,4,CV_32FC1,MouthTracker->KalmanXY->PosterErrorCovariance) );
		CV_CALL( PriState = cvMat(4,1,CV_32FC1,MouthTracker->KalmanXY->PriorState) );
		cvmSetZero(&PNC);				// Q (process)
		cvmSet(&PNC, 0, 0, Q);			// Q
		cvmSet(&PNC, 1, 1, Q);
		cvmSet(&PNC, 2, 2, Q);
		cvmSet(&PNC, 3, 3, Q);
		cvmSetIdentity(&PriErr);		// Pk-
		cvmSetIdentity(&PostErr);		// Pk
		cvmSetZero(&MNC);				// R (measure)
		cvmSet(&MNC, 0, 0, R);			// R
		cvmSet(&MNC, 1, 1, R);
		cvmSet(&MNC, 2, 2, R);
		cvmSet(&MNC, 3, 3, R);
		cvmSetZero(&PriState);			// Xk-
		cvmSetIdentity(&Dyn);			// A
		cvmSet(&Dyn,0,1,Tx);
		cvmSet(&Dyn,2,3,Ty);
		cvmSetIdentity(&Mes);			// H
		CV_CALL( MouthTracker->MeasureVectXY = cvCreateMat(4, 1, CV_32FC1) );
		CV_CALL( cvSetZero(MouthTracker->MeasureVectXY) );

		MouthTracker->PrevX = 0;
		MouthTracker->PrevY = 0;		
	}
	else if (KinOrder == 3) 
	{
		CV_CALL( icvReleaseKalmanXY(MouthTracker) );
		MouthTracker->KinematicOrder = 3;
		MouthTracker->KalmanXY = cvCreateKalman(6,6);
		CV_CALL( Dyn = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->DynamMatr) );
		CV_CALL( Mes = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->MeasurementMatr) );
		CV_CALL( PNC = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->PNCovariance) );
		CV_CALL( MNC = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->MNCovariance) );
		CV_CALL( PriErr = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->PriorErrorCovariance) );
		CV_CALL( PostErr = cvMat(6,6,CV_32FC1,MouthTracker->KalmanXY->PosterErrorCovariance) );
		CV_CALL( PriState = cvMat(6,1,CV_32FC1,MouthTracker->KalmanXY->PriorState) );
		cvmSetZero(&PNC);				// Q (process)
		cvmSet(&PNC, 0, 0, Q);			// Q
		cvmSet(&PNC, 1, 1, Q);
		cvmSet(&PNC, 2, 2, Q);
		cvmSet(&PNC, 3, 3, Q);
		cvmSet(&PNC, 4, 4, Q);
		cvmSet(&PNC, 5, 5, Q);
		cvmSetIdentity(&PriErr);		// Pk-
		cvmSetIdentity(&PostErr);		// Pk
		cvmSetZero(&MNC);				// R (measure)
		cvmSet(&MNC, 0, 0, R);			// R
		cvmSet(&MNC, 1, 1, R);
		cvmSet(&MNC, 2, 2, R);
		cvmSet(&MNC, 3, 3, R);
		cvmSet(&MNC, 4, 4, R);
		cvmSet(&MNC, 5, 5, R);
		cvmSetZero(&PriState);			// Xk-
		cvmSetIdentity(&Dyn);			// A
		cvmSet(&Dyn, 0, 1, Tx);	
		cvmSet(&Dyn, 0, 2, 0.5*Tx*Tx);
		cvmSet(&Dyn, 1, 2, Tx);
		cvmSet(&Dyn, 3, 4, Ty);	
		cvmSet(&Dyn, 3, 5, 0.5*Ty*Ty);
		cvmSet(&Dyn, 4, 5, Ty);
		cvmSetIdentity(&Mes);			// H
		CV_CALL( MouthTracker->MeasureVectXY = cvCreateMat(6, 1, CV_32FC1) );
		CV_CALL( cvSetZero(MouthTracker->MeasureVectXY) );

		MouthTracker->PrevX = 0;
		MouthTracker->PrevY = 0;
		MouthTracker->PrevVx = 0;
		MouthTracker->PrevVy = 0;
		return CV_StsOk;
	}
	else
	{
		CV_ERROR(CV_StsBadArg, "Invalid kinematic order! Must be 2 or 3!");
	}
	
	return CV_StsOk;

	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//		icvInitKalmanXY(const float X0, const float Y0, 
//						const float Vx0, const float Vy0,
//						CvMouthTracker* MouthTracker );
//    Purpose:
//			Set the initial state (in fact the measurement) of the kalman filter
//    Context:
//    Parameters:
//			X0, Y0		---- [I] initial position (in previous step)
//			Vx0, Vy0	---- [I] initial velocity (in previous step)
//			MouthTracker --- [I/O] contains the kalman filter to initialize
//    Returns: 
//			return CV_StsOk if successfully, else return CV_StsError;
//    Notes:
//F*/
CV_IMPL CVStatus
icvInitKalmanXY( const float X0, const float Y0, const float Vx0, const float Vy0, CvMouthTracker* MouthTracker)
{

	CV_FUNCNAME( "icvInitKalmanXY" );

	__BEGIN__;

	if (MouthTracker->KalmanXY == NULL) 
		CV_ERROR(CV_StsBadArg, "Invalid input Kalman filter!");

	MouthTracker->PrevX = X0;
	MouthTracker->PrevY = Y0;
	MouthTracker->PrevVx = Vx0;
	MouthTracker->PrevVy = Vy0;

	if (MouthTracker->KinematicOrder == 2) 
	{
		MouthTracker->KalmanXY->PriorState[0] = X0;
		MouthTracker->KalmanXY->PriorState[1] = Vx0;
		MouthTracker->KalmanXY->PriorState[2] = Y0;
		MouthTracker->KalmanXY->PriorState[3] = Vy0;
	}
	else if (MouthTracker->KinematicOrder == 3) 
	{
		MouthTracker->KalmanXY->PriorState[0] = X0;
		MouthTracker->KalmanXY->PriorState[1] = Vx0;
		MouthTracker->KalmanXY->PriorState[2] = 0;
		MouthTracker->KalmanXY->PriorState[3] = Y0;
		MouthTracker->KalmanXY->PriorState[4] = Vy0;
		MouthTracker->KalmanXY->PriorState[5] = 0;
	}

	return CV_StsOk;
	
	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvUpdateKalmanXY(const float X, const float Y, CvMouthTracker* MouthTracker, float &SmX, float &SmY)
//    Purpose:
//			Do measure update (correct)
//    Context:
//    Parameters:
//			X, Y		---- [I] new detection position
//			MouthTracker --- [I/O] KalmanXY filter in it to be updated.
//			SmX, SmY	---- [O] smoothed state
//    Returns: 
//		  return CV_StsOk if successfully, else return CV_StsError;
//    Notes:
//F*/
CV_IMPL CVStatus
icvUpdateKalmanXY(const float X, const float Y, CvMouthTracker* MouthTracker, float &SmX, float &SmY)
{

	CV_FUNCNAME( "icvUpdateKalmanXY" );

	__BEGIN__;

	if (MouthTracker->KalmanXY == NULL) 
		CV_ERROR(CV_StsBadArg, "Invalid input Kalman filter!");

	if (MouthTracker->KinematicOrder == 2)
	{
		// count measurement vector
		MouthTracker->MeasureVectXY->data.fl[0] = X;
		MouthTracker->MeasureVectXY->data.fl[1] = X - MouthTracker->PrevX;
		MouthTracker->MeasureVectXY->data.fl[2] = Y;
		MouthTracker->MeasureVectXY->data.fl[3] = Y - MouthTracker->PrevY;
		// save data as previous step's
		MouthTracker->PrevX = X;
		MouthTracker->PrevY = Y;
		MouthTracker->PrevVx = MouthTracker->MeasureVectXY->data.fl[1];
		MouthTracker->PrevVy = MouthTracker->MeasureVectXY->data.fl[3];
		// update
		CV_CALL( cvKalmanUpdateByMeasurement(MouthTracker->KalmanXY, MouthTracker->MeasureVectXY) );
		// count smoothed obs (omit multiple by H)
		SmX = MouthTracker->KalmanXY->PosterState[0];
		SmY = MouthTracker->KalmanXY->PosterState[2];

		return CV_StsOk;
	}
	
	if (MouthTracker->KinematicOrder == 3) 
	{
		// count measurement vector
		MouthTracker->MeasureVectXY->data.fl[0] = X;
		MouthTracker->MeasureVectXY->data.fl[1] = X - MouthTracker->PrevX;
		MouthTracker->MeasureVectXY->data.fl[2] = MouthTracker->MeasureVectXY->data.fl[1] - MouthTracker->PrevVx;
		MouthTracker->MeasureVectXY->data.fl[3] = Y;
		MouthTracker->MeasureVectXY->data.fl[4] = Y - MouthTracker->PrevY;
		MouthTracker->MeasureVectXY->data.fl[5] = MouthTracker->MeasureVectXY->data.fl[4] - MouthTracker->PrevVy;
		// save data as previous step's
		MouthTracker->PrevX = X;
		MouthTracker->PrevY = Y;
		MouthTracker->PrevVx = MouthTracker->MeasureVectXY->data.fl[1];
		MouthTracker->PrevVy = MouthTracker->MeasureVectXY->data.fl[4];
		// update
		CV_CALL( cvKalmanUpdateByMeasurement(MouthTracker->KalmanXY, MouthTracker->MeasureVectXY) );
		// count smoothed obs (omit multiple by H)
		SmX = MouthTracker->KalmanXY->PosterState[0];
		SmY = MouthTracker->KalmanXY->PosterState[3];

		return CV_StsOk;
	}

	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			icvReleaseKalmanXY(CvKalman *KalmanXY)
//    Purpose:
//          Release Kalman filter
//    Context:
//    Parameters:
//			MouthTracker --- [I/O] containing Kalman filter to be released
//    Returns:
//    Notes:
//F*/
CV_IMPL void
icvReleaseKalmanXY(CvMouthTracker* MouthTracker)
{
	CV_FUNCNAME( "icvReleaseKalmanXY" );

	__BEGIN__;
	
	if (MouthTracker->KalmanXY != NULL) 
	{
		cvReleaseKalman(&MouthTracker->KalmanXY);
		MouthTracker->KalmanXY = NULL;
	}

	if (MouthTracker->MeasureVectXY != NULL) {
		CV_CALL( cvReleaseMat(&MouthTracker->MeasureVectXY) );
		MouthTracker->MeasureVectXY = NULL;
	}

	__END__;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			icvCheckMouthTrackResult(CvMouthLocateResult &result);
//    Purpose:
//          Check if the mouth tracking result is a valid one.
//    Context:
//    Parameters:
//			result ---- [I] mouth locate result to be checked
//    Returns:
//			return TRUE if valid, else FALSE
//    Notes:
//F*/
OPENCVAUXAPI bool
icvCheckMouthTrackResult(CvMouthLocateResult &result)
{
	CV_FUNCNAME( "icvCheckMouthTrackResult" );

	__BEGIN__;

	if ((result.Width <= 0) || (result.Height <= 0)) 
		return FALSE;
	if (result.RotAngle < -360.0f || result.RotAngle > 360.0f) 
		return FALSE;
	if (result.Likeness < 0.0f || result.Likeness > 1.0f) 
		return FALSE;
	
	return TRUE;

	__END__;
	
	return FALSE;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvRegularTrack(CvMouthLocateResult * Trajectory, const int FrameNum)
//    Purpose:
//          Refularize the mouth track.
//    Context:
//    Parameters:
//			Trajectory ---- [I/O] lip track
//			FrameNum ---- [I] frame number of the track
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//F*/
CV_IMPL CVStatus
icvRegularTrack(CvMouthLocateResult * Trajectory, const int FrameNum)
{
	int i, j;
	int indexPre, indexNext;
	int isFound;
	int HaveValid = FALSE;
	CvMouthLocateResult resultNext, resultPre;
	CvMouthLocateResult *Track;
	float A;
	float B;
	float Div;

	CV_FUNCNAME( "icvRegularTrack" );

	__BEGIN__;

	if (Trajectory == NULL || FrameNum < 1) 
		CV_ERROR(CV_StsBadArg, "Invlid input argument!");

	// fill the blank items
	for (i=0; i<FrameNum; i++) 
	{
		if ( icvCheckMouthTrackResult(Trajectory[i]) ) 
		{
			HaveValid = TRUE;
			continue;
		}
		// search forward
		isFound =FALSE;
		for (j=i+1; j < FrameNum && (!isFound); j++) 
		{
			if ( !icvCheckMouthTrackResult(Trajectory[j]) ) 
				continue;
			resultNext = Trajectory[j];
			indexNext = j;
			isFound = TRUE;
		}
		// get forward
		if (i > 0) 
		{
			resultPre = Trajectory[i-1];
			indexPre = i - 1;
			if (isFound) 
			{
				Track = &(Trajectory[i]);
				A = float(indexNext - i);		// next
				B = float(i - indexPre);		// pre
				Div = A + B;
				Track->CenterX = (A * resultPre.CenterX + B * resultNext.CenterX) / Div;
				Track->CenterY = (A * resultPre.CenterY + B * resultNext.CenterY) / Div;
				Track->Width = (A * resultPre.Width + B * resultNext.Width) / Div;
				Track->Height = (A * resultPre.Height + B * resultNext.Height) / Div;
				Track->RotAngle = (A * resultPre.RotAngle + B * resultNext.RotAngle) / Div;
				Track->Likeness = (A * resultPre.Likeness + B * resultNext.Likeness) / Div;
			}
			else 
			{
				Trajectory[i] = resultPre;
			}
		}
		else 
		{
			if (isFound) 
			{
				Trajectory[i] = resultNext;
			}
		} // if (i>0)
	}

	return CV_StsOk;

	__END__;

	return CV_StsError;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvGaussianSmooth(double* Array, const int ArrayLength, double *Filter, const int FilterLength, const int Variance)
//    Purpose:
//          smooth array using gaussian filter.
//    Context:
//    Parameters:
//			Array ---- [I/O] input data array to be smoothed
//			ArrayLength ---- [I] array length
//			Filter ---- [I] input filter buffer
//			FilterLength ---- [I] length of filter
//			Variance ---- [I] variance of filter
//    Returns:
//    Notes:
//F*/
CV_IMPL void 
icvGaussianSmooth(double* Array, const int ArrayLength, double *Filter, const int FilterLength, const int Variance)
{
	int i, j, n;
	int WndRadius;
	int Variable;
	double temp0, temp1;
	double total = 0.0, sum, fdiv;
	double *ResultArray;

	CV_FUNCNAME( "icvGaussianSmooth" );

	__BEGIN__;
	
	if(Variance <=0) 
		CV_ERROR(CV_StsBadArg, "Invalid input variance data!");
	
	if(FilterLength % 2 != 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input filter length data!");
	
	WndRadius = FilterLength / 2;
	temp0 = 1.0 / (2 * CV_PI * Variance);
	temp1 = 1.0 / (Variance * Variance);
	
	for(i=-WndRadius; i<= WndRadius; i++)
	{
		Filter[i+WndRadius] = temp0 * exp(-i * i * temp1); 
		total += Filter[i+WndRadius];
	}
	
	for(i=-WndRadius; i<= WndRadius; i++)
		Filter[i+WndRadius] /= total;

	CV_CALL( ResultArray = (double *)cvAlloc(ArrayLength * sizeof(double)) );

	Variable = (FilterLength - 1) / 2;
	
	for(i=Variable; i<ArrayLength-Variable; i++)
	{
		sum = 0.0;
		for(j=0; j<FilterLength; j++)
			sum += Filter[j] * Array[i - Variable + j];
		ResultArray[i] = sum;
	}

	// process out of bound
	for(i=0; i<Variable; i++)
	{
		sum = 0.0;
		fdiv = 0.0;
		for(j=0; j<=Variable+i; j++) 
		{
			sum += Filter[Variable-i+j] * Array[j];
			fdiv += Filter[Variable-i+j];
		}
		ResultArray[i] = sum / fdiv;
	}

	for(i=ArrayLength-Variable; i<ArrayLength; i++)
	{
		sum = 0.0;
		fdiv = 0.0;
		n = 0;
		for(j=i-Variable; j<ArrayLength; j++, n++) 
		{
			sum += Filter[n] * Array[j];
			fdiv += Filter[n];
		}
		ResultArray[i] = sum / fdiv;
	}

	for(i=0; i<ArrayLength; i++)
		Array[i] = ResultArray[i];

	__END__;
	
	cvFree((void**)&ResultArray);
	cvFree((void**)&ResultArray);
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvGaussianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance)
//    Purpose:
//			smooth the X Y trajectory using the Gaussian filter.
//    Context:
//    Parameters:
//			Trajectory	---- [I/O] lip track
//			FrameNum	---- [I] frame number of the track
//			WndRadius	---- [I] middle value filter window radius
//			Variance	---- [I] filter variance
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//F*/
CV_IMPL CVStatus
icvGaussianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance)
{
	int i;
	int WndWidth;
	double *filterArray, *data;
	CV_FUNCNAME( "icvGaussianSmoothTrackXY" );

	__BEGIN__;

	if (Trajectory == NULL || (Variance<=0) || (WndRadius <=0))
		CV_ERROR(CV_StsBadArg, "Invalid input trajectory data!");
	
	WndWidth = WndRadius * 2 + 1;
	if (FrameNum <= WndWidth) 
		CV_ERROR(CV_StsBadArg, "Insufficient data point for smoothing!");

	CV_CALL( filterArray = (double *)cvAlloc(WndWidth * sizeof(double)) );

	CV_CALL( data = (double *)cvAlloc(FrameNum * sizeof(double)) );

	// smoothing location
	// X
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].CenterX;
	
	CV_CALL( icvGaussianSmooth(data, FrameNum, filterArray, WndWidth, Variance) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].CenterX = (float) data[i];

	// Y
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].CenterY;
	
	CV_CALL( icvGaussianSmooth(data, FrameNum, filterArray, WndWidth, Variance) );
	
	for(i=0; i<FrameNum; i++) 		
		Trajectory[i].CenterY = (float) data[i];

	cvFree((void**)&data);
	cvFree((void**)&filterArray);

	return CV_StsOk;

	__END__;
	
	cvFree((void**)&data);
	cvFree((void**)&filterArray);
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvGaussianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance)
//    Purpose:
//			Smooth the sizes in the mouth trajectory using Gaussian filter
//    Context:
//    Parameters:
//			Trajectory	---- [I/O] mouth track result to be smoothed
//			FrameNum	---- [I] frame number of the track
//			WndRadius	---- [I] middle value filter window radius
//			Variance	---- [I] filter variance
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//F*/
CV_IMPL CVStatus
icvGaussianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius, const int Variance)
{
	int i;
	int WndWidth;
	double *filterArray, *data;
	
	CV_FUNCNAME( "icvGaussianSmoothTrackSize" );

	__BEGIN__;

	if (Trajectory == NULL || (Variance <= 0) || (WndRadius <= 0))
		CV_ERROR(CV_StsBadArg, "Invalid input trajectory data!");

	WndWidth = WndRadius * 2 + 1;
	if (FrameNum < WndWidth) 
		CV_ERROR(CV_StsBadArg, "Insufficient data point for smoothing!");

	CV_CALL( filterArray = (double *)cvAlloc(WndWidth * sizeof(double)) );

	CV_CALL( data = (double *)cvAlloc(FrameNum * sizeof(double)) );

	// smoothing size
	// width
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].Width;
	
	CV_CALL( icvGaussianSmooth(data, FrameNum, filterArray, WndWidth, Variance) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].Width = (float) data[i];

	// height
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].Height;
	
	CV_CALL( icvGaussianSmooth(data, FrameNum, filterArray, WndWidth, Variance) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].Height = (float) data[i];

	cvFree((void**)&data);
	cvFree((void**)&filterArray);

	return CV_StsOk;

	__END__;
	
	cvFree((void**)&data);
	cvFree((void**)&filterArray);
	
	return CV_StsError;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvMediavalueSmooth(double *Array, const int ArrayLength, const int FilterRadius)
//    Purpose:
//          smooth array using median value filter.
//    Context:
//    Parameters:
//			Array		---- [I/O] input data to be smoothed
//			ArrayLength ---- [I] array length
//			FilterRadius --- [I] 
//    Returns:
//    Notes:
//F*/
CV_IMPL void 
icvMediavalueSmooth(double *Array, const int ArrayLength, const int FilterRadius)
{
	int WndWidth, i, j, k;
	int start_index, end_index, index;
	double *buf,*wnd;
	double maxValue, midValue;

	CV_FUNCNAME( "icvMediavalueSmooth" );

	__BEGIN__;
	
	if (Array == NULL || ArrayLength < 1) 
		CV_ERROR(CV_StsBadArg, "Invalid input data!");

	// allocate buffer
	CV_CALL( buf = (double *)cvAlloc(ArrayLength * sizeof(double)) );
	
	WndWidth = FilterRadius * 2 + 1;
	
	CV_CALL( wnd = (double *)cvAlloc(WndWidth * sizeof(double)) );
	
	for (i=0; i<ArrayLength; i++) 
	{
		// computer boundary
		start_index = i - FilterRadius;
		start_index = start_index < 0 ? 0 : start_index;
		end_index = i + FilterRadius + 1;
		end_index = end_index > ArrayLength ? ArrayLength : end_index;
		WndWidth = end_index - start_index;
		
		// copy window data
		for (j=0; j<WndWidth; j++) 
			wnd[j] = Array[start_index + j];

		// get middle value		
		if (WndWidth == 1) 
			midValue = wnd[0];
		else
		{
			for (j=0; j<FilterRadius; j++) 
			{
				maxValue = - CV_POS_INF;
				index = 0;
				for (k=0; k<WndWidth; k++) 
				{
					if (wnd[k] > maxValue) 
					{
						maxValue = wnd[k];
						index = k;
					}
				}
				wnd[index] = - CV_POS_INF;
			} // end for j
	
			midValue = - CV_POS_INF;
			for (j=0; j<WndWidth; j++) 
			{
				if (wnd[j] > midValue) 
				{
					midValue = wnd[j];
				}
			}	
		}// end if
		
		// set data
		buf[i] = midValue;
	} // end for i

	for (i=0; i<ArrayLength; i++) 
		Array[i] = buf[i];
	
	__END__;
	
	cvFree((void**)&buf);
	cvFree((void**)&wnd);
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvMedianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius)
//    Purpose:
//          Smooth the X Y mouth trajectory using median filter
//    Context:
//    Parameters:
//			Trajectory ---- [I/O] lip track
//			FrameNum  ---- [I] frame number of the track
//			WndRadius ---- [I] middle value filter window radius
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//F*/
CV_IMPL CVStatus
icvMedianSmoothTrackXY(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius)
{

	int i;
	double *data;

	CV_FUNCNAME( "icvMedianSmoothTrackXY" );

	__BEGIN__;

	if (Trajectory == NULL || (WndRadius <=0)) 
		CV_ERROR(CV_StsBadArg, "Invalid input trajectory data!");

	CV_CALL( data = (double *)cvAlloc(FrameNum * sizeof(double)) );

	// smoothing location
	// X
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].CenterX;
	
	CV_CALL( icvMediavalueSmooth(data, FrameNum, WndRadius) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].CenterX = (float) data[i];

	// Y
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].CenterY;
	
	CV_CALL( icvMediavalueSmooth(data, FrameNum, WndRadius) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].CenterY = (float) data[i];

	cvFree((void**)&data);
	
	return CV_StsOk;

	__END__;
	
	cvFree((void**)&data);
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			icvMedianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius)
//    Purpose:
//			Smooth the sizes in the mouth trajectory using Median filter
//    Context:
//    Parameters:
//			Trajectory ---- [I/O] mouth track result to be smoothed
//			FrameNum ---- [I] frame number of the track
//			WndRadius ---- [I] middle value filter window radius
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//F*/
CV_IMPL CVStatus
icvMedianSmoothTrackSize(CvMouthLocateResult *Trajectory, const int FrameNum, const int WndRadius)
{
	int i;
	double *data;
	
	CV_FUNCNAME( "icvMedianSmoothTrackSize" );

	__BEGIN__;

	if (Trajectory == NULL || (WndRadius <=0))
		CV_ERROR(CV_StsBadArg, "Invalid input trajectory data!");
	
	CV_CALL( data = (double *)cvAlloc(FrameNum * sizeof(double)) );

	// smoothing location
	// Width
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].Width;
	
	CV_CALL( icvMediavalueSmooth(data, FrameNum, WndRadius) );
		
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].Width = (float) data[i];

	// Height
	for(i=0; i<FrameNum; i++) 
		data[i] = Trajectory[i].Height;
	
	CV_CALL( icvMediavalueSmooth(data, FrameNum, WndRadius) );
	
	for(i=0; i<FrameNum; i++) 
		Trajectory[i].Height = (float) data[i];

	cvFree((void**)&data);
	
	return CV_StsOk;

	__END__;
	
	cvFree((void**)&data);
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			cvEqualizeHist(IplImage *img)
//    Purpose:
//          The function cvEqualizeHist applies the histogram equalization to the image img. 
//			The ROI of the image will be ignored. 
//    Context:
//    Parameters:
//			img	---- [I/O] The input and output 8-bit grayscale image.
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//			input image must be a allocated gray scale image
//F*/
CV_IMPL CVStatus
cvEqualizeHist(IplImage *img)
{
	int width, height, size;
	uchar *imgArray;
	uchar *array;
	int count[256];	
	int table[256];
	int i, x, y;
	
	CV_FUNCNAME( "cvEqualizeHist" );

	__BEGIN__;

	if (img == NULL || img->nChannels != 1)
		CV_ERROR(CV_StsBadArg, "Invalid input image! This function only supports gray images!");

	width = img->width;
	height = img->height;
	imgArray = (uchar *)(img->imageData);

	for (i=0; i<256; i++) 
		count[i] = 0;
	
	for (y=0; y<height; y++)
	{	
		// calc histogram
		array = imgArray + img->widthStep * y;
		for (x=0; x<width; x++) 
		{
			count[array[x]]++;
		}
	}
	size = height * width;
	table[0] = count[0] / size;
	for (i=1; i<256; i++) 
	{	
		// calc acculated histogram and look up table
		count[i] += count[i-1];
		table[i] = (256 * count[i]) / size;
		if (table[i] > 255)
			table[i] = 255;
	}
	// equalization
	for (y=0; y<height; y++) 
	{	
		array = imgArray + img->widthStep * y;
		for (x=0; x<width; x++) 
		{
			array[x] = table[array[x]];
		}
	}

	return CV_StsOk;

	__END__;

	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			cvNormalizeIllum(IplImage *img, const float destMean, const float destMse)
//    Purpose:
//          The function cvNormalizeIllum implements the illumination gradient correction 
//			to reduce heavy shadows in the image due to extreme lighting angles. The function 
//			adjusts the mean and mean square error of the image img according to destMean and 
//			destMse parameters. The ROI of the image is ignored.
//    Context:
//    Parameters:
//			img		---- [I/O] The input and output 8-bit grayscale image.
//			destMean --- [I] Input the given value of the means of destination image.
//			destMse ---- [I] Input the given value of the mean square error of destination image.
//    Returns:
//			return CV_StsOk if successes, else return CV_StsError
//    Notes:
//			input image must be a allocated gray scale image
//F*/
CV_IMPL CVStatus
cvNormalizeIllum(IplImage *img, const float destMean, const float destMse)
{
	uchar* src_data;
	float* dst_data;
	CvRect src_roi;
	int width_step, width, height, roi_size, img_index, data_index;
	CvSize data_size;
	float temp_float, mse_ratio;
	int i, x, y;
	float A11=0, A12=0, A13=0, A21=0, A22=0, A23=0, A31=0, A32=0, A33=0;
	float I_point;
	float B1=0, B2 = 0, B3 = 0;
	float Det;
	float B11, B12, B13, B22, B23, B33;
	float A1, A2, A3;
	float cal_mean, cal_mse;

	CV_FUNCNAME("cvNormalizeIllum");

	__BEGIN__;

	// check arguments
	if (destMean < 0.0f || destMse < 0.0f)
		CV_ERROR( CV_StsBadArg, "Invalid target mean and MSE." );

	// check image	
	if(img->depth != IPL_DEPTH_8U)
		CV_ERROR( CV_BadDepth, icvUnsupportedFormat );
	if (img->nChannels != 1)
		CV_ERROR( CV_BadNumChannels, icvUnsupportedFormat );

	// Get raw data
	cvGetImageRawData(img, (uchar **) &src_data, &width_step, &data_size);
	src_roi = cvGetImageROI(img);

	width = data_size.width;
	height = data_size.height;
	roi_size = width * height;
	CV_CALL( dst_data = (float *)cvAlloc(roi_size * sizeof(float)) );

	//1. Caculate the A'A
	for(y=0; y<height; y++)
	{
		for(x=0; x<width; x++)
		{
			A11 += x * x;
			A12 += x * y;
			A13 += x;
			A22 += y * y;
			A23 += y;
		}
	}
	A33 = (float)roi_size;

	//2. Caculate the A'B
	img_index = 0;
	for(y=0; y<height; y++)
	{
		for(x=0; x<width; x++)
		{
			I_point = (float)src_data[img_index + x];
			B1 += x * I_point;
			B2 += y * I_point;
			B3 += I_point;
		}
		img_index += width_step;
	}

	//3. Caculate the inverse matrix (A'A)
	Det = - A11*A22*A33 + A11*A23*A23 + A12*A12*A33 - 2*A12*A13*A23 + A13*A13*A22;
	if(Det == 0)
		CV_ERROR( CV_StsError, "Fail to fit the illumination plain." );
	B11 =  - (A22*A33 - A23*A23) / Det;
	B12 = (A12*A33 - A13*A23) / Det;
	B13 = - (A12*A23 - A13*A22) / Det;
	B22 = - (A11*A33 - A13*A13) / Det;
	B23 = - ( - A11*A23 + A12*A13) / Det;
	B33 = ( - A11*A22+A12*A12) / Det;

	//4. Solve equations and find a1, a2, a3
	A1 = B11 * B1 + B12 * B2 + B13 * B3;
	A2 = B12 * B1 + B22 * B2 + B23 * B3;
	A3 = B13 * B1 + B23 * B2 + B33 * B3;

	//5. Brightness correction
	cal_mean = 0;
	data_index = 0;
	img_index = 0;
	for (y=0; y<height; y++)
	{
		for (x=0; x<width; x++)
		{
				I_point = (float)src_data[img_index + x] - (A1 * x + A2 * y + A3);
				dst_data[data_index] = I_point;
				data_index ++;
				cal_mean += I_point;
		}
		img_index += width_step;
	}
	cal_mean /= (width * height);

	// MSE Caculation
	cal_mse = 0;
	for( i=0; i<roi_size; i++)
	{
		temp_float = dst_data[i] - cal_mean;
		cal_mse += temp_float * temp_float;
	}
	cal_mse = (float)sqrt( cal_mse / roi_size );

	// MSE normalization and write back to image buffer
	if (cal_mse == 0) {
		data_index = 0;
		img_index = 0;
		for (y=0; y<height; y++)
		{
			for (x=0; x<width; x++)
			{
				src_data[img_index + x] = (unsigned char)destMean;
			}
			img_index += width_step;
		}
	}
	else {
		mse_ratio = destMse / cal_mse;
		data_index = 0;
		img_index = 0;
		for (y=0; y<height; y++)
		{
			for (x=0; x<width; x++)
			{
				temp_float = (dst_data[data_index] - cal_mean) * mse_ratio + destMean;
				data_index ++;
				if(temp_float > 255.0f)
					src_data[img_index + x] = 255;
				else if (temp_float < 0.0f) 
					src_data[img_index + x] = 0;
				else
					src_data[img_index + x] = (uchar)(temp_float + 0.5f);
			}
			img_index += width_step;
		}
	}
		
	cvFree((void**)&dst_data);
	
	return CV_StsOk;

	__END__;

	return CV_StsError;
}
