/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/


#include "cvvisualspeech.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvInitPCA(const IplImage** srcImg, const int blockImgNum, CvMat* sumVector, CvMat* pCovMatrix);
//    Purpose:
//          Calculates the sum vector and the partial covariance matrix for a set of IPL images
//    Context:
//    Parameters:
//			srcImg ---- [I] pointer to the array of IplImage input images. All the images must be of 
//							the same size, 8-bit and without ROI selected.
//			blockImgNum ---- [I] number of images in srcImage. 
//			sumVector ---- [O] summary of the vectors corresponding to the input images
//			pcovMatrix ---- [O] partial covariance of the vectors corresponding to the input images
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvInitPCA(const IplImage** srcImg, const int blockImgNum, CvMat* sumVector, CvMat* pcovMatrix)
{
	int i, j, k, x, y, index;
	int width, height, widthstep, vecsize, band, outtype;
	unsigned char *imagebuf;
	float *inputdata, tmpfl;
	double tmpdb;
	bool first;

	CV_FUNCNAME( "cvInitPCA" );

   	__BEGIN__;

	// check arguments
	if( !CV_IS_MAT(pcovMatrix) ) 
	{
		CV_ERROR(CV_StsBadArg, "Inalid input matrix data.");
	}
	
	if( !CV_ARE_TYPES_EQ(sumVector, pcovMatrix) )
	{
		CV_ERROR(CV_StsBadArg, "All output matrices should have same type.");
	}

	vecsize = sumVector->rows * sumVector->cols;
	if (pcovMatrix->cols != vecsize || pcovMatrix->rows != vecsize)
	{
		CV_ERROR(CV_StsBadArg, "Size of output matrices do not match.");
	}

	CV_CALL( outtype = cvGetElemType(sumVector) );
	if (outtype != CV_64F && outtype != CV_32F)
	{
		CV_ERROR(CV_StsBadArg, "Inalid data type in input feature data.");
	}

	if(blockImgNum < 1) 
	{
		CV_ERROR(CV_StsBadArg, "Invalid input image data.");
	}

	for (i=0; i<blockImgNum; i++) 
	{
		if (srcImg[i] == NULL) 
		{
			CV_ERROR(CV_StsBadArg, "Invalid input image data.");
		}
	}

	// initialize matrices to 0
	cvSetZero(sumVector);
	cvSetZero(pcovMatrix);
			
	// calculate
	first = true;
	for (i=0; i<blockImgNum; i++) 
	{
		// check image
		if (first) {
			width = srcImg[i]->width;
			height = srcImg[i]->height;
			if (vecsize != width * height) {
				CV_ERROR(CV_StsBadSize, "Size of input image and output matrices do not match.");
			}
			widthstep = srcImg[i]->widthStep;
			band = srcImg[i]->nChannels;
			if (band != 1) 
			{
				CV_ERROR(CV_StsBadSize, "Invalid channel number of input image data.");
			}
			CV_CALL( inputdata = (float *) cvAlloc( vecsize * sizeof(float) ) );
			first = false;
		}
		else
		{
			if (srcImg[i]->width != width || srcImg[i]->height != height || srcImg[i]->nChannels != band)
			{
				CV_ERROR(CV_StsBadArg, "Input images are not identical.");
			}
		}
		// get input data
		imagebuf = (unsigned char *)srcImg[i]->imageData;
		j = 0;
		for (y=0; y<height; y++) 
		{
			index = y * widthstep;
			for (x=0; x<width; x++)
			{
				inputdata[j] = (float)(imagebuf[index+x]);
				j++;
			}
		}
		// compute
		if (outtype == CV_32F) 
		{
			for (j=0; j<vecsize; j++) 
			{
				// total
				tmpfl = inputdata[j];
				sumVector->data.fl[j] += tmpfl;
				index = j * vecsize;
				for (k=0; k<=j; k++) 
				{
					pcovMatrix->data.fl[index+k] += tmpfl * inputdata[k];
				}
			}
		}
		else		// CV_64F
		{
			for (j=0; j<vecsize; j++) 
			{
				// total
				tmpdb = (double)inputdata[j];
				sumVector->data.db[j] += tmpdb;
				index = j * vecsize;
				for (k=0; k<=j; k++) 
				{
					pcovMatrix->data.db[index+k] += tmpdb * inputdata[k];
				}
			}
		}
	}
		
	CV_CALL( cvFree( (void **) &inputdata ) );

	// copy the lower triangle to upper
	if (outtype == CV_32F) 
	{
		for (i=0; i<vecsize; i++) 
		{
			for (j=0; j<i; j++) 
			{
				pcovMatrix->data.fl[j*vecsize+i] = pcovMatrix->data.fl[i*vecsize+j];
			}
		}
	}
	else
	{
		for (i=0; i<vecsize; i++) 
		{
			for (j=0; j<i; j++) 
			{
				pcovMatrix->data.db[j*vecsize+i] = pcovMatrix->data.db[i*vecsize+j];
			}
		}
	}

	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvUpdatePCA(const IplImage** srcImg, const int blockImgNum, int &totalImgNum, 
//						CvMat* sumVector, CvMat* pcovMatrix);
//    Purpose:
//          Update the sum vector and the partial covariance matrix by using the group of input images.
//    Context:
//    Parameters:
//			srcImg ---- [I] pointer to the array of input IplImage images. All the images must be of 
//							the same size, 8-bit and with no ROI selected.
//			blockImgNum ---- [I] number of images in srcImage.
//			totalImgNum ---- [O] number of images used to calculate the input matrices, output the updated number. 
//			sumVector ---- [I/O] the sum vector. The function updates the current value of sumVector .
//			pcovMatrix ---- [I/O] the partial covariance matrix. The function updates updates 
//									the current value of pcovMatrix.
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvUpdatePCA(const IplImage** srcImg, const int blockImgNum, int &totalImgNum, CvMat* sumVector, CvMat* pcovMatrix)
{
	int i, j, k, x, y, index;
	int width, height, widthstep, vecsize, band, outtype;
	unsigned char *imagebuf;
	float *inputdata, tmpfl;
	double tmpdb;
	bool first;

	CV_FUNCNAME( "cvUpdatePCA" );

   	__BEGIN__;

	// check arguments
	if( !CV_IS_MAT(sumVector) || !CV_IS_MAT(pcovMatrix) ) 
	{
		CV_ERROR(CV_StsBadArg, "Inalid input matrix data.");
	}
	
	if( !CV_ARE_TYPES_EQ(sumVector, pcovMatrix) )
	{
		CV_ERROR(CV_StsBadArg, "All output matrices should have same type.");
	}

	vecsize = sumVector->cols * sumVector->rows;
	if (pcovMatrix->cols != vecsize || pcovMatrix->rows != vecsize)
	{
		CV_ERROR(CV_StsBadArg, "Size of output matrices do not match.");
	}

	CV_CALL( outtype = cvGetElemType(sumVector) );
	if (outtype != CV_64F && outtype != CV_32F)
	{
		CV_ERROR(CV_StsBadArg, "Inalid data type in input feature data.");
	}

	if(blockImgNum < 1) 
	{
		CV_ERROR(CV_StsBadArg, "Invalid input image data.");
	}

	for (i=0; i<blockImgNum; i++) 
	{
		if (srcImg[i] == NULL) 
		{
			CV_ERROR(CV_StsBadArg, "Invalid input image data.");
		}
	}

	// calculate
	first = true;
	for (i=0; i<blockImgNum; i++) 
	{
		// check image
		if (first) {
			width = srcImg[i]->width;
			height = srcImg[i]->height;
			if (vecsize != width * height) {
				CV_ERROR(CV_StsBadSize, "Size of input image and output matrices do not match.");
			}
			widthstep = srcImg[i]->widthStep;
			band = srcImg[i]->nChannels;
			if (band != 1) 
			{
				CV_ERROR(CV_StsBadSize, "Invalid channel number of input image data.");
			}
			CV_CALL( inputdata = (float *) cvAlloc( vecsize * sizeof(float) ) );
			first = false;
		}
		else
		{
			if (srcImg[i]->width != width || srcImg[i]->height != height || srcImg[i]->nChannels != band)
			{
				CV_ERROR(CV_StsBadArg, "Input images are not identical.");
			}
		}
		// get input data
		imagebuf = (unsigned char *)srcImg[i]->imageData;
		j = 0;
		for (y=0; y<height; y++) 
		{
			index = y * widthstep;
			for (x=0; x<width; x++)
			{
				inputdata[j] = (float)(imagebuf[index+x]);
				j++;
			}
		}
		// compute
		if (outtype == CV_32F) 
		{
			for (j=0; j<vecsize; j++) 
			{
				// total
				tmpfl = inputdata[j];
				sumVector->data.fl[j] += tmpfl;
				index = j * vecsize;
				for (k=0; k<=j; k++) 
				{
					pcovMatrix->data.fl[index+k] += tmpfl * inputdata[k];
				}
			}
		}
		else		// CV_64F
		{
			for (j=0; j<vecsize; j++) 
			{
				// total
				tmpdb = (double)inputdata[j];
				sumVector->data.db[j] += tmpdb;
				index = j * vecsize;
				for (k=0; k<=j; k++) 
				{
					pcovMatrix->data.db[index+k] += tmpdb * inputdata[k];
				}
			}
		}
	}
		
	CV_CALL( cvFree( (void **) &inputdata ) );

	// copy the lower triangle to upper
	if (outtype == CV_32F) 
	{
		for (i=0; i<vecsize; i++) 
		{
			for (j=0; j<i; j++) 
			{
				pcovMatrix->data.fl[j*vecsize+i] = pcovMatrix->data.fl[i*vecsize+j];
			}
		}
	}
	else
	{
		for (i=0; i<vecsize; i++) 
		{
			for (j=0; j<i; j++) 
			{
				pcovMatrix->data.db[j*vecsize+i] = pcovMatrix->data.db[i*vecsize+j];
			}
		}
	}

	totalImgNum += blockImgNum;

	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvCalcPCAMatrix( const int totalImgNum, const CvMat* sumVector, const CvMat* pcovMatrix,
							 CvMat* meanVector, CvMat* covMatrix, CvMat* eigenValues, CvMat* eigenVectors);
//    Purpose:
//			The function cvCalcPCAMatrix calculates the mean vector, the covariance matrix, 
//			and  the eigen values and eigen vectors of the covariance marix. The function takes as 
//			inputs totalNumImgs, sumVector and pcovMatrix obtained with cvInitPCA or cvUpdatePCA.
//    Context:
//    Parameters:
//			totalImgNum	---- [I] the number of input images used to calculate the matrices sumVector and pcovMatrix.
//			sumVector	---- [I] the summary vector calculated by function cvInitPCA or cvUpdatePCA. 
//			pcovMatrix	---- [I] the partial covariance matrix calculated by function cvInitPCA or cvUpdatePCA.
//			meanVector	---- [O] the mean vector. 
//			covMatrix	---- [O] the covariance matrix.
//			eigenValues ---- [O] the eigenvalues in descending order.
//			eigenVectors --- [O] the eigenvectors corresponding to the eigenvalues in eigenValues. The eigenvectors are stored 
//								 as the rows of the eigenVectors matrix.
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvCalcPCAMatrix( const int totalImgNum, const CvMat* sumVector, const CvMat* pcovMatrix,
				 CvMat* meanVector, CvMat* covMatrix, CvMat* eigenValues, CvMat* eigenVectors)
{
	int vecsize, type;
	int i, j, rltcopy;
	float tmpfl;
	double tmpdb;
	CvMat *tmpValue, *tmpVectors, *tmpCovar;

	CV_FUNCNAME("cvCalcPCAMatrix");

	__BEGIN__;
	
	// check arguments
	if ( totalImgNum < 1) 
	{
		CV_ERROR(CV_StsBadArg, "Object number must be larger than 1.");
	}

	if( !CV_IS_MAT(sumVector) || !CV_IS_MAT(pcovMatrix) || !CV_IS_MAT(meanVector) || 
		!CV_IS_MAT(covMatrix) || !CV_IS_MAT(eigenValues) || !CV_IS_MAT(eigenVectors) ) 
	{
		CV_ERROR(CV_StsBadArg, "Inalid input matrix data.");
	}
	
	if( !CV_ARE_TYPES_EQ(sumVector, pcovMatrix) || !CV_ARE_TYPES_EQ(sumVector, meanVector) || 
		!CV_ARE_TYPES_EQ(sumVector, covMatrix) || !CV_ARE_TYPES_EQ(sumVector, eigenValues) || 
		!CV_ARE_TYPES_EQ(sumVector, eigenVectors) )
	{
		CV_ERROR(CV_StsBadArg, "All input and output matrices should have same type.");
	}

	vecsize = sumVector->rows * sumVector->cols;
	if ( pcovMatrix->rows != vecsize || pcovMatrix->cols != vecsize || 
		 meanVector->rows != vecsize || covMatrix->rows != vecsize || 
		 covMatrix->cols != vecsize || eigenVectors->cols != vecsize )
	{
		CV_ERROR(CV_StsBadSize, "Sizes of input and output matrices do not match.");
	}

	CV_CALL( type = cvGetElemType(sumVector) );
	if ( type == CV_32F)
	{
		// compute mean and mean square error
		for (i=0; i<vecsize; i++) 
		{
			tmpfl = sumVector->data.fl[i] / totalImgNum;
			meanVector->data.fl[i] = tmpfl;
		}
		// compute covariance
		for (i=0; i<vecsize; i++) 
		{
			tmpfl = meanVector->data.fl[i];
			for (j=0; j<vecsize; j++) 
			{
				covMatrix->data.fl[i*vecsize+j] = pcovMatrix->data.fl[i*vecsize+j] / totalImgNum - tmpfl * meanVector->data.fl[j];
			}
		}
		// decomposition
		CV_CALL( tmpValue = cvCreateMat(vecsize, 1, CV_32F) );
		CV_CALL( tmpVectors = cvCreateMat(vecsize, vecsize, CV_32F) );
		CV_CALL( tmpCovar = cvCreateMat(vecsize, vecsize, CV_32F) );
		CV_CALL( cvmCopy(covMatrix, tmpCovar) );
		CV_CALL ( cvmEigenVV( tmpCovar, tmpVectors, tmpValue, 0) );
		// copy eigen vectors
		rltcopy = eigenVectors->rows > vecsize ? vecsize : eigenVectors->rows;
		for (i=0; i<rltcopy; i++)
		{
			for (j=0; j<vecsize; j++) 
			{
				eigenVectors->data.fl[i*vecsize+j] = tmpVectors->data.fl[i*vecsize+j]; 
			}
		}
		// copy eigen value
		rltcopy = eigenValues->rows > vecsize ? vecsize : eigenValues->rows;
		for (i=0; i<rltcopy; i++)
		{
			eigenValues->data.fl[i] = tmpValue->data.fl[i];
		}
		CV_CALL( cvReleaseMat(&tmpValue) );
		CV_CALL( cvReleaseMat(&tmpVectors) );
		CV_CALL( cvReleaseMat(&tmpCovar) );
	}
	else if (type == CV_64F)
	{
		// compute mean and mean square error
		for (i=0; i<vecsize; i++) 
		{
			tmpdb = sumVector->data.db[i];
			meanVector->data.db[i] = tmpdb / totalImgNum;
		}
		// compute covariance
		for (i=0; i<vecsize; i++) 
		{
			tmpdb = meanVector->data.db[i];
			for (j=0; j<vecsize; j++) 
			{
				covMatrix->data.db[i*vecsize+j] = pcovMatrix->data.db[i*vecsize+j] / totalImgNum - tmpdb * meanVector->data.db[j];
			}
		}
		// decomposition
		CV_CALL( tmpValue = cvCreateMat(vecsize, 1, CV_64F) );
		CV_CALL( tmpVectors = cvCreateMat(vecsize, vecsize, CV_64F) );
		CV_CALL( tmpCovar = cvCreateMat(vecsize, vecsize, CV_64F) );
		CV_CALL( cvmCopy(covMatrix, tmpCovar) );
		CV_CALL ( cvmEigenVV( tmpCovar, tmpVectors, tmpValue, 0) );
		// copy eigen vectors
		rltcopy = eigenVectors->rows > vecsize ? vecsize : eigenVectors->rows;
		for (i=0; i<rltcopy; i++)
		{
			for (j=0; j<vecsize; j++) 
			{
				eigenVectors->data.db[i*vecsize+j] = tmpVectors->data.db[i*vecsize+j]; 
			}
		}
		// copy eigen value
		rltcopy = eigenValues->rows > vecsize ? vecsize : eigenValues->rows;
		for (i=0; i<rltcopy; i++)
		{
			eigenValues->data.db[i] = tmpValue->data.db[i];
		}
		CV_CALL( cvReleaseMat(&tmpValue) );
		CV_CALL( cvReleaseMat(&tmpVectors) );
		CV_CALL( cvReleaseMat(&tmpCovar) );
	}
	else 
	{
		CV_ERROR(CV_StsBadArg, "Invalid data type of matrices.");
	}
	
	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:
//			cvPCAProjection(const IplImage** srcImg, const int imgNum, 
//							const CvMat* eigenValues, const CvMat* eigenVectors, const CvMat* meanVector, 
							CvMat* projVectors, bool eigenNorm  CV_DEFAULT(false));
//    Purpose:
//			Transform the input images to the subspace obtained by PCA.
//    Context:
//    Parameters:
//			srcImg ---- [I] pointer to the array of input IplImage images. All the images must be 
//							of the same size, and have 8-bit depth.
//			imgNum ---- [I] the number of input images. 
//			eigenValues ---- [I] the p largest eigenvalues obtained in PCA if the images will be 
//								 transformed to p-dimensional subspace.
//			eigenVectors --- [I] the matrix in which each row is the eigenvector corresponding to 
//								 the eigenvalue in eigenValue
//			meanVector	---- [I] the mean vector of the training images that used in PCA to obtain 
//								 the eigenvalues and eigenvectors.
//			projVectors ---- [O] output matrix in which each row is one feature vector in 
//								 p-dimensional subspace transformed from the corresponding image in srcImg.
//			eigenNorm	---- [I] Select if the transform result will be normalized using eigenvalues.
//    Returns:
//			return CV_StsOk if successes, CV_StsError for failed or error
//    Notes:
//F*/
CV_IMPL CVStatus 
cvPCAProjection(const IplImage** srcImg, const int imgNum, 
				const CvMat* eigenValues, const CvMat* eigenVectors, const CvMat* meanVector, 
				CvMat* projVectors, bool eigenNorm)
{
	int n, i, j, x, y, index, width, height, widthstep, band;
	float *inputdata, tmpfl;
	unsigned char *imagebuf;
	double tmpdb;
	bool first;
	double *eigenvsqrt;
	
	CV_FUNCNAME("cvPCAProjection");

	__BEGIN__;
	
	// check arguments
	if ( imgNum < 1 || srcImg == NULL) 
	{
		CV_ERROR(CV_StsBadArg, "Invalid input images.");
	}

	if( (eigenNorm && !CV_IS_MAT(eigenValues)) || !CV_IS_MAT(eigenVectors) || 
		!CV_IS_MAT(meanVector) || !CV_IS_MAT(projVectors)  ) 
	{
		CV_ERROR(CV_StsBadArg, "Inalid input matrix data.");
	}
	
	if( !CV_ARE_TYPES_EQ(eigenVectors, meanVector) || (eigenNorm && !CV_ARE_TYPES_EQ(eigenVectors, eigenValues)) )
	{
		CV_ERROR(CV_StsBadArg, "All input matrices should have same type.");
	}

	int typein, typeout;
	CV_CALL( typein = cvGetElemType(eigenVectors) );
	CV_CALL( typeout = cvGetElemType(projVectors) );
	if (typein != CV_32F && typein != CV_64F && typeout != CV_32F && typeout != CV_64F)
	{
		CV_ERROR(CV_StsBadArg, "Invalid data type of input or output matrices.");
	}

	int invecsize, outvecsize;
	invecsize = eigenVectors->cols;
	outvecsize = projVectors->cols;
	if ( outvecsize > eigenVectors->rows )
	{
		CV_ERROR(CV_StsBadSize, "Size of output vector is larger than number of eigen vectors.");
	}
	if ( projVectors->rows != imgNum || meanVector->rows != invecsize || 
		 (eigenNorm && eigenValues->rows != invecsize) )
	{
		CV_ERROR(CV_StsBadSize, "Size of input and output data do not match.");
	}

	// compute square root of eigen values for normalization
	if (eigenNorm)
	{
		CV_CALL( eigenvsqrt = (double *) cvAlloc( invecsize * sizeof(double) ) );
		if (typein == CV_32F)
		{
			for ( i=0; i<invecsize; i++ )
			{
				eigenvsqrt[i] = sqrt(eigenValues->data.fl[i]);
			}
		}
		else
		{
			for ( i=0; i<invecsize; i++ )
			{
				eigenvsqrt[i] = sqrt(eigenValues->data.db[i]);
			}
		}
	}
	// compute projection
	first = true;
	for (n=0; n<imgNum; n++)
	{
		// check image
		if (first) {
			width = srcImg[n]->width;
			height = srcImg[n]->height;
			if (invecsize != width * height) {
				CV_ERROR(CV_StsBadSize, "Size of input image and output matrices do not match.");
			}
			widthstep = srcImg[n]->widthStep;
			band = srcImg[n]->nChannels;
			if (band != 1) 
			{
				CV_ERROR(CV_StsBadSize, "Invalid channel number of input image data.");
			}
			CV_CALL( inputdata = (float *) cvAlloc( invecsize * sizeof(float) ) );
			first = false;
		}
		else
		{
			if (srcImg[n]->width != width || srcImg[n]->height != height || srcImg[n]->nChannels != band)
			{
				CV_ERROR(CV_StsBadArg, "Input images are not identical.");
			}
		}
		// get input data
		imagebuf = (unsigned char *)srcImg[n]->imageData;
		index = 0;
		for (y=0; y<height; y++) 
		{
			for (x=0; x<width; x++)
			{
				inputdata[index] = (float)(imagebuf[y*widthstep+x]);
				index ++;
			}
		}
		// compute
		if ( typein == CV_32F)
		{
			if ( typeout == CV_32F )
			{
				// minus mean
				for (j=0; j<invecsize; j++) 
				{
					inputdata[j] = inputdata[j] - meanVector->data.fl[j];
				}
				// multiply matrix
				for (i=0; i<outvecsize; i++)
				{
					tmpfl = 0;
					index = i * invecsize;
					for (j=0; j<invecsize; j++) {
						tmpfl += eigenVectors->data.fl[index+j] * inputdata[j];
					}
					if (eigenNorm)
					{
						projVectors->data.fl[n*outvecsize+i] = tmpfl / (float)eigenvsqrt[i];
					}
					else
					{
						projVectors->data.fl[n*outvecsize+i] = tmpfl;
					}
				}
			}
			else		// typeout == CV_64F
			{
				// minus mean
				for (j=0; j<invecsize; j++) 
				{
					inputdata[j] = inputdata[j] - meanVector->data.fl[j];
				}
				// multiply matrix
				for (i=0; i<outvecsize; i++)
				{
					tmpfl = 0;
					index = i * invecsize;
					for (j=0; j<invecsize; j++) {
						tmpfl += eigenVectors->data.fl[index+j] * inputdata[j];
					}
					if (eigenNorm)
					{
						projVectors->data.db[n*outvecsize+i] = tmpfl / eigenvsqrt[i];
					}
					else
					{
						projVectors->data.db[n*outvecsize+i] = tmpfl;
					}
				}
			}
		}
		else	// typein == CV_64F
		{
			if ( typeout == CV_32F )
			{
				// minus mean
				for (j=0; j<invecsize; j++) 
				{
					inputdata[j] = inputdata[j] - (float)meanVector->data.db[j];
				}
				// multiply matrix
				for (i=0; i<outvecsize; i++)
				{
					tmpdb = 0;
					index = i * invecsize;
					for (j=0; j<invecsize; j++) {
						tmpdb += eigenVectors->data.db[index+j] * inputdata[j];
					}
					if (eigenNorm)
					{
						projVectors->data.fl[n*outvecsize+i] = (float)(tmpdb / eigenvsqrt[i]);
					}
					else
					{
						projVectors->data.fl[n*outvecsize+i] = (float)tmpdb;
					}
				}
			}
			else	// typeout == CV_64F
			{
				// minus mean
				for (j=0; j<invecsize; j++) 
				{
					inputdata[j] = inputdata[j] - (float)meanVector->data.db[j];
				}
				// multiply matrix
				for (i=0; i<outvecsize; i++)
				{
					tmpdb = 0;
					index = i * invecsize;
					for (j=0; j<invecsize; j++) {
						tmpdb += eigenVectors->data.db[index+j] * inputdata[j];
					}
					if (eigenNorm)
					{
						projVectors->data.db[n*outvecsize+i] = tmpdb / eigenvsqrt[i];
					}
					else
					{
						projVectors->data.db[n*outvecsize+i] = tmpdb;
					}
				}
			}
		}
	}

	if (eigenNorm)
	{
		CV_CALL( cvFree( (void **) &eigenvsqrt ) );
	}
	CV_CALL( cvFree( (void **) &inputdata ) );
	
	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}


