/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "cvvisualspeech.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			icvFT1D( const CvMat* inData, CvMat* outRe, CvMat* outIm )
//    Purpose: 
//			Internal function of doing 1-dimention FT
//    Context:
//    Parameters:
//			inData -- [I] input vector.
//			outRe  -- [O] real part of output vector. 
//			outIm  -- [O] image part of output vector. 
//    Returns:
//    Notes:
//F*/
CV_IMPL void
icvFT1D( const CvMat* inData, CvMat* outRe, CvMat* outIm )
{
	int N,i,j;
	int type;
	CvMat *sin_coe = NULL;
	CvMat *cos_coe = NULL;
	double angle,temp;	
	
    CV_FUNCNAME( "icvFT1D" );

    __BEGIN__;
	
	// check arguments
	if( !CV_IS_MAT(inData) || !CV_IS_MAT(outRe) || !CV_IS_MAT(outIm) )
		if( !CV_ARE_TYPES_EQ( inData, outRe ) || !CV_ARE_TYPES_EQ( inData, outIm ))
			CV_ERROR( CV_StsUnmatchedFormats, "All matrices must have the same type." );

	if( !(inData->cols == outRe->cols) || !(outRe->cols == outIm->cols) 
		|| !(inData->rows == outRe->rows) || !(outRe->rows == outIm->rows) )
        CV_ERROR( CV_StsUnmatchedSizes, "All matrices must have the same size." );

    if( inData->cols < 1 || outRe->cols < 1 || outIm->cols < 1 )
        CV_ERROR( CV_StsBadSize, "Invalid vector length." );


    if ( MIN( inData->rows, inData->cols ) < 1 
		 || MIN( outRe->rows, outRe->cols ) < 1 
		 || MIN( outIm->rows, outIm->cols ) < 1 )
		CV_ERROR( CV_StsBadSize, "Number of points must be >= 1." );
	    
	CV_CALL( type = cvGetElemType(inData) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR( CV_StsUnsupportedFormat, "Input data type must be float or double." );

	N = inData->cols;
	angle = 2.0 * CV_PI / N;	
	CV_CALL( sin_coe = cvCreateMat(1, N, type) );
	CV_CALL( cos_coe = cvCreateMat(1, N, type) );
	if(type == CV_32F )
	{
		for(i=0; i<N; i++)
		{
			temp = angle * i;
			sin_coe->data.fl[i] = (float)sin(temp);
			cos_coe->data.fl[i] = (float)cos(temp);
		}

		for(i=0;i<N;i++)
		{
			outRe->data.fl[i] = outIm->data.fl[i] = 0.0;
			for(j=0; j<N; j++) 
			{
				int index = (i * j) % N;		
				float sin_val = sin_coe->data.fl[index];
				float cos_val = cos_coe->data.fl[index];
				outRe->data.fl[i] += cos_val * inData->data.fl[j];
				outIm->data.fl[i] += sin_val * inData->data.fl[j];
			}
		}
	}
	else
	{
		for(i=0; i<N; i++)
		{
			temp = angle * i;
			sin_coe->data.db[i] = sin(temp);
			cos_coe->data.db[i] = cos(temp);
		}
		for(i=0;i<N;i++)
		{
			outRe->data.db[i] = outIm->data.db[i] = 0.0;
			for(j=0; j<N; j++) 
			{
				int index = (i * j) % N;		
				double sin_val = sin_coe->data.db[index];
				double cos_val = cos_coe->data.db[index];
				outRe->data.db[i] += cos_val * inData->data.db[j];
				outIm->data.db[i] += sin_val * inData->data.db[j];
			}
		}
	}
	
	cvReleaseMat(&sin_coe);
	cvReleaseMat(&cos_coe);

	__END__;
	
	cvFree((void**)&sin_coe);
	cvFree((void**)&cos_coe);
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			icvIFT1D( const CvMat* inRe, const CvMat* inIm, CvMat* outData )
//    Purpose: 
//			Internal function of doing 1-dimention inverse FT
//    Context:
//    Parameters:
//			inRe    -- [I] real part of input vector. 
//			inIm    -- [I] image part of output vector. 
//			outData -- [O] output vector.
//    Returns:
//    Notes:
//F*/

CV_IMPL void
icvIFT1D( const CvMat* inRe, const CvMat* inIm, CvMat* outData )
{
	int N,i,j;
	int type;
	CvMat *sin_coe = NULL;
	CvMat *cos_coe = NULL;
	double angle,temp;	
	
    CV_FUNCNAME( "icvFT1D" );

    __BEGIN__;
	
	// check arguments
	if( !CV_IS_MAT(outData) || !CV_IS_MAT(inRe) || !CV_IS_MAT(inIm) )
		if( !CV_ARE_TYPES_EQ( outData, inRe ) || !CV_ARE_TYPES_EQ( outData, inIm ))
			CV_ERROR( CV_StsUnmatchedFormats, "All matrices must have the same type." );

	if( !(outData->cols == inRe->cols) || !(inRe->cols == inIm->cols) 
		|| !(outData->rows == inRe->rows) || !(inRe->rows == inIm->rows) )
         CV_ERROR( CV_StsUnmatchedSizes, "All matrices must have the same size." );

    if( outData->cols < 1 || inRe->cols < 1 || inIm->cols < 1 )
        CV_ERROR( CV_StsBadSize, "Invalid vector length." );


    if ( MIN( outData->rows, outData->cols ) < 1 
		 || MIN( inRe->rows, inRe->cols ) < 1 
		 || MIN( inIm->rows, inIm->cols ) < 1 )
		CV_ERROR( CV_StsBadSize, "Number of points must >= 1." );

	CV_CALL( type = cvGetElemType(outData) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR( CV_StsUnsupportedFormat, "Input data type must be float or double." );

	N = inRe->cols;
	angle = -2.0 * CV_PI / N;
	CV_CALL( sin_coe = cvCreateMat(1, N, type) );
	CV_CALL( cos_coe = cvCreateMat(1, N, type) );
	if(type == CV_32F )
	{	
		for(i=0; i<N; i++)
		{
			temp = angle * i;
			sin_coe->data.fl[i] = (float)sin(temp);
			cos_coe->data.fl[i] = (float)cos(temp);
		}

		for(i=0;i<N;i++)
		{
			outData->data.fl[i] = 0.0;
			for(j=0; j<N; j++) 
			{
				int index = (i * j) % N;			
				float sin_val = sin_coe->data.fl[index];
				float cos_val = cos_coe->data.fl[index];
				outData->data.fl[i] += cos_val * inRe->data.fl[j] - sin_val * inIm->data.fl[j];			
			}
			outData->data.fl[i] = outData->data.fl[i] / N;
		}
	}
	else
	{
		for(i=0; i<N; i++)
		{
			temp = angle * i;
			sin_coe->data.db[i] = sin(temp);
			cos_coe->data.db[i] = cos(temp);
		}

		for(i=0;i<N;i++)
		{
			outData->data.fl[i] = 0.0;
			for(j=0; j<N; j++) 
			{
				int index = (i * j) % N;			
				double sin_val = sin_coe->data.db[index];
				double cos_val = cos_coe->data.db[index];
				outData->data.db[i] += cos_val * inRe->data.db[j] - sin_val * inIm->data.db[j];
			}
			outData->data.db[i] = outData->data.db[i] / N;
		}
	}

	cvReleaseMat(&sin_coe);
	cvReleaseMat(&cos_coe);
	
	__END__;

	cvFree((void**)&sin_coe);
	cvFree((void**)&cos_coe);
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name:    
//			cvInterpolateFT( const CvMat* inVectors, const float R, CvMat* outVectors );
//    Purpose: 
//			The function cvInterpolateFT interpolates to the input vectors using the Fourier 
//			transform. The interpolation is done on each column of the input feature inVectors 
//			and the result is stored in outVectors that must be allocated before calling 
//			this function. Assuming M N-dimensional input vectors stored in a MN matrix 
//			inVectors, the matrix outVectors must be size of (MR) N.
//    Context:
//    Parameters:
//			inVectors -- [I] Input vectors, stored as rows of inVectors
//			R		-- [I] Interpolation rate, must be larger than 1 
//			ldaout	-- [O] Output vectors stored as the rows of outVectors   
//    Returns:
//			return CV_StsOk if upsample successfully, else report error.
//    Notes:
//F*/

CV_IMPL CVStatus
cvInterpolateFT( const CvMat* inVectors, const float R, CvMat* ldaout )
{
	CvMat *VectorFFTRe = 0; 
	CvMat *VectorFFTIm = 0;
	CvMat *NyqstFFTRe = 0;
	CvMat *NyqstFFTIm = 0;
	CvMat *InColumn = 0;
	CvMat *OutColumn = 0;
	int type, vecLen, vecNum, featureNum;
	int NyqstNum, i, j, n, m;
	bool IsEven;

	CV_FUNCNAME( "cvInterpolateFT" );

    __BEGIN__;
	
	// check arguments
	if( !CV_IS_MAT(inVectors) || !CV_IS_MAT(ldaout) )
		if( !CV_ARE_TYPES_EQ( inVectors, ldaout ) )
			CV_ERROR( CV_StsUnmatchedFormats, "All matrices must have the same type." );

	if ( inVectors->cols < 1 )
		CV_ERROR( CV_StsBadSize, "Vector length must be >= 1." );
	
	if ( inVectors->rows < 1 )
		CV_ERROR( CV_StsBadSize, "Vector number must be >= 1." );
	
	if ( R < 1 )
		CV_ERROR( CV_StsBadArg, "Invalid upsample rate." );
	
	CV_CALL( type = cvGetElemType(inVectors) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR( CV_StsUnsupportedFormat, "Input data type must be float or double." );

	vecLen = inVectors->cols;
	vecNum = inVectors->rows;
	featureNum = ldaout->rows;

	CV_CALL( VectorFFTRe  =  cvCreateMat(1, vecNum, type) );
	CV_CALL( VectorFFTIm  =  cvCreateMat(1, vecNum, type) );
	CV_CALL( InColumn  =  cvCreateMat(1, vecNum, type) );
	
	CV_CALL( NyqstFFTRe  =  cvCreateMat(1, featureNum, type) );
	CV_CALL( NyqstFFTIm  =  cvCreateMat(1, featureNum, type) );
	CV_CALL( OutColumn  =  cvCreateMat(1, featureNum, type) );
	
	IsEven = (vecNum % 2 == 0);		
	NyqstNum = cvCeil((vecNum + 1) / 2.0f);
	
	if( type == CV_32F )
	{
		// column by column
		for (i=0; i<vecLen; i++) 
		{
			// extract column and expend			
			for (j=0, n=0; j<vecNum; j++, n++) 
			{
				InColumn->data.fl[n] = inVectors->data.fl[i+j*vecLen];
			}
			
			// do FFT
			CV_CALL( icvFT1D( InColumn, VectorFFTRe, VectorFFTIm ) );
			
			// Nyqst expend
			n = 0;
			for ( ; n<(NyqstNum-1); n++) 
			{
				NyqstFFTRe->data.fl[n] = VectorFFTRe->data.fl[n];
				NyqstFFTIm->data.fl[n] = VectorFFTIm->data.fl[n];
			}
			m = n;
			NyqstFFTRe->data.fl[n] = IsEven ? VectorFFTRe->data.fl[m] / 2 : VectorFFTRe->data.fl[m];
			NyqstFFTIm->data.fl[n] = IsEven ? VectorFFTIm->data.fl[m] / 2 : VectorFFTIm->data.fl[m];
			n++;	// m move back
			for ( ; n < ( featureNum-(vecNum-NyqstNum+1) ); n++)
			{
				NyqstFFTRe->data.fl[n] = 0.0f;
				NyqstFFTIm->data.fl[n] = 0.0f;
			}
			NyqstFFTRe->data.fl[n] = IsEven ? VectorFFTRe->data.fl[m] / 2 : 0;
			NyqstFFTIm->data.fl[n] = IsEven ? VectorFFTIm->data.fl[m] / 2 : 0;
			n++; m++;
			for ( ; n<featureNum; n++, m++) 
			{
				NyqstFFTRe->data.fl[n] = VectorFFTRe->data.fl[m];
				NyqstFFTIm->data.fl[n] = VectorFFTIm->data.fl[m];
			}
			
			// invert FFT
			CV_CALL( icvIFT1D( NyqstFFTRe, NyqstFFTIm, OutColumn) );
			
			// extract result data			
			float factor = (float)featureNum / (float)vecNum;
			for (j=0, n=0; j<featureNum; j++, n++) 
			{
				ldaout->data.fl[i+j*vecLen] = OutColumn->data.fl[n] * factor;				
			}
		}
	}
	else
	{
		// column by column
		for (i=0; i<vecLen; i++) 
		{
			// extract column and expend			
			for (j=0, n=0; j<vecNum; j++, n++) 
			{
				InColumn->data.db[n] = inVectors->data.db[i+j*vecLen];	
			}
			
			// do FFT
			CV_CALL( icvFT1D( InColumn, VectorFFTRe, VectorFFTIm ) );
			
			// Nyqst expend
			n = 0;
			for ( ; n<(NyqstNum-1); n++) 
			{
				NyqstFFTRe->data.db[n] = VectorFFTRe->data.db[n];
				NyqstFFTIm->data.db[n] = VectorFFTIm->data.db[n];
			}
			m = n;
			NyqstFFTRe->data.db[n] = IsEven ? VectorFFTRe->data.db[m] / 2 : VectorFFTRe->data.db[m];
			NyqstFFTIm->data.db[n] = IsEven ? VectorFFTIm->data.db[m] / 2 : VectorFFTIm->data.db[m];
			n++;	// m move back
			for ( ; n < ( featureNum-(vecNum-NyqstNum+1) ); n++)
			{
				NyqstFFTRe->data.db[n] = 0.0f;
				NyqstFFTIm->data.db[n] = 0.0f;
			}
			NyqstFFTRe->data.db[n] = IsEven ? VectorFFTRe->data.db[m] / 2 : 0;
			NyqstFFTIm->data.db[n] = IsEven ? VectorFFTIm->data.db[m] / 2 : 0;
			n++; m++;
			for ( ; n<featureNum; n++, m++) 
			{
				NyqstFFTRe->data.db[n] = VectorFFTRe->data.db[m];
				NyqstFFTIm->data.db[n] = VectorFFTIm->data.db[m];
			}
			
			// invert FFT
			CV_CALL( icvIFT1D( NyqstFFTRe, NyqstFFTIm, OutColumn) );
			
			// extract result data			
			double factor = (double)featureNum / (double)vecNum;
			for (j=0, n=0; j<featureNum; j++, n++) 
			{
				ldaout->data.db[i+j*vecLen] = OutColumn->data.db[n] * factor;
			}
		}
	}
	
	/* Free data */
	cvReleaseMat(&VectorFFTRe);
    cvReleaseMat(&VectorFFTIm);
	cvReleaseMat(&InColumn);
	cvReleaseMat(&NyqstFFTRe);
	cvReleaseMat(&NyqstFFTIm);
	cvReleaseMat(&OutColumn);

	return CV_StsOk;

	__END__;
	
	cvFree((void**)&VectorFFTRe);
	cvFree((void**)&VectorFFTIm);
	cvFree((void**)&InColumn);
	cvFree((void**)&NyqstFFTRe);
	cvFree((void**)&NyqstFFTIm);
	cvFree((void**)&OutColumn);
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			cvConcatVectors(const CvMat* inVectors, const int J, CvMat* outVectors );
//    Purpose:
//			The function cvConcatVectors concatenates the J consecutive feature vectors to 
//			one new vector to capture the dynamic information in the feature data.  
//			The matrix outVectors must be allocated before the function call. 
//    Context:
//    Parameters:
//			inVectors	--	[I] the input vectors stored as the rows of the inVectors matrix.
//			J			-- [I] the number of input vectors concatenated to form one output vector.
//			outVectors	--  [O] the output concatenated vector stored as the rows of outVectors.
//    Returns:
//			return CV_StsOk if concatenation successfully, else report error.
//    Notes:
//F*/

CV_IMPL CVStatus  
cvConcatVectors( const CvMat* inVectors, const int J, CvMat* outVectors )
{
	int WindowRadius, i, j, index;
	int type, invecLen, invecNum, outvecLen;

	CV_FUNCNAME( "cvConcatVectors" );

    __BEGIN__;
	
	// check arguments
	if ( !CV_IS_MAT(inVectors) || !CV_IS_MAT_HDR(outVectors) )
		if ( !CV_ARE_TYPES_EQ( inVectors, outVectors ) )
			CV_ERROR( CV_StsUnmatchedFormats, "All input/output matrices must have the same type." );

    if ( MIN( inVectors->rows, inVectors->cols ) < 1 || MIN( outVectors->rows, outVectors->cols ) < 1)
		CV_ERROR( CV_StsBadSize, "Number of sample data must be >= 1." );

	if ( J > inVectors->rows )
		CV_ERROR( CV_StsBadArg, "Window Width must be <= sample number." );
	
	CV_CALL( type = cvGetElemType(inVectors) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR( CV_StsUnsupportedFormat, "Input data type must be float or double." );

	if ( J % 2 !=0 )	// J is odd.
		WindowRadius = ( J - 1 ) / 2;
	else
		WindowRadius = J / 2;
	
	invecLen = inVectors->cols;
	invecNum = inVectors->rows;
	outvecLen = outVectors->cols;

	// concatenation
	if( type == CV_32F )
	{
		// set Fout(i) (i>= m) to be [Fin(i-m), ..., Fin(i+m)]
		for (i=WindowRadius; i < invecNum-WindowRadius; i++)
		{
			float *BufOut = &( outVectors->data.fl[outvecLen * i] );
			float *BufIn = &( inVectors->data.fl[invecLen * (i-WindowRadius)] );
			for (j=0; j<J; j++)
			{
				index = invecLen * j;
				memcpy( &(BufOut[index]), &(BufIn[index]), invecLen * sizeof(float) );
			}
		}
		// set the first m frames Fout(i) be the Fout(m+1)
		for (i = 0; i < WindowRadius; i++) 
		{
			memcpy( &(outVectors->data.fl[outvecLen * i]), &(outVectors->data.fl[outvecLen * WindowRadius]), outvecLen * sizeof(float));
		}
		// set the last m frames Fout(i) be the Fout(n-m-1)
		for (i = invecNum - WindowRadius; i < invecNum; i++) 
		{
			memcpy( &(outVectors->data.fl[outvecLen * i]), &(outVectors->data.fl[outvecLen * (invecNum - WindowRadius-1)]), outvecLen * sizeof(float));
		}
	}
	else
	{
		// set Fout(i) (i>= m) to be [Fin(i-m), ..., Fin(i+m)]
		for (i=WindowRadius; i < invecNum-WindowRadius; i++)
		{
			double *BufOut = &( outVectors->data.db[outvecLen * i] );
			double *BufIn = &( inVectors->data.db[invecLen * (i-WindowRadius)] );
			for (j=0; j<J; j++)
			{
				index = invecLen * j;
				memcpy( &(BufOut[index]), &(BufIn[index]), invecLen * sizeof(double) );
			}
		}
		// set the first m frames Fout(i) be the Fout(m+1)
		for (i = 0; i < WindowRadius; i++) 
		{
			memcpy( &(outVectors->data.db[outvecLen * i]), &(outVectors->data.db[outvecLen * WindowRadius]), outvecLen * sizeof(double));
		}
		// set the last m frames Fout(i) be the Fout(n-m-1)
		for (i = invecNum - WindowRadius; i < invecNum; i++) 
		{
			memcpy( &(outVectors->data.db[outvecLen * i]), &(outVectors->data.db[outvecLen * (invecNum - WindowRadius-1)]), outvecLen * sizeof(double));
		}
	}

	return CV_StsOk;

	__END__;
	
	return CV_StsError;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//
//    Name: 
//			cvNormalizeMean( CvMat* inVectors );
//    Purpose:
//			The function cvNormalizeMean subtracts the mean from the input vectors to the feature
//			 vectors. The mean is computed over the whole set of input vectors.
//    Context:
//    Parameters:
//			inVectors -- [I/O] Feature data matrix, each row is a feature vector
//    Returns: 
//			return CV_StsOk if FMN successfully, else report error.
//    Notes:
//F*/

CV_IMPL CVStatus
cvNormalizeMean( CvMat* inVectors )
{
	int i, j, index, vecLen, vecNum;
	int type;
	CvMat *mean = NULL;

	CV_FUNCNAME( "cvNormalizeMean" );

    __BEGIN__;
	
	// check arguments
	if ( !CV_IS_MAT(inVectors) )
		CV_ERROR( CV_StsUnmatchedFormats, "Invalid input argument." );

    if ( MIN( inVectors->rows, inVectors->cols ) < 1 )
		CV_ERROR( CV_StsBadSize, "Number of sample data must be >= 1." );

	CV_CALL( type = cvGetElemType(inVectors) );
	if( type != CV_32F && type != CV_64F )
		CV_ERROR( CV_StsUnsupportedFormat, "Input data type must be float or double." );
	
	vecLen = inVectors->cols;
	vecNum = inVectors->rows;
	// allocate mean buffer
	CV_CALL( mean = cvCreateMat(1, vecLen, type) );
	cvSetZero(mean);
	if(type == CV_32F)
	{
		// calculate mean vector
		for (i=0; i < vecNum; i++) 
		{
			index = vecLen * i;
			for (j=0; j < vecLen; j++) 
			{
				mean->data.fl[j] += inVectors->data.fl[index+j];
			}
		}
		for (j=0; j < vecLen; j++) 
		{
			mean->data.fl[j] /=  vecNum;
		}

		// normalize
		for (i=0; i < vecNum; i++) 
		{
			index = vecLen * i;
			for (j=0; j < vecLen; j++) 
			{
				inVectors->data.fl[index+j] -= mean->data.fl[j];
			}
		}
	}
	else
	{
		// calculate mean vector
		for (i=0; i < vecNum; i++) 
		{
			index = vecLen * i;
			for (j=0; j < vecLen; j++) 
			{
				mean->data.db[j] += inVectors->data.db[index+j];
			}
		}
		for (j=0; j < vecLen; j++) 
		{
			mean->data.db[j] /=  vecNum;
		}

		// normalize
		for (i=0; i < vecNum; i++) 
		{
			index = vecLen * i;
			for (j=0; j < vecLen; j++) 
			{
				inVectors->data.db[index+j] -= mean->data.db[j];
			}
		}
	}

	cvReleaseMat(&mean);
	
	return CV_StsOk;

	__END__;
	
	cvFree((void**)&mean);

	return CV_StsError;
}
